@@ -104,7 +104,7 @@ run_test() {
 
     edo \
         timeout -sKILL ${SYD_TEST_TIMEOUT} \
-        "${SYD}" -ppaludis -m 'allow/all+/***' -mlock:on \
+        ${SYD} -ppaludis -m 'allow/all+/***' -mlock:on \
         -- make -j check
     r=$?
 
