@@ -39,13 +39,14 @@ use nix::{
 use once_cell::sync::Lazy;
 
 use crate::{
-    compat::{STATX_MNT_ID, STATX_MNT_ID_UNIQUE},
+    check_cross_memory_attach,
+    compat::{STATX_MNT_ID, STATX_MNT_ID_UNIQUE, STATX_MODE},
     fs::{duprand, FileMap, FileMapEntry, FileType},
     fstatx, info,
     path::{XPath, XPathBuf},
     proc::proc_mmap_min_addr,
+    sandbox::Version,
     sysinfo::RandTimer,
-    workers::PidFdMap,
 };
 
 // Convenience type to list {io,pr}ctls with their names.
@@ -65,6 +66,8 @@ pub const ESYD_SH: &str = include_str!("esyd.sh");
 /// The contents of the file `syd.el`.
 pub const SYD_EL: &str = include_str!("syd.el");
 
+/// The environment variable to override the host Linux kernel version.
+pub const ENV_ASSUME_KERNEL: &str = "SYD_ASSUME_KERNEL";
 /// The environment variable to read the default shell from.
 pub const ENV_SH: &str = "SYD_SHELL";
 /// The environment variable to read the log level from.
@@ -132,6 +135,14 @@ pub const ENV_SYNC_SCMP: &str = "SYD_SYNC_SCMP";
 /// Exherbo during Paludis' generate metadata phase.
 pub const ENV_QUICK_BOOT: &str = "SYD_QUICK_BOOT";
 
+/// The environment variable to make syd-oci skip configuration parsing.
+///
+/// This is primarily intended for testing.
+pub const ENV_OCI_NO_CONFIG: &str = "SYD_OCI_NO_CONFIG";
+
+/// The environment variable to read the parent-death signal from.
+pub const ENV_PDS: &str = "SYD_PDS";
+
 /// SafeSetID does not allow UID transitions for UIDs less than or equal to UID_MIN.
 /// 11 is typically the `operator` user.
 pub const UID_MIN: u64 = 11;
@@ -154,11 +165,17 @@ pub static VERSION: Lazy<&'static str> = Lazy::new(|| {
     }
 });
 
-/// Api version of the configuration.
-pub const API_VERSION: &str = env!("CARGO_PKG_VERSION_MAJOR");
+/// Api version of the syd(2) configuration.
+pub const API_VERSION: Version = Version::new(3, 1);
+
+/// Api major version of the syd(2) configuration.
+pub const API_MAJOR_VERSION: &str = env!("CARGO_PKG_VERSION_MAJOR");
+
+/// Api minor version of the syd(2) configuration.
+pub const API_MINOR_VERSION: &str = "1";
 
 /// Api version for the IPC service.
-pub const IPC_VERSION: &str = "1";
+pub const IPC_MINOR_VERSION: &str = "1";
 
 /// File format marker for Crypt sandboxing.
 pub const CRYPT_MAGIC: &[u8] = &[0x7F, b'S', b'Y', b'D', 3];
@@ -258,8 +275,8 @@ pub const MON_CYCLE_TIME: Duration = Duration::from_millis(25);
 pub const INT_CYCLE_TIME: Duration = Duration::from_millis(10);
 
 /// Specify the cycle period of the aes thread.
-/// Defaults to 250 milliseconds.
-pub const AES_CYCLE_TIME: Duration = Duration::from_millis(250);
+/// Defaults to 10 milliseconds.
+pub const AES_CYCLE_TIME: Duration = Duration::from_millis(10);
 
 /// Stack size for the syscall handler threads.
 /// Defaults to 208k.
@@ -350,6 +367,17 @@ pub const SEGVGUARD_MAXCRASHES: u8 = 5;
 
 // Sandbox Restrictions
 
+/// List of allowlisted personality requests.
+///
+/// This is consistent with podman and docker.
+pub const ALLOWLIST_PERSONALITY: &[KeyValue] = &[
+    ("PER_LINUX", 0x0000),
+    ("PER_LINUX32", 0x0800000),
+    ("UNAME26", 0x0020000),
+    ("PER_LINUX32|UNAME26", 0x0800000 | 0x0020000),
+    ("GET_PERSONALITY", 0xffffffff),
+];
+
 /// List of allowlisted prctl requests.
 pub const ALLOWLIST_PRCTL: &[KeyValue] = &[
     ("PR_SET_PDEATHSIG", 1),
@@ -508,31 +536,32 @@ pub const UNSAFE_ENV: &[&[u8]] = &[
  * 12. FS_IOC32_GETFLAGS, FS_IOC32_SETFLAGS are forwarded to their device
  *     implementations.
  */
+#[allow(clippy::unnecessary_cast)]
 pub const DEFAULT_IOCTL: &[(u64, bool)] = &[
-    (0x5451, false),     // FIOCLEX
-    (0x5450, false),     // FIONCLEX
-    (0x5421, false),     // FIONBIO
-    (0x5452, false),     // FIOASYNC
-    (0x5460, false),     // FIOQSIZE
-    (0xc0045877, true),  // FIFREEZE: CAP_SYS_ADMIN!
-    (0xc0045878, true),  // FITHAW:   ditto!
-    (0xc020660b, true),  // FS_IOC_FIEMAP: Leaks on-disk layout.
-    (0x2, true),         // FIGETBSZ: ditto!
-    (0x40049409, false), // FICLONE
-    (0x4020940d, false), // FICLONERANGE
-    (0xc0189436, false), // FIDEDUPERANGE
-    (0x80111500, true),  // FS_IOC_GETFSUUID: Leaks on-disk layout.
-    (0x80811501, true),  // FS_IOC_GETFSSYSFSPATH: ditto!
-    (0x1, true),         // FIBMAP: ditto!
-    (0x4b4d, true),      // KDSETKEYCODE
-    (0x4b4e, true),      // KDSIGACCEPT
-    (0x5423, true),      // TIOCSETD
-    (0x5412, true),      // TIOCSTI
-    (0x541c, true),      // TIOCCONS
-    (0x541d, true),      // TIOCLINUX
-    (0x801c581f, true),  // FS_IOC_FSGETXATTR, (SAFETY: xattr restrictions)
-    (0x401c5820, true),  // FS_IOC_FSSETXATTR, (ditto)
-    (0x40086602, true),  // FS_IOC_SETFLAGS (to deny immutable and append-only flag changes)
+    (libc::FIOCLEX as u64, false),
+    (libc::FIONCLEX as u64, false),
+    (libc::FIONBIO as u64, false),
+    (libc::FIOASYNC as u64, false),
+    (crate::fs::FIOQSIZE as u64, false),
+    (crate::fs::FIFREEZE as u64, true),      // CAP_SYS_ADMIN!
+    (crate::fs::FITHAW as u64, true),        // ditto!
+    (crate::fs::FS_IOC_FIEMAP as u64, true), // Leaks on-disk layout.
+    (crate::fs::FIGETBSZ as u64, true),      // ditto!
+    (libc::FICLONE as u64, false),
+    (libc::FICLONERANGE as u64, false),
+    (crate::fs::FIDEDUPERANGE as u64, false),
+    (crate::fs::FS_IOC_GETFSUUID as u64, true), // Leaks on-disk layout.
+    (crate::fs::FS_IOC_GETFSSYSFSPATH as u64, true), // ditto!
+    (crate::fs::FIBMAP as u64, true),           // ditto!
+    (crate::fs::KDSETKEYCODE as u64, true),
+    (crate::fs::KDSIGACCEPT as u64, true),
+    (libc::TIOCSETD as u64, true),
+    (libc::TIOCSTI as u64, true),
+    (libc::TIOCCONS as u64, true),
+    (libc::TIOCLINUX as u64, true),
+    (crate::fs::FS_IOC_FSGETXATTR as u64, true), // (SAFETY: xattr restrictions)
+    (crate::fs::FS_IOC_FSSETXATTR as u64, true), // ditto!
+    (crate::fs::FS_IOC_SETFLAGS as u64, true),   // To deny immutable & append-only flag changes
 ];
 
 // Sandboxing profiles
@@ -616,6 +645,10 @@ pub const PROFILE_NOMEM: &[&str] = &["trace/allow_unsafe_memory:true"];
 /// No PIE sandboxing profile.
 pub const PROFILE_NOPIE: &[&str] = &["trace/allow_unsafe_nopie:true"];
 
+/// Random sandboxing profile, activated with:
+/// `syd --profile rand`.
+pub const PROFILE_RAND: &[&str] = &["trace/force_rand_fd:true"];
+
 /// Firefox sandboxing profile.
 /// `syd --profile ff`
 pub const PROFILE_FIREFOX: &[&str] = &[
@@ -638,7 +671,7 @@ pub const PROFILE_CONTAINER: &[&str] = &["unshare/all:true"];
 /// `syd --profile immutable`
 pub const PROFILE_IMMUTABLE: &[&str] = &[
     "include_profile container",
-    "bind+/var/empty:/boot:ro,nodev,noexec,nosuid",
+    "bind+tmpfs:/boot:nodev,nosuid,noexec,nosymfollow,ro,mode=0",
     "bind+tmpfs:/dev/shm:nodev,nosuid,noexec,nosymfollow,mode=1777",
     "bind+tmpfs:/run:nodev,nosuid,nosymfollow,mode=1777",
     "bind+tmpfs:/tmp:nodev,nosuid,nosymfollow,mode=1777",
@@ -696,7 +729,11 @@ pub const PROFILE_PALUDIS: &[&str] = &[
     "trace/allow_unsafe_exec:true",
     "trace/allow_unsafe_filename:true", // sydbox#118
     "trace/allow_unsafe_libc:true",
+    "trace/allow_unsafe_create:true",     // poppler-data src_install fails without this.
+    "trace/allow_unsafe_hardlinks:true",  // gnulib linkat test fails without this.
     "trace/allow_unsafe_magiclinks:true", // tests love to access /proc/1.
+    "trace/allow_unsafe_symlinks:true",   // git's tests fail without this.
+    "trace/allow_unsafe_memfd:true",      // executable memory-fds are OK.
     "trace/allow_unsafe_memory:true",
     "trace/allow_unsafe_msgsnd:true",
     "trace/allow_unsafe_nice:true",
@@ -709,6 +746,7 @@ pub const PROFILE_PALUDIS: &[&str] = &[
     "trace/allow_unsafe_ptrace:true",
     "trace/allow_unsafe_prctl:true", // perl tests want to set process name.
     "trace/allow_unsafe_prlimit:true",
+    "trace/allow_unsafe_shm:true",   // perl tests map executable shared memory.
     "trace/allow_unsafe_sigreturn:true",
     "trace/allow_unsafe_spec_exec:true",
     "trace/allow_unsafe_socket:true",
@@ -716,6 +754,7 @@ pub const PROFILE_PALUDIS: &[&str] = &[
     "trace/allow_unsafe_sysinfo:true", // disables sysinfo(2) randomizer, >3.23.4
     "trace/allow_unsupp_socket:true",
     "trace/allow_unsafe_time:true",
+    "trace/allow_unsafe_xattr:true", // setcap requires access to security.capability.*
     // Avoid annoying open(/, O_DIRECTORY) failures.
     // See: #208, #212, and #219.
     // no-op unless Landlock is enabled.
@@ -769,6 +808,11 @@ pub const PROFILE_PALUDIS: &[&str] = &[
     // in the default secure `linux` profile.
     "allow/stat+/proc/[0-9]*",
     "allow/read,stat+/proc/filesystems",
+    // SAFETY: proc_pid_maps(5) access allows an attacker
+    // to easily circumvent ASLR, therefore the two rules
+    // below have been moved from Linux to Paludis profile.
+    "allow/read,stat+/proc/self/maps",
+    "allow/read,stat+/proc/self/task/[0-9]*/maps",
     "allow/read,stat+/proc/self/mounts",
     "allow/read,stat+/proc/self/task/[0-9]*/mounts",
     "allow/read,stat+/proc/mounts", // symlink to self/mounts
@@ -788,8 +832,9 @@ pub const PROFILE_PALUDIS: &[&str] = &[
     "allow/lock/bind+0",
     "allow/lock/bind+1024-65535",
     "allow/lock/connect+1024-65535",
-    // Allow creation of memory file descriptors.
-    "allow/create+/memfd:**",
+    // Allow interaction with memory file descriptors.
+    "allow/stat,exec,read,write,create,chmod,truncate+/memfd:**",
+    "allow/stat,exec,read,write,create,chmod,truncate+/memfd-hugetlb:**",
     // Repository directory
     "allow/read,stat,chdir,readdir+/var/db/paludis/***",
 ];
@@ -862,6 +907,9 @@ pub const PROFILE_LINUX: &[&str] = &[
     "allow/exec,read,stat,chdir,readdir+/lib*/***",
     "allow/exec,read,stat,chdir,readdir+/usr/***",
     "allow/exec,read,stat,chdir,readdir+/opt/***",
+    // SAFETY: Do not leak kernel config/module information.
+    // This overrides the lib rule above.
+    "deny/exec,read,stat,chdir,readdir+/lib*/modules/***",
     // Safe defaults for Network sandboxing
     "allow/net/link+route", // allow NETLINK_ROUTE.
     "allow/net/connect+/run/nscd/socket",
@@ -984,8 +1032,6 @@ pub const PROFILE_LINUX: &[&str] = &[
     r"allow/read,stat,write,truncate,ioctl+/proc/self/task/[0-9]*/fd/socket:\[[0-9]*\]",
     "allow/read,stat,chdir,readdir+/proc/self/fdinfo/[0-9]*",
     "allow/read,stat,chdir,readdir+/proc/self/task/[0-9]*/fdinfo/[0-9]*",
-    "allow/read,stat+/proc/self/maps",
-    "allow/read,stat+/proc/self/task/[0-9]*/maps",
     "allow/stat,chdir,readdir+/proc/self/root",
     "allow/stat,chdir,readdir+/proc/self/task/[0-9]*/root",
     // /run
@@ -1069,8 +1115,11 @@ pub const PROFILE_TTY: &[&str] = &[
     "allow/read,stat,write,ioctl+/dev/tty",
     "allow/read,stat,write,ioctl+${SYD_TTY}",
     // PTY handling
-    "ioctl/allow+0x5410",     // TIOCSPGRP
+    "ioctl/allow+0x540f",     // TIOCGPGRP: tcgetpgrp(3)
+    "ioctl/allow+0x5410",     // TIOCSPGRP: tcsetpgrp(3)
+    "ioctl/allow+0x5429",     // TIOCGSID: tcgetsid(3)
     "ioctl/allow+0x5401",     // TCGETS: isatty(3), tcgetattr(3) etc.
+    "ioctl/allow+0x802c542a", // TCGETS2: ditto.
     "ioctl/allow+0x5432",     // TCGETX
     "ioctl/allow+0x5433",     // TCSETX
     "ioctl/allow+0x5434",     // TCSETXF
@@ -1226,12 +1275,18 @@ pub const EBPF_SYSCALLS: &[&str] = &["bpf", "uretprobe"];
 
 /// The list of mount family syscalls
 pub const MOUNT_SYSCALLS: &[&str] = &[
+    "fsconfig",
+    "fsmount",
+    "fsopen",
+    "fspick",
+    "listmount",
     "mount",
     "mount_setattr",
+    "open_tree",
+    "open_tree_attr",
+    "statmount",
     "umount",
     "umount2",
-    "listmount",
-    "statmount",
 ];
 
 /// The list of process, i/o priority changing system calls.
@@ -1454,60 +1509,6 @@ pub const HOOK_SYSCALLS: &[&str] = &[
 /// As of 3.30.0 we use SafeChroot so chroot(2) is not here anymore.
 pub const NOOP_SYSCALLS: &[&str] = &["pivot_root"];
 
-/// The list of system calls that are denied in syd parent seccomp filter.
-/// This filter is inherited by the sandbox process.
-/// These system calls must not be used by syd.
-pub const DEAD_SYSCALLS: &[&str] = &[
-    "_sysctl", // deprecated
-    "acct",
-    // Hard disabled due to pointer-indirection during NS check.
-    "clone3",
-    "create_module",
-    "delete_module",
-    "finit_module",
-    "fsconfig",
-    "fsmount",
-    "fsopen",
-    "fspick",
-    "get_kernel_syms",
-    "init_module",
-    "ioperm",
-    "iopl",
-    "kexec_file_load",
-    "kexec_load",
-    "lsm_get_self_attr",
-    "lsm_set_self_attr",
-    "lsm_list_modules",
-    "mbind",
-    "migrate_pages",
-    "move_pages",
-    // As of 3.35.2, mincore may be enabled using
-    // trace/allow_unsafe_page_cache:1 at startup.
-    //"mincore",
-    "name_to_handle_at",
-    "nfsservctl",
-    "open_tree",
-    "pciconfig_iobase",
-    "pciconfig_read",
-    "pciconfig_write",
-    "query_module",
-    "quotactl",
-    "quotactl_fd",
-    "reboot",
-    "request_key",
-    "stime", // deprecated use clock_settime
-    "swapoff",
-    "swapon",
-    "sysfs",
-    "uselib",
-    // Added in Linux-4.3
-    // (Ab)used by most Project Zero Linux kernel exploits.
-    "userfaultfd",
-    "ustat", // deprecated
-    "vhangup",
-    "vmsplice", // https://lore.kernel.org/linux-mm/X+PoXCizo392PBX7@redhat.com/
-];
-
 /// The list of system calls which are confined by the Stat sandboxing category.
 ///
 /// Note, this list _must_ be sorted!
@@ -1617,7 +1618,7 @@ pub const SAFE_SYSCALLS: &[&str] = &[
     //"_sysctl",
     //"accept",
     //"accept4",
-    //"acct",
+    //"acct", // CAP_SYS_PACCT
     //add_key, restrictions applied, see setup_seccomp.
     //adjtimex, restrictions applied, see setup_seccomp.
     //afs_syscall
@@ -1751,8 +1752,8 @@ pub const SAFE_SYSCALLS: &[&str] = &[
     "io_uring_setup",
     */
     //"ioctl", restrictions applied, see setup_seccomp.
-    //"ioperm",
-    //"iopl",
+    //"ioperm", // CAP_SYS_RAWIO
+    //"iopl", // CAP_SYS_RAWIO
     "ioprio_get",
     //"ioprio_set", restrictions applied, see setup_seccomp.
     //"kcmp", part of PTRACE_SYSCALLS now as of 3.33.0.
@@ -1786,7 +1787,7 @@ pub const SAFE_SYSCALLS: &[&str] = &[
     "mprotect",
     "mq_getsetattr",
     "mq_notify",
-    "mq_open",
+    //"mq_open", restrictions applied, see setup_seccomp.
     "mq_timedreceive",
     "mq_timedreceive_time64",
     "mq_timedsend",
@@ -1794,8 +1795,8 @@ pub const SAFE_SYSCALLS: &[&str] = &[
     "mq_unlink",
     //"mremap",
     "mseal",
-    "msgctl",
-    "msgget",
+    //"msgctl", restrictions applied, see setup_seccomp.
+    //"msgget", restrictions applied, see setup_seccomp.
     "msgrcv",
     //"msgsnd", restrictions applied, see setup_seccomp.
     "msync",
@@ -1809,7 +1810,7 @@ pub const SAFE_SYSCALLS: &[&str] = &[
     //open_tree
     "pause",
     //"perf_event_open", restrictions applied see load_seccomp_parent and setup_seccomp.
-    "personality", // restrictions applied in load_seccomp_parent.
+    //"personality", restrictions applied, see setup_seccomp.
     //"pidfd_getfd", added to ptrace set as of 3.35.2.
     //"pidfd_open", restrictions applied, see setup_seccomp.
     "pidfd_send_signal",
@@ -1852,12 +1853,12 @@ pub const SAFE_SYSCALLS: &[&str] = &[
     "recvmmsg",
     "recvmmsg_time64",
     "recvmsg",
-    "remap_file_pages",
+    "remap_file_pages", // prot _must_ be 0, not security critical.
     //request_key, restrictions applied, see setup_seccomp.
     "restart_syscall",
     "riscv_flush_icache",
     //"riscv_hwprobe", info leak, see: https://www.kernel.org/doc/html/v6.6/riscv/hwprobe.html
-    "rseq",
+    //"rseq", restrictions applied, see setup_seccomp.
     //"rt_sigaction", SA_RESTART tracking.
     "rt_sigpending",
     "rt_sigprocmask",
@@ -1886,8 +1887,8 @@ pub const SAFE_SYSCALLS: &[&str] = &[
     //"seccomp", restrictions applied, see setup_seccomp.
     //security
     "select",
-    "semctl",
-    "semget",
+    //"semctl", restrictions applied, see setup_seccomp.
+    //"semget", restrictions applied, see setup_seccomp.
     "semop",
     "semtimedop",
     "semtimedop_time64",
@@ -1932,10 +1933,10 @@ pub const SAFE_SYSCALLS: &[&str] = &[
     "set_tls", // arm
     //"sgetmask", // x86, OBSOLETE!
     //"ssetmask", // x86, OBSOLETE!
-    "shmat",
-    "shmctl",
+    //"shmat", restrictions applied, see setup_seccomp.
+    //"shmctl", restrictions applied, see setup_seccomp.
     "shmdt",
-    "shmget",
+    //"shmget", restrictions applied, see setup_seccomp.
     "shutdown",
     "signal",
     //"sigaction", SA_RESTART tracking.
@@ -1955,6 +1956,7 @@ pub const SAFE_SYSCALLS: &[&str] = &[
     //swapon
     //"sync",
     "sync_file_range",
+    "sync_file_range2", // arm & ppc
     //"syncfs",
     //"sysinfo", information-leak, see setup_seccomp.
     //"syslog",
@@ -1986,11 +1988,13 @@ pub const SAFE_SYSCALLS: &[&str] = &[
     "unshare", // restrictions applied in load_seccomp_parent.
     //"uretprobe", restrictions applied, see setup_seccomp.
     //"uselib",
+    // Added in Linux-4.3
+    // (Ab)used by most Project Zero Linux kernel exploits.
     //"userfaultfd",
-    //"ustat",
+    //"ustat", deprecated
     "vfork",
-    //"vhangup",
-    //"vmsplice", // see DEAD_SYSCALLS!
+    //"vhangup", // CAP_SYS_TTY_CONFIG
+    //"vmsplice", // https://lore.kernel.org/linux-mm/X+PoXCizo392PBX7@redhat.com/
     //vserver
     "wait4",
     "waitid",
@@ -2003,7 +2007,7 @@ pub const SAFE_SYSCALLS: &[&str] = &[
 pub const EMU_SYSCALLS: &[&str] = &[
     "_llseek",
     "accept4",
-    "bind",
+    "bind", // TODO: protect with syscall cookies!
     "brk",
     "clock_gettime",
     "clock_gettime64",
@@ -2011,15 +2015,14 @@ pub const EMU_SYSCALLS: &[&str] = &[
     "clone",
     "clone3",
     "close",
-    "connect",
+    "connect", // TODO: protect with syscall cookies!
     "exit",
     "exit_group",
-    "fallocate",
+    "fallocate", // TODO: protect with syscall cookies!
     "fanotify_mark",
     "fchdir",
     "fchmod",
     "fchown",
-    "fdatasync", // for logging.
     "fgetxattr",
     "flistxattr",
     "fremovexattr",
@@ -2042,11 +2045,11 @@ pub const EMU_SYSCALLS: &[&str] = &[
     "gettid",
     "getxattrat",
     "inotify_add_watch",
-    //"kcmp", only KCMP_VM and KCMP_FILE are allowed.
+    "kcmp",
     "landlock_create_ruleset",
     "landlock_restrict_self",
     "lgetxattr",
-    "linkat",
+    //linkat, protected with syscall argument cookies.
     "listxattrat",
     "llistxattr",
     "lremovexattr",
@@ -2077,6 +2080,8 @@ pub const EMU_SYSCALLS: &[&str] = &[
     "restart_syscall",
     "rseq",
     "rt_sigprocmask",
+    "rt_sigtimedwait",        // Used by interrupt handling.
+    "rt_sigtimedwait_time64", // ditto.
     "sched_getaffinity",
     "sched_setaffinity",
     "sched_yield",
@@ -2089,7 +2094,7 @@ pub const EMU_SYSCALLS: &[&str] = &[
     "sigprocmask",
     //"socket", protected with syscall argument cookies.
     "splice",
-    "symlinkat",
+    "symlinkat", // TODO: protect with syscall cookies!
     "tee",
     "tgkill",
     "tkill",
@@ -2101,16 +2106,15 @@ pub const EMU_SYSCALLS: &[&str] = &[
     // Required to unblock FIFOs.
     "sigreturn",
     "rt_sigreturn",
-    //TODO:deny installing new signal handlers!
-    "sigaction",
-    "rt_sigaction",
+    //"sigaction", installing new signal handlers is not permitted.
+    //"rt_sigaction", ditto.
     // fd calls
     "faccessat2",
-    "fchmodat",
-    "fchmodat2",
-    "fchownat",
-    "mkdirat",
-    "mknodat",
+    "fchmodat",  // TODO: protect with syscall cookies!
+    "fchmodat2", // TODO: ditto.
+    "fchownat",  // TODO: ditto.
+    "mkdirat",   // TODO: ditto.
+    "mknodat",   // TODO: ditto.
     "newfstatat",
     //"openat2", protected with syscall argument cookies.
     "readlinkat",
@@ -2135,53 +2139,74 @@ pub const INT_SYSCALLS: &[&str] = &[
     "clock_nanosleep",
     "close",
     "exit",
-    "exit_group", // to exit if inter-thread signaling does not work.
+    "exit_group",      // to exit if inter-thread signaling does not work.
+    "clock_gettime",   // required for logging.
+    "clock_gettime64", // ditto.
     // "futex", See: FUTEX_SYSCALLS for futex family.
     "getpid",
     "gettid",
-    "munmap", // memory deallocation is allowed but allocation is not.
+    // can {{dr}e,}allocate memory.
+    // mmap{,2} and mprotect are further confined to disable PROT_EXEC.
+    "brk",
+    "madvise",
+    "mremap",
+    "munmap",
     "nanosleep",
     "lseek",   // seek in /proc/$pid/status file.
     "_llseek", // ditto.
     "read",
     "restart_syscall",
     "rt_sigprocmask",
+    "rseq",
     "sched_yield",
     "sigaltstack", // rare but necessary.
     "sigprocmask", // ditto.
 ];
 
 /// Fcntl operations allowed for Interrupter thread.
-pub const INT_FCNTL_OPS: &[u64] = &[libc::F_GETFD as u64];
+pub const INT_FCNTL_OPS: &[u64] = &[
+    libc::F_GETFD as u64,
+    libc::F_SETFD as u64,
+    libc::F_OFD_SETLK as u64,
+    libc::F_OFD_SETLKW as u64,
+];
 
 /// System calls allowed for the IPC thread.
 pub const IPC_SYSCALLS: &[&str] = &[
     "close",
     "exit",
+    "clock_gettime",   // required for logging.
+    "clock_gettime64", // ditto.
     // "futex", See: FUTEX_SYSCALLS for futex family.
+    "getpid",
+    "gettid",
     "getrandom",
+    "getsockopt",
     "accept4",
-    "recvfrom",
-    "write",
-    // In case sandbox is locked through IPC,
-    // we need mmap family.
+    "recvmsg",
+    "sendmsg",
+    // can {{dr}e,}allocate and seal memory.
+    // mmap{,2} and mprotect are further confined to disable PROT_EXEC.
+    "brk",
     "madvise",
-    "mmap",
-    "mmap2",
-    "mprotect",
     "mremap",
     "mseal",
     "munmap",
-    //
     "restart_syscall",
     "rt_sigprocmask",
+    "rseq",
     "sched_yield",
     "sigaltstack", // rare but necessary.
     "sigprocmask", // ditto.
 ];
 
 /// Fcntl operations allowed for IPC thread.
-pub const IPC_FCNTL_OPS: &[u64] = &[libc::F_GETFD as u64];
+pub const IPC_FCNTL_OPS: &[u64] = &[
+    libc::F_GETFD as u64,
+    libc::F_SETFD as u64,
+    libc::F_OFD_SETLK as u64,
+    libc::F_OFD_SETLKW as u64,
+];
 
 /// System calls allowed for AES threads.
 pub const AES_SYSCALLS: &[&str] = &[
@@ -2235,9 +2260,8 @@ pub const AES_SYSCALLS: &[&str] = &[
     "sigprocmask",
     "splice",
     "tee",
-    //TODO:deny installing new signal handlers!
-    "sigaction",
-    "rt_sigaction",
+    //"sigaction", installing new signal handlers is not permitted.
+    //"rt_sigaction", ditto.
 ];
 
 /// Fcntl operations allowed for AES threads.
@@ -2257,7 +2281,6 @@ pub const MAIN_SYSCALLS: &[&str] = &[
     "clock_nanosleep",
     "close",
     "exit_group",
-    "fdatasync", // for logging.
     "fstat",
     "fstat64",
     "fstatfs",
@@ -2303,9 +2326,8 @@ pub const MAIN_SYSCALLS: &[&str] = &[
     "tgkill",
     "tkill",
     "waitid",
-    //TODO:deny installing new signal handlers!
-    "sigaction",
-    "rt_sigaction",
+    //"sigaction", installing new signal handlers is not permitted.
+    //"rt_sigaction", ditto.
     // fd-calls
     "faccessat2",
     "newfstatat",
@@ -2323,6 +2345,20 @@ pub const MAIN_FCNTL_OPS: &[u64] = &[
     libc::F_OFD_SETLKW as u64,
 ];
 
+/// System calls that must be specifically allowed for syd-oci.
+///
+/// Note, this list _must_ be sorted!
+pub const OCI_SYSCALLS: &[&str] = &[
+    "kcmp",
+    "pidfd_getfd",
+    "process_mrelease",
+    "process_vm_readv",
+    "process_vm_writev",
+    "ptrace",
+    "syslog",
+    "unshare",
+];
+
 /// Define a static (global) Lazy value for the kernel version.
 #[allow(clippy::disallowed_methods)]
 pub static KERNEL_VERSION: Lazy<(u32, u32)> = Lazy::new(|| {
@@ -2332,9 +2368,14 @@ pub static KERNEL_VERSION: Lazy<(u32, u32)> = Lazy::new(|| {
     use memchr::memchr;
     use nix::sys::utsname::uname;
 
-    let version = uname().unwrap();
-    let version = version.release().as_bytes();
-    let nextdot = memchr(b'.', version).unwrap();
+    // As of version 3.36.1 we allow overriding kernel version with SYD_ASSUME_KERNEL.
+    let version = if let Some(version) = std::env::var_os(ENV_ASSUME_KERNEL) {
+        version.as_bytes().to_owned()
+    } else {
+        let version = uname().unwrap();
+        version.release().as_bytes().to_owned()
+    };
+    let nextdot = memchr(b'.', &version).unwrap();
     let major = btoi::<u32>(&version[..nextdot]).unwrap();
     let version = &version[nextdot + 1..];
     let nextdot = memchr(b'.', version).unwrap();
@@ -2343,6 +2384,12 @@ pub static KERNEL_VERSION: Lazy<(u32, u32)> = Lazy::new(|| {
     (major, minor)
 });
 
+/// Check for MFD_NOEXEC_SEAL support (Linux 6.3 or newer).
+pub static HAVE_MFD_NOEXEC_SEAL: Lazy<bool> = Lazy::new(|| {
+    let (major, minor) = *KERNEL_VERSION;
+    major > 6 || (major == 6 && minor >= 3)
+});
+
 /// Check for PIDFD_THREAD support (Linux-6.9 or newer).
 pub static HAVE_PIDFD_THREAD: Lazy<bool> = Lazy::new(|| {
     let (major, minor) = *KERNEL_VERSION;
@@ -2367,11 +2414,50 @@ pub static HAVE_AT_EXECVE_CHECK: Lazy<bool> = Lazy::new(|| {
     major > 6 || (major == 6 && minor >= 14)
 });
 
+/// Check for PROCMAP_QUERY ioctl(2) (Linux-6.11 or newer).
+pub static HAVE_PROCMAP_QUERY: Lazy<bool> = Lazy::new(|| {
+    let (major, minor) = *KERNEL_VERSION;
+    major > 6 || (major == 6 && minor >= 11)
+});
+
+/// Check for MADV_GUARD_{INSTALL,REMOVE} madvise(2) (Linux-6.13 or newer).
+pub static HAVE_MADV_GUARD: Lazy<bool> = Lazy::new(|| {
+    let (major, minor) = *KERNEL_VERSION;
+    major > 6 || (major == 6 && minor >= 13)
+});
+
+/// Check for CONFIG_CROSS_MEMORY_ATTACH support.
+pub static HAVE_CROSS_MEMORY_ATTACH: Lazy<bool> = Lazy::new(check_cross_memory_attach);
+
+/// Check for stat.st_size support in /proc/$pid/fd (Linux-6.2 or newer).
+pub static HAVE_PROC_PID_FD_STAT_SIZE: Lazy<bool> = Lazy::new(|| {
+    let (major, minor) = *KERNEL_VERSION;
+    major > 6 || (major == 6 && minor >= 2)
+});
+
 //
 // Below is internal territory, you have been warned.
 
+/// Unsafe mode mask for SHM hardening.
+pub(crate) const SHM_UNSAFE_MASK: u64 = 0o177;
+
+/// Name prefix for memory file descriptors.
+pub const MFD_NAME_PREFIX: &[u8] = b"/memfd:";
+
+/// Name prefix for hugetlbfs memory file descriptors.
+pub const MFD_NAME_PREFIX_HUGETLB: &[u8] = b"/memfd-hugetlb:";
+
+/// BIGTCP maximum for loopback:
+///
+/// https://git.kernel.org/pub/scm/linux/kernel/git/torvalds/linux.git/commit/?id=d6f938ce52f9adb23f4c31cc371654a5f18ff328
+pub const LOOPBACK_BIGTCP_MAX: u32 = 0x30d40;
+
 // Determine minimum allowed mmap address by reading `/proc/sys/vm/mmap_min_addr`.
-pub(crate) static MMAP_MIN_ADDR: Lazy<u64> = Lazy::new(|| proc_mmap_min_addr().unwrap_or(4096));
+//
+// Enforce OpenBSD's hard-coded minimum mmap address by capping at PAGE_SIZE;
+// See: https://github.com/openbsd/src/blob/94a7e27d04f23871848a8f225ef286e84969197a/sys/uvm/uvm_map.c#L235
+pub(crate) static MMAP_MIN_ADDR: Lazy<u64> =
+    Lazy::new(|| proc_mmap_min_addr().unwrap_or(*PAGE_SIZE).max(*PAGE_SIZE));
 
 // Trace data for ptrace(2) hooked system calls.
 //
@@ -2420,6 +2506,16 @@ pub(crate) fn ROOT_FD() -> RawFd {
     *_ROOT_FD.get().unwrap()
 }
 
+/// Returns a reference to the static `/` mode.
+///
+/// Calling this before calling `proc_init` will panic!
+#[allow(clippy::disallowed_methods)]
+#[allow(non_snake_case)]
+#[inline(always)]
+pub(crate) fn ROOT_F_MODE() -> u16 {
+    *_ROOT_F_MODE.get().unwrap()
+}
+
 /// Returns a reference to the static `/` unique mount id.
 ///
 /// Calling this before calling `proc_init` will panic!
@@ -2451,6 +2547,16 @@ pub(crate) fn PROC_FD() -> RawFd {
     *_PROC_FD.get().unwrap()
 }
 
+/// Returns a reference to the static `/proc` mode.
+///
+/// Calling this before calling `proc_init` will panic!
+#[allow(clippy::disallowed_methods)]
+#[allow(non_snake_case)]
+#[inline(always)]
+pub(crate) fn PROC_F_MODE() -> u16 {
+    *_PROC_F_MODE.get().unwrap()
+}
+
 /// Returns a reference to the static `/proc` unique mount id.
 ///
 /// Calling this before calling `proc_init` will panic!
@@ -2482,6 +2588,16 @@ pub(crate) fn NULL_FD() -> RawFd {
     *_NULL_FD.get().unwrap()
 }
 
+/// Returns a reference to the static `/dev/null` mode.
+///
+/// Calling this before calling `proc_init` will panic!
+#[allow(clippy::disallowed_methods)]
+#[allow(non_snake_case)]
+#[inline(always)]
+pub(crate) fn NULL_F_MODE() -> u16 {
+    *_NULL_F_MODE.get().unwrap()
+}
+
 /// Returns a reference to the static `/dev/null` unique mount id.
 ///
 /// Calling this before calling `proc_init` will panic!
@@ -2508,26 +2624,37 @@ pub(crate) fn NULL_FILE() -> BorrowedFd<'static> {
 /// File descriptor to `/`, ie the root file system.
 pub(crate) static _ROOT_FD: OnceLock<RawFd> = OnceLock::new();
 
+/// Mode to `/`, ie the root file system.
+pub(crate) static _ROOT_F_MODE: OnceLock<u16> = OnceLock::new();
+
 /// Unique mount id to `/`, ie the root file system.
 pub(crate) static _ROOT_MNT_ID: OnceLock<u64> = OnceLock::new();
 
 /// File descriptor to /proc file system.
 pub(crate) static _PROC_FD: OnceLock<RawFd> = OnceLock::new();
 
+/// Mode to the /proc file system.
+pub(crate) static _PROC_F_MODE: OnceLock<u16> = OnceLock::new();
+
 /// Unique mount id to /proc file system.
 pub(crate) static _PROC_MNT_ID: OnceLock<u64> = OnceLock::new();
 
 /// File descriptor to /dev/null character device.
 pub(crate) static _NULL_FD: OnceLock<RawFd> = OnceLock::new();
 
+/// Mode to the /dev/null character device.
+pub(crate) static _NULL_F_MODE: OnceLock<u16> = OnceLock::new();
+
 /// Unique mount id to /dev/null character device.
 pub(crate) static _NULL_MNT_ID: OnceLock<u64> = OnceLock::new();
 
 /// Initialize static file descriptors for use by syd::proc any friends.
+#[allow(clippy::cast_possible_truncation)]
 #[allow(clippy::cognitive_complexity)]
 #[allow(clippy::disallowed_methods)]
 pub fn proc_init() -> Result<(), Errno> {
-    let mask = if *HAVE_STATX_MNT_ID_UNIQUE {
+    let mut mask = STATX_MODE;
+    mask |= if *HAVE_STATX_MNT_ID_UNIQUE {
         STATX_MNT_ID_UNIQUE
     } else {
         STATX_MNT_ID
@@ -2538,7 +2665,8 @@ pub fn proc_init() -> Result<(), Errno> {
         OFlag::O_PATH | OFlag::O_DIRECTORY | OFlag::O_CLOEXEC,
         Mode::empty(),
     )?;
-    let id_root = fstatx(&fd_root, mask).map(|stx| stx.stx_mnt_id)?;
+    let (f_mode_root, mnt_id_root) = fstatx(&fd_root, mask)
+        .map(|stx| (stx.stx_mode & !(libc::S_IFMT as u16), stx.stx_mnt_id))?;
 
     // SAFETY: To make this file descriptor harder to spot by an
     // attacker we duplicate it to a random fd number.
@@ -2547,9 +2675,11 @@ pub fn proc_init() -> Result<(), Errno> {
     info!("ctx": "run", "op": "opendir_root",
         "msg": "opened root directory",
         "fd": fd_root,
-        "id": id_root);
+        "f_mode": f_mode_root,
+        "mnt_id": mnt_id_root);
     _ROOT_FD.set(fd_root).or(Err(Errno::EAGAIN))?;
-    _ROOT_MNT_ID.set(id_root).or(Err(Errno::EAGAIN))?;
+    _ROOT_F_MODE.set(f_mode_root).or(Err(Errno::EAGAIN))?;
+    _ROOT_MNT_ID.set(mnt_id_root).or(Err(Errno::EAGAIN))?;
 
     // SAFETY: fd_root is a valid fd for process lifetime.
     let fd_root = unsafe { BorrowedFd::borrow_raw(fd_root) };
@@ -2562,7 +2692,8 @@ pub fn proc_init() -> Result<(), Errno> {
         OFlag::O_DIRECTORY | OFlag::O_CLOEXEC,
         Mode::empty(),
     )?;
-    let id_proc = fstatx(&fd_proc, mask).map(|stx| stx.stx_mnt_id)?;
+    let (f_mode_proc, mnt_id_proc) = fstatx(&fd_proc, mask)
+        .map(|stx| (stx.stx_mode & !(libc::S_IFMT as u16), stx.stx_mnt_id))?;
 
     // SAFETY: To make this file descriptor harder to spot by an
     // attacker we duplicate it to a random fd number.
@@ -2571,9 +2702,11 @@ pub fn proc_init() -> Result<(), Errno> {
     info!("ctx": "run", "op": "opendir_proc",
         "msg": "opened /proc directory",
         "fd": fd_proc,
-        "id": id_proc);
+        "f_mode": f_mode_proc,
+        "mnt_id": mnt_id_proc);
     _PROC_FD.set(fd_proc).or(Err(Errno::EAGAIN))?;
-    _PROC_MNT_ID.set(id_proc).or(Err(Errno::EAGAIN))?;
+    _PROC_F_MODE.set(f_mode_proc).or(Err(Errno::EAGAIN))?;
+    _PROC_MNT_ID.set(mnt_id_proc).or(Err(Errno::EAGAIN))?;
 
     let fd_null = nix::fcntl::openat(
         fd_root,
@@ -2581,7 +2714,8 @@ pub fn proc_init() -> Result<(), Errno> {
         OFlag::O_PATH | OFlag::O_CLOEXEC,
         Mode::empty(),
     )?;
-    let id_null = fstatx(&fd_null, mask).map(|stx| stx.stx_mnt_id)?;
+    let (f_mode_null, mnt_id_null) = fstatx(&fd_null, mask)
+        .map(|stx| (stx.stx_mode & !(libc::S_IFMT as u16), stx.stx_mnt_id))?;
 
     // SAFETY: Validate what we've opened is indeed `/dev/null`.
     crate::validate_dev_null(fd_null.as_raw_fd())?;
@@ -2593,9 +2727,11 @@ pub fn proc_init() -> Result<(), Errno> {
     info!("ctx": "run", "op": "opendev_null",
         "msg": "opened /dev/null",
         "fd": fd_null,
-        "id": id_null);
+        "f_mode": f_mode_null,
+        "mnt_id": mnt_id_null);
     _NULL_FD.set(fd_null).or(Err(Errno::EAGAIN))?;
-    _NULL_MNT_ID.set(id_null).or(Err(Errno::EAGAIN))?;
+    _NULL_F_MODE.set(f_mode_null).or(Err(Errno::EAGAIN))?;
+    _NULL_MNT_ID.set(mnt_id_null).or(Err(Errno::EAGAIN))?;
 
     Ok(())
 }
@@ -2624,6 +2760,7 @@ pub(crate) static FILEMAP: Lazy<FileMap> = Lazy::new(|| {
     let entry = FileMapEntry::new(
         ROOT_FD().into(),
         Some(FileType::Dir),
+        Some(ROOT_F_MODE()),
         Some(ROOT_MNT_ID()),
         None,
     );
@@ -2632,6 +2769,7 @@ pub(crate) static FILEMAP: Lazy<FileMap> = Lazy::new(|| {
     let entry = FileMapEntry::new(
         PROC_FD().into(),
         Some(FileType::Dir),
+        Some(PROC_F_MODE()),
         Some(PROC_MNT_ID()),
         None,
     );
@@ -2640,6 +2778,7 @@ pub(crate) static FILEMAP: Lazy<FileMap> = Lazy::new(|| {
     let entry = FileMapEntry::new(
         NULL_FD().into(),
         Some(FileType::Chr),
+        Some(NULL_F_MODE()),
         Some(NULL_MNT_ID()),
         None,
     );
@@ -2650,6 +2789,3 @@ pub(crate) static FILEMAP: Lazy<FileMap> = Lazy::new(|| {
 
 // FileMap default capacity.
 pub(crate) const FILECAP: usize = 8;
-
-/// Reference to the global PidFd map.
-pub static PIDFD_MAP: OnceLock<PidFdMap> = OnceLock::new();
