@@ -131,6 +131,8 @@ pub(crate) static UNLINKAT_COOKIE_ARG4: Lazy<Cookie> =
 #[allow(clippy::disallowed_methods)]
 pub(crate) static UNLINKAT_COOKIE_ARG5: Lazy<Cookie> =
     Lazy::new(|| getcookie().expect("getcookie"));
+#[allow(clippy::disallowed_methods)]
+pub(crate) static LINKAT_COOKIE_ARG5: Lazy<Cookie> = Lazy::new(|| getcookie().expect("getcookie"));
 
 /// These are used in `syd::fs::seccomp_notify_addfd`.
 #[allow(clippy::disallowed_methods)]
@@ -154,6 +156,17 @@ pub(crate) static SECCOMP_IOCTL_NOTIF_SEND_COOKIE_ARG4: Lazy<Cookie> =
 pub(crate) static SECCOMP_IOCTL_NOTIF_SEND_COOKIE_ARG5: Lazy<Cookie> =
     Lazy::new(|| getcookie().expect("getcookie"));
 
+/// These are used in `syd::proc::procmap_query`.
+#[allow(clippy::disallowed_methods)]
+pub(crate) static PROCMAP_QUERY_COOKIE_ARG3: Lazy<Cookie> =
+    Lazy::new(|| getcookie().expect("getcookie"));
+#[allow(clippy::disallowed_methods)]
+pub(crate) static PROCMAP_QUERY_COOKIE_ARG4: Lazy<Cookie> =
+    Lazy::new(|| getcookie().expect("getcookie"));
+#[allow(clippy::disallowed_methods)]
+pub(crate) static PROCMAP_QUERY_COOKIE_ARG5: Lazy<Cookie> =
+    Lazy::new(|| getcookie().expect("getcookie"));
+
 /// Safe openat2 confined by syscall cookies.
 #[inline(always)]
 pub(crate) fn safe_openat2<Fd: AsFd>(
@@ -544,3 +557,30 @@ pub(crate) fn safe_unlinkat<Fd: AsFd>(
 
     Errno::result(res).map(drop)
 }
+
+/// Safe linkat(2) confined by syscall cookies.
+#[inline(always)]
+pub(crate) fn safe_linkat<Fd1: AsFd, Fd2: AsFd>(
+    olddirfd: Fd1,
+    oldpath: &XPath,
+    newdirfd: Fd2,
+    newpath: &XPath,
+    flag: AtFlags,
+) -> Result<(), Errno> {
+    let res = oldpath.with_nix_path(|oldcstr| {
+        // SAFETY: In libc we trust.
+        newpath.with_nix_path(|newcstr| unsafe {
+            libc::syscall(
+                libc::SYS_linkat,
+                olddirfd.as_fd().as_raw_fd(),
+                oldcstr.as_ptr(),
+                newdirfd.as_fd().as_raw_fd(),
+                newcstr.as_ptr(),
+                flag.bits(),
+                *LINKAT_COOKIE_ARG5,
+            )
+        })
+    })??;
+
+    Errno::result(res).map(drop)
+}
