@@ -16,6 +16,7 @@ use std::{
     net::AddrParseError,
     num::{ParseIntError, TryFromIntError},
     str::Utf8Error,
+    thread::JoinHandle,
 };
 
 #[cfg(feature = "oci")]
@@ -40,6 +41,9 @@ use crate::{caps::errors::CapsError, elf::ElfError};
 /// Convenience type to use for functions returning a SydError.
 pub type SydResult<T> = std::result::Result<T, SydError>;
 
+/// Convenience type to use for join handlers returning a SydError.
+pub type SydJoinHandle<T> = JoinHandle<SydResult<T>>;
+
 /// A macro to create a SydError from the last errno.
 #[macro_export]
 macro_rules! lasterrno {
@@ -108,10 +112,7 @@ impl SydError {
         match self {
             Self::Nix(errno) => Some(*errno),
             Self::Proc(error) => proc_error_to_errno(error),
-            Self::Scmp(error) => error
-                .sysrawrc()
-                .map(|errno| errno.abs())
-                .map(Errno::from_raw),
+            Self::Scmp(error) => scmp2no(error),
             _ => None,
         }
     }
@@ -389,12 +390,10 @@ pub fn err2no(err: &std::io::Error) -> Errno {
         .unwrap_or(Errno::ENOSYS)
 }
 
-/// Convert a std::io::Error reference into a nix::Errno.
+/// Convert SeccompError to a nix::Errno.
 #[inline(always)]
-pub fn err2no_ref(err: &std::io::Error) -> Errno {
-    err.raw_os_error()
-        .map(Errno::from_raw)
-        .unwrap_or(Errno::ENOSYS)
+pub fn scmp2no(err: &SeccompError) -> Option<Errno> {
+    err.sysrawrc().map(|errno| errno.abs()).map(Errno::from_raw)
 }
 
 /// Convert proc errno to nix errno.
@@ -403,7 +402,7 @@ pub fn proc_error_to_errno(error: &ProcError) -> Option<Errno> {
     match error {
         ProcError::PermissionDenied(_) => Some(Errno::EACCES),
         ProcError::NotFound(_) => Some(Errno::ESRCH),
-        ProcError::Io(error, _) => Some(err2no_ref(error)),
+        ProcError::Io(error, _) => Some(err2no(error)),
         ProcError::Other(_) => None,
         ProcError::Incomplete(_) => None,
         ProcError::InternalError(_) => None,
