@@ -9,14 +9,14 @@
 use std::{
     hash::BuildHasher,
     io::{IoSlice, Read},
-    os::fd::{AsFd, AsRawFd, BorrowedFd, FromRawFd, OwnedFd, RawFd},
+    os::fd::{AsFd, AsRawFd, BorrowedFd, FromRawFd, IntoRawFd, OwnedFd, RawFd},
 };
 
 #[allow(clippy::disallowed_types)]
 use ahash::{AHasher, RandomState};
 use constant_time_eq::constant_time_eq;
 use crc::{Crc, CRC_32_ISO_HDLC, CRC_64_ECMA_182};
-use hex::{DisplayHex, FromHex};
+use data_encoding::{HEXLOWER, HEXLOWER_PERMISSIVE, HEXUPPER};
 use lexis::ToName;
 use memchr::arch::all::is_equal;
 use nix::{
@@ -125,14 +125,14 @@ impl Key {
 
     /// Creates an IV from a hex-encoded string.
     pub fn from_hex(hex: &[u8]) -> Result<Self, Errno> {
-        let key = <[u8; KEY_SIZE]>::from_hex(std::str::from_utf8(hex).or(Err(Errno::EINVAL))?)
-            .or(Err(Errno::EINVAL))?;
+        let key = HEXLOWER_PERMISSIVE.decode(hex).or(Err(Errno::EINVAL))?;
+        let key = key.as_slice().try_into().or(Err(Errno::EINVAL))?;
         Ok(Self::new(key))
     }
 
     /// Returns a hex-encoded string of the KEY.
     pub fn as_hex(&self) -> String {
-        self.0.unsecure().to_lower_hex_string()
+        HEXLOWER.encode(self.0.unsecure())
     }
 
     /// Check if the KEY is all zeros.
@@ -189,14 +189,14 @@ impl IV {
 
     /// Creates an IV from a hex-encoded string.
     pub fn from_hex(hex: &[u8]) -> Result<Self, Errno> {
-        let iv = <[u8; IV_SIZE]>::from_hex(std::str::from_utf8(hex).or(Err(Errno::EINVAL))?)
-            .or(Err(Errno::EINVAL))?;
+        let iv = HEXLOWER_PERMISSIVE.decode(hex).or(Err(Errno::EINVAL))?;
+        let iv = iv.as_slice().try_into().or(Err(Errno::EINVAL))?;
         Ok(Self::new(iv))
     }
 
     /// Returns a hex-encoded string of the IV.
     pub fn as_hex(&self) -> String {
-        self.0.unsecure().to_lower_hex_string()
+        HEXLOWER.encode(self.0.unsecure())
     }
 
     /// Check if the IV is all zeros.
@@ -286,7 +286,7 @@ impl AsMut<[u8]> for IV {
 /// `Key` turns into `Alg` and is wiped from memory at startup.
 pub enum Secret {
     /// Encryption & Authentication sockets
-    Alg(OwnedFd, OwnedFd),
+    Alg(RawFd, RawFd),
     /// Uninitialized master key
     Key(Key),
 }
@@ -321,7 +321,7 @@ impl Secret {
 
         // SAFETY: Free key from secure memory,
         // and replace it with the KCAPI connection.
-        *self = Self::Alg(enc_fd, tag_fd);
+        *self = Self::Alg(enc_fd.into_raw_fd(), tag_fd.into_raw_fd());
 
         Ok(())
     }
@@ -378,10 +378,7 @@ pub fn hash<R: Read>(mut reader: R, func: HashAlgorithm) -> SydResult<Vec<u8>> {
     let digest = match hasher_state {
         HashState::Crc32(d) => d.finalize().to_be_bytes().to_vec(),
         HashState::Crc64(d) => d.finalize().to_be_bytes().to_vec(),
-        HashState::Md5(c) => {
-            let dg = c.compute();
-            dg.0.to_vec()
-        }
+        HashState::Md5(s) => s.finalize().to_vec(),
         HashState::Sha1(s) => s.finalize().to_vec(),
         HashState::Sha3_256(s) => s.finalize().to_vec(),
         HashState::Sha3_384(s) => s.finalize().to_vec(),
@@ -394,7 +391,7 @@ pub fn hash<R: Read>(mut reader: R, func: HashAlgorithm) -> SydResult<Vec<u8>> {
 /// Returns a concise summary of hmac(sha256) shash support in the kernel.
 pub fn hmac_sha256_info() -> String {
     match proc_crypto() {
-        Err(e) => format!("HMAC‐SHA256: failed to read /proc/crypto: {e}!"),
+        Err(e) => format!("HMAC-SHA256: failed to read /proc/crypto: {e}!"),
         Ok(table) => {
             if let Some(blocks) = table.crypto_blocks.get("hmac(sha256)") {
                 for block in blocks {
@@ -404,20 +401,20 @@ pub fn hmac_sha256_info() -> String {
                             SelfTest::Unknown => "unknown",
                         };
                         let internal = if block.internal {
-                            "in‐kernel"
+                            "in-kernel"
                         } else {
                             "external"
                         };
                         let fips = if block.fips_enabled {
                             "FIPS"
                         } else {
-                            "no‐FIPS"
+                            "no-FIPS"
                         };
 
                         return format!(
-                            "HMAC‐SHA256: Secure hash is supported via '{}' driver; \
+                            "HMAC-SHA256: Secure hash is supported via '{}' driver; \
 module '{}'; prio {}; refcnt {}; \
-self‐test: {}; {}; {}; \
+self-test: {}; {}; {}; \
 blocksize {}B; digestsize {}B.",
                             block.driver,
                             block.module,
@@ -432,7 +429,7 @@ blocksize {}B; digestsize {}B.",
                     }
                 }
             }
-            "HMAC‐SHA256: Secure hash is unsupported!".to_string()
+            "HMAC-SHA256: Secure hash is unsupported!".to_string()
         }
     }
 }
@@ -540,21 +537,21 @@ pub fn aes_ctr_info() -> String {
                             SelfTest::Unknown => "unknown",
                         };
                         let internal = if block.internal {
-                            "in‐kernel"
+                            "in-kernel"
                         } else {
                             "external"
                         };
                         let fips = if block.fips_enabled {
                             "FIPS"
                         } else {
-                            "no‐FIPS"
+                            "no-FIPS"
                         };
                         let async_cap = if sk.async_capable { "async" } else { "sync" };
 
                         return format!(
-                            "AES‐CTR: Symmetric‐key cipher is supported via '{}' driver; \
+                            "AES-CTR: Symmetric-key cipher is supported via '{}' driver; \
 module '{}'; prio {}; refcnt {}; \
-self‐test: {}; {}; {}; {}; \
+self-test: {}; {}; {}; {}; \
 key {}–{}B; iv {}B; chunk {}B; walk {}B.",
                             block.driver,
                             block.module,
@@ -573,7 +570,7 @@ key {}–{}B; iv {}B; chunk {}B; walk {}B.",
                     }
                 }
             }
-            "AES‐CTR: Symmetric‐key cipher is unsupported!".to_string()
+            "AES-CTR: Symmetric-key cipher is unsupported!".to_string()
         }
     }
 }
@@ -1078,9 +1075,9 @@ pub fn get_at_random_u64() -> (u64, u64) {
 pub fn get_at_random_hex(upper: bool) -> String {
     let rnd = get_at_random();
     if upper {
-        rnd.to_upper_hex_string()
+        HEXUPPER.encode(rnd)
     } else {
-        rnd.to_lower_hex_string()
+        HEXLOWER.encode(rnd)
     }
 }
 
@@ -1149,7 +1146,6 @@ impl BuildHasher for SydRandomState {
 mod tests {
     use std::io::Cursor;
 
-    use hex::DisplayHex;
     use nix::{fcntl::open, sys::stat::Mode};
 
     use super::*;
@@ -1346,7 +1342,7 @@ mod tests {
         for case in HASH_TEST_CASES {
             let input_cursor = Cursor::new(case.0);
             let result = match hash(input_cursor, case.2) {
-                Ok(hash) => hash.to_upper_hex_string(),
+                Ok(hash) => HEXUPPER.encode(&hash),
                 Err(e) => {
                     errors.push(format!(
                         "Hashing failed for {:?} with error: {:?}",
@@ -1383,7 +1379,7 @@ mod tests {
         for case in cases {
             let input_cursor = Cursor::new(input.clone());
             let result = match hash(input_cursor, case.0) {
-                Ok(hash) => hash.to_upper_hex_string(),
+                Ok(hash) => HEXUPPER.encode(&hash),
                 Err(e) => {
                     errors.push(format!(
                         "Hashing failed for {:?} with error: {:?}",
@@ -1422,7 +1418,7 @@ mod tests {
         for case in cases {
             let input_cursor = Cursor::new(input.clone());
             let result = match hash(input_cursor, case.0) {
-                Ok(hash) => hash.to_upper_hex_string(),
+                Ok(hash) => HEXUPPER.encode(&hash),
                 Err(e) => {
                     errors.push(format!(
                         "Hashing failed for {:?} with error: {:?}",
@@ -1530,7 +1526,7 @@ mod tests {
             };
 
             // Step 6: Convert the HMAC tag to a hex string
-            let computed_hex = hmac_result.unsecure().to_lower_hex_string();
+            let computed_hex = HEXLOWER.encode(hmac_result.unsecure());
 
             // Step 7: Compare with the expected output
             if expected_hmac.len() < 64 {
