@@ -7,14 +7,18 @@
 // SPDX-License-Identifier: GPL-3.0
 
 use libseccomp::ScmpNotifResp;
-use nix::errno::Errno;
+use nix::{
+    errno::Errno,
+    sys::signal::{kill, Signal},
+    unistd::Pid,
+};
 
 use crate::{
     compat::{fstatx, statx},
     config::PROC_FILE,
     debug, error,
     fs::{readlinkat, CanonicalPath, FileInfo, FsFlags},
-    hook::{PathArgs, RemoteProcess, SysArg, SysFlags, UNotifyEventRequest},
+    hook::{RemoteProcess, SysArg, SysFlags, UNotifyEventRequest},
     kernel::{sandbox_path, syscall_path_handler},
     path::XPathBuf,
     ptrace::{ptrace_get_error, ptrace_syscall_info, ptrace_syscall_info_seccomp},
@@ -24,7 +28,7 @@ use crate::{
 // Note, chdir is a ptrace(2) hook, not a seccomp hook!
 // The seccomp hook is only used with trace/allow_unsafe_ptrace:1.
 pub(crate) fn sysenter_chdir<'a>(
-    process: &RemoteProcess,
+    pid: Pid,
     sandbox: &SandboxGuard,
     data: ptrace_syscall_info_seccomp,
 ) -> Result<CanonicalPath<'a>, Errno> {
@@ -39,12 +43,10 @@ pub(crate) fn sysenter_chdir<'a>(
     }
 
     // Read remote path.
-    let (path, _, _) =
-        // SAFETY: PidFd is validated.
-        unsafe { process.read_path(sandbox, data.args, arg, false, None) }?;
-    if !process.is_alive() {
-        return Err(Errno::ESRCH);
-    }
+    let process = RemoteProcess::new(pid);
+
+    // SAFETY: This is a ptrace hook, the PID cannot be validated.
+    let (path, _, _, _) = unsafe { process.read_path(sandbox, data.args, arg, false, None) }?;
 
     // Check for chroot, allow for the
     // common `cd /` use case.
@@ -56,6 +58,7 @@ pub(crate) fn sysenter_chdir<'a>(
         };
     }
 
+    let hide = sandbox.enabled(Capability::CAP_STAT);
     let mut caps = Capability::empty();
     if let Some(typ) = path.typ.as_ref() {
         if typ.is_dir() {
@@ -65,11 +68,10 @@ pub(crate) fn sysenter_chdir<'a>(
         return Err(Errno::ENOENT);
     }
 
-    sandbox_path(None, sandbox, process, path.abs(), caps, true, "chdir")?;
+    sandbox_path(None, sandbox, pid, path.abs(), caps, hide, "chdir")?;
 
     if !caps.contains(Capability::CAP_CHDIR) {
-        // SAFETY: Return this after sandboxing
-        // to honour hidden paths.
+        // SAFETY: Return this after sandboxing to honour hidden paths.
         return Err(Errno::ENOTDIR);
     }
 
@@ -78,12 +80,12 @@ pub(crate) fn sysenter_chdir<'a>(
 
 #[allow(clippy::cognitive_complexity)]
 pub(crate) fn sysexit_chdir(
-    process: RemoteProcess,
+    pid: Pid,
     info: ptrace_syscall_info,
     path: CanonicalPath,
 ) -> Result<(), Errno> {
     // Check for successful sigaction exit.
-    match ptrace_get_error(process.pid, info.arch) {
+    match ptrace_get_error(pid, info.arch) {
         Ok(None) => {
             // Successful chdir call, validate CWD magiclink.
         }
@@ -91,16 +93,16 @@ pub(crate) fn sysexit_chdir(
             // Unsuccessful chdir call, continue process.
             return Ok(());
         }
+        Err(Errno::ESRCH) => return Err(Errno::ESRCH),
         Err(_) => {
-            // SAFETY: Failed to get return value,
-            // terminate the process.
-            let _ = process.pidfd_kill(libc::SIGKILL);
+            // SAFETY: Failed to get return value, terminate the process.
+            let _ = kill(pid, Some(Signal::SIGKILL));
             return Err(Errno::ESRCH);
         }
     };
 
     // SAFETY: Validate /proc/$pid/cwd against TOCTTOU!
-    let mut pfd = XPathBuf::from_pid(process.pid);
+    let mut pfd = XPathBuf::from_pid(pid);
     pfd.push(b"cwd");
 
     #[allow(clippy::disallowed_methods)]
@@ -113,10 +115,10 @@ pub(crate) fn sysexit_chdir(
             // assume TOCTTOU: terminate the process.
             error!("ctx": "chdir", "op": "fstat_dir_fd",
                 "err": format!("failed to fstat dir-fd for `{path}': {errno}"),
-                "pid": process.pid.as_raw(),
+                "pid": pid.as_raw(),
                 "path": &path,
                 "errno": errno as i32);
-            let _ = process.pidfd_kill(libc::SIGKILL);
+            let _ = kill(pid, Some(Signal::SIGKILL));
             return Err(Errno::ESRCH);
         }
     };
@@ -128,10 +130,10 @@ pub(crate) fn sysexit_chdir(
             // assume TOCTTOU: terminate the process.
             error!("ctx": "chdir", "op": "stat_cwd_symlink",
                 "err": format!("failed to stat cwd-symlink for `{path}': {errno}"),
-                "pid": process.pid.as_raw(),
+                "pid": pid.as_raw(),
                 "path": &path,
                 "errno": errno as i32);
-            let _ = process.pidfd_kill(libc::SIGKILL);
+            let _ = kill(pid, Some(Signal::SIGKILL));
             return Err(Errno::ESRCH);
         }
     };
@@ -157,19 +159,19 @@ pub(crate) fn sysexit_chdir(
             .unwrap_or_else(|| XPathBuf::from("?"));
         error!("ctx": "chdir", "op": "dir_mismatch",
             "err": format!("dir mismatch detected for directory `{path}' -> `{cwd}': assume TOCTTOU!"),
-            "pid": process.pid.as_raw(),
+            "pid": pid.as_raw(),
             "path": &path,
             "real": cwd,
             "cwd_mount_id": stx_cwd.stx_mnt_id,
             "dir_mount_id": stx_fd.stx_mnt_id,
             "cwd_inode": stx_cwd.stx_ino,
             "dir_inode": stx_fd.stx_ino);
-        let _ = process.pidfd_kill(libc::SIGKILL);
+        let _ = kill(pid, Some(Signal::SIGKILL));
         return Err(Errno::ESRCH);
     } else {
         debug!("ctx": "chdir", "op": "verify_chdir",
             "msg": format!("dir change to `{path}' approved"),
-            "pid": process.pid.as_raw(),
+            "pid": pid.as_raw(),
             "path": &path,
             "cwd_mount_id": stx_cwd.stx_mnt_id,
             "dir_mount_id": stx_fd.stx_mnt_id,
@@ -187,58 +189,55 @@ pub(crate) fn sys_chdir(request: UNotifyEventRequest) -> ScmpNotifResp {
         ..Default::default()
     }];
 
-    syscall_path_handler(
-        request,
-        "chdir",
-        argv,
-        |path_args: PathArgs, request, sandbox| {
-            drop(sandbox); // release the read-lock.
+    syscall_path_handler(request, "chdir", argv, |path_args, request, sandbox| {
+        drop(sandbox); // release the read-lock.
 
-            // SAFETY: SysArg has one element.
-            #[allow(clippy::disallowed_methods)]
-            if let Some(typ) = path_args.0.as_ref().unwrap().typ.as_ref() {
-                if !typ.is_dir() {
-                    return Ok(request.fail_syscall(Errno::ENOTDIR));
-                }
-            } else {
-                return Ok(request.fail_syscall(Errno::ENOENT));
+        // SAFETY: SysArg has one element.
+        #[allow(clippy::disallowed_methods)]
+        if let Some(typ) = path_args.0.as_ref().unwrap().typ.as_ref() {
+            if !typ.is_dir() {
+                return Ok(request.fail_syscall(Errno::ENOTDIR));
             }
+        } else {
+            return Ok(request.fail_syscall(Errno::ENOENT));
+        }
 
-            // SAFETY: This is vulnerable to TOCTTOU.
-            // We only use this hook with trace/allow_unsafe_ptrace:1
-            // hence the user is aware of the consequences.
-            Ok(unsafe { request.continue_syscall() })
-        },
-    )
+        // SAFETY: This is vulnerable to TOCTTOU.
+        // We only use this hook with trace/allow_unsafe_ptrace:1
+        // hence the user is aware of the consequences.
+        Ok(unsafe { request.continue_syscall() })
+    })
 }
 
 pub(crate) fn sys_fchdir(request: UNotifyEventRequest) -> ScmpNotifResp {
-    // SAFETY: fchdir is fd-only, so UNSAFE_CONT is ok.
+    // SAFETY:
+    // 1. fchdir is fd-only, so UNSAFE_CONT is ok.
+    // 2. We do not pass `fsflags` which defaults to MUST_PATH.
     let argv = &[SysArg {
         dirfd: Some(0),
         flags: SysFlags::UNSAFE_CONT,
         ..Default::default()
     }];
 
-    syscall_path_handler(
-        request,
-        "fchdir",
-        argv,
-        |path_args: PathArgs, request, sandbox| {
-            drop(sandbox); // release the read-lock.
+    syscall_path_handler(request, "fchdir", argv, |path_args, request, sandbox| {
+        drop(sandbox); // release the read-lock.
 
-            // SAFETY: SysArg has one element.
-            #[allow(clippy::disallowed_methods)]
-            if let Some(typ) = path_args.0.as_ref().unwrap().typ.as_ref() {
-                if !typ.is_dir() {
-                    return Ok(request.fail_syscall(Errno::ENOTDIR));
-                }
-            } else {
-                return Ok(request.fail_syscall(Errno::ENOENT));
+        // SAFETY: SysArg has one element.
+        #[allow(clippy::disallowed_methods)]
+        let path = path_args.0.as_ref().unwrap();
+
+        // Check file type.
+        if let Some(typ) = path.typ.as_ref() {
+            if !typ.is_dir() {
+                // Deny non-directory with ENOTDIR.
+                return Ok(request.fail_syscall(Errno::ENOTDIR));
             }
+        } else {
+            // No file type, file disappeared mid-way?
+            return Ok(request.fail_syscall(Errno::ENOTDIR));
+        }
 
-            // SAFETY: fchdir is fd-only.
-            Ok(unsafe { request.continue_syscall() })
-        },
-    )
+        // SAFETY: fchdir is fd-only.
+        Ok(unsafe { request.continue_syscall() })
+    })
 }
