@@ -9,13 +9,14 @@
 use std::{
     fs::File,
     io::{Seek, SeekFrom},
-    os::fd::AsFd,
 };
 
 use libseccomp::ScmpNotifResp;
 use nix::{
     errno::Errno,
-    fcntl::{AtFlags, OFlag},
+    fcntl::AtFlags,
+    sys::signal::{kill, Signal},
+    unistd::Pid,
     NixPath,
 };
 
@@ -23,7 +24,7 @@ use crate::{
     config::{PTRACE_DATA_EXECVE, PTRACE_DATA_EXECVEAT},
     elf::{ElfError, ElfFileType, ElfType, ExecutableFile, LinkingType},
     err2no, error,
-    fs::{is_executable, safe_open, set_nonblock, FsFlags, MaybeFd, AT_EXECVE_CHECK},
+    fs::{is_executable, set_nonblock, FileType, FsFlags, MaybeFd, AT_EXECVE_CHECK},
     hook::{RemoteProcess, SysArg, SysFlags, UNotifyEventRequest},
     kernel::{sandbox_path, to_atflags},
     ptrace::ptrace_syscall_info,
@@ -35,10 +36,10 @@ use crate::{
 // The seccomp hooks are only used with trace/allow_unsafe_ptrace:1.
 #[allow(clippy::cognitive_complexity)]
 pub(crate) fn sysenter_exec(
-    process: &RemoteProcess,
+    pid: Pid,
     sandbox: &SandboxGuard,
     info: ptrace_syscall_info,
-) -> Result<ExecutableFile, Errno> {
+) -> Result<(File, ExecutableFile), Errno> {
     let data = if let Some(data) = info.seccomp() {
         data
     } else {
@@ -89,12 +90,11 @@ pub(crate) fn sysenter_exec(
     };
 
     // Read remote path.
+    let process = RemoteProcess::new(pid);
+
+    // SAFETY: This is a ptrace hook, the PID cannot be validated.
     let (mut path, _, _, empty_path) =
-        // SAFETY: We will validate the PidFd afterwards.
         unsafe { process.read_path(sandbox, data.args, arg, false, None) }?;
-    if !process.is_alive() {
-        return Err(Errno::ESRCH);
-    }
 
     // Call sandbox access checker if Exec sandboxing is on.
     let caps = sandbox.getcaps(Capability::CAP_EXEC | Capability::CAP_TPE);
@@ -104,7 +104,7 @@ pub(crate) fn sysenter_exec(
         sandbox_path(
             None,
             sandbox,
-            process,
+            pid,
             path.abs(),
             Capability::CAP_EXEC,
             hide,
@@ -125,6 +125,16 @@ pub(crate) fn sysenter_exec(
         return Err(Errno::ELOOP);
     }
 
+    // SAFETY: Return EACCES without any more processing
+    // if the file is not a regular file or a memory fd.
+    // Mfd check depends on trace/allow_unsafe_memfd option.
+    match path.typ.as_ref() {
+        Some(FileType::Reg) => {}
+        Some(FileType::Mfd) if sandbox.allow_unsafe_memfd() => {}
+        Some(FileType::Dir) => return Err(Errno::EISDIR),
+        _ => return Err(Errno::EACCES),
+    };
+
     // SAFETY:
     // 1. Return EACCES without any more processing
     //    if the file is not executable.
@@ -143,7 +153,7 @@ pub(crate) fn sysenter_exec(
                 "err": format!("max crashes {} exceeded, execution of `{bin}' denied",
                     sandbox.segvguard_maxcrashes),
                 "tip": "increase `segvguard/maxcrashes'",
-                "pid": process.pid.as_raw(), "path": path.abs());
+                "pid": pid.as_raw(), "path": path.abs());
         }
 
         match action {
@@ -152,15 +162,15 @@ pub(crate) fn sysenter_exec(
             Action::Panic => panic!(),
             Action::Exit => std::process::exit(libc::EACCES),
             Action::Stop => {
-                let _ = process.pidfd_kill(libc::SIGSTOP);
+                let _ = kill(process.pid, Some(Signal::SIGSTOP));
                 return Err(Errno::EACCES);
             }
             Action::Abort => {
-                let _ = process.pidfd_kill(libc::SIGABRT);
+                let _ = kill(process.pid, Some(Signal::SIGABRT));
                 return Err(Errno::EACCES);
             }
             Action::Kill => {
-                let _ = process.pidfd_kill(libc::SIGKILL);
+                let _ = kill(process.pid, Some(Signal::SIGKILL));
                 return Err(Errno::EACCES);
             }
         }
@@ -168,12 +178,16 @@ pub(crate) fn sysenter_exec(
 
     // Trusted Path Execution.
     if caps.contains(Capability::CAP_TPE) {
-        let action = sandbox.check_tpe(path.abs());
+        // MUST_PATH ensures path.dir is Some.
+        #[allow(clippy::disallowed_methods)]
+        let file = path.dir.as_ref().unwrap();
+        let (action, msg) = sandbox.check_tpe(file, path.abs());
         if !matches!(action, Action::Allow | Action::Filter) {
             // TODO: Fix proc_mmap to work in ptrace hooks.
+            let msg = msg.as_deref().unwrap_or("?");
             error!("ctx": "trusted_path_execution",
-                "err": "exec from untrusted path blocked",
-                "pid": process.pid.as_raw(), "path": path.abs(),
+                "err": format!("exec from untrusted path blocked: {msg}"),
+                "pid": pid.as_raw(), "path": path.abs(),
                 "sys": syscall_name, "arch": info.arch, "args": data.args,
                 "tip": "move the binary to a safe location or use `sandbox/tpe:off'");
         }
@@ -183,50 +197,31 @@ pub(crate) fn sysenter_exec(
             Action::Panic => panic!(),
             Action::Exit => std::process::exit(libc::EACCES),
             Action::Stop => {
-                let _ = process.pidfd_kill(libc::SIGSTOP);
+                let _ = kill(pid, Some(Signal::SIGSTOP));
                 return Err(Errno::EACCES);
             }
             Action::Abort => {
-                let _ = process.pidfd_kill(libc::SIGABRT);
+                let _ = kill(pid, Some(Signal::SIGABRT));
                 return Err(Errno::EACCES);
             }
             Action::Kill => {
-                let _ = process.pidfd_kill(libc::SIGKILL);
+                let _ = kill(pid, Some(Signal::SIGKILL));
                 return Err(Errno::EACCES);
             }
         }
     }
 
-    // SAFETY: We will read from the regular files only and parse ELF.
-    if let Some(file_type) = path.typ.as_ref() {
-        if file_type.is_dir() {
-            return Err(Errno::EISDIR);
-        } else if !file_type.is_file() {
-            return Err(Errno::EACCES);
+    assert!(
+        path.base.is_empty(),
+        "BUG: MUST_PATH returned a directory for exec, report a bug!"
+    );
+    let mut file = match path.dir.take() {
+        Some(MaybeFd::Owned(fd)) => {
+            set_nonblock(&fd, false)?;
+            File::from(fd)
         }
-    } else {
-        return Err(Errno::ENOENT);
-    }
-
-    // SAFETY: Use safe open to avoid TOCTOU!
-    let flags = OFlag::O_RDONLY | OFlag::O_NOFOLLOW | OFlag::O_NOCTTY | OFlag::O_CLOEXEC;
-
-    let mut file = if path.base.is_empty() {
-        match path.dir.take() {
-            Some(MaybeFd::Owned(fd)) => {
-                set_nonblock(&fd, false)?;
-                Ok(fd)
-            }
-            _ => return Err(Errno::ENOEXEC),
-        }
-    } else {
-        safe_open(
-            path.dir.as_ref().map(|fd| fd.as_fd()).ok_or(Errno::EBADF)?,
-            path.base,
-            flags,
-        )
-    }
-    .map(File::from)?;
+        _ => return Err(Errno::ENOEXEC),
+    };
 
     // Parse ELF as necessary for restrictions.
     let deny_script = sandbox.deny_script();
@@ -237,15 +232,25 @@ pub(crate) fn sysenter_exec(
     let restrict_xs = !sandbox.allow_unsafe_stack();
 
     // Shared library execution depends on trace/allow_unsafe_exec:1.
-    // unsafe_exec also means no ptrace, hence we can never be here.
-    let restrict_ldd = true /* !sandbox.allow_unsafe_exec() */;
+    let restrict_ldd = !sandbox.allow_unsafe_exec();
 
     let check_linking = restrict_ldd || restrict_dyn || restrict_sta || restrict_pie || restrict_xs;
 
+    // Ensure the file offset is maintained,
+    // as the file might be sharing the OFD
+    // with the sandbox process.
+    //
+    // SAFETY: Deny non-seekable files such as FIFOs with EACCES.
+    let offset = file.stream_position().or(Err(Errno::EACCES))?;
+
     let result = (|| -> Result<ExecutableFile, ElfError> {
         // Parse ELF and reset the file offset.
+        if offset != 0 {
+            file.rewind().map_err(ElfError::IoError)?;
+        }
         let result = ExecutableFile::parse(&file, check_linking);
-        file.seek(SeekFrom::Start(0)).map_err(ElfError::IoError)?;
+        file.seek(SeekFrom::Start(offset))
+            .map_err(ElfError::IoError)?;
         result
     })();
 
@@ -256,7 +261,7 @@ pub(crate) fn sysenter_exec(
             if !sandbox.filter_path(Capability::CAP_EXEC, path.abs()) {
                 error!("ctx": "parse_elf",
                     "err": format!("io error: {errno}"),
-                    "pid": process.pid.as_raw(), "path": path.abs());
+                    "pid": pid.as_raw(), "path": path.abs());
             }
             return Err(errno);
         }
@@ -264,7 +269,7 @@ pub(crate) fn sysenter_exec(
             if !sandbox.filter_path(Capability::CAP_EXEC, path.abs()) {
                 error!("ctx": "parse_elf",
                     "err": "invalid ELF file",
-                    "pid": process.pid.as_raw(), "path": path.abs());
+                    "pid": pid.as_raw(), "path": path.abs());
             }
             return Err(Errno::ENOEXEC);
         }
@@ -272,7 +277,7 @@ pub(crate) fn sysenter_exec(
             if !sandbox.filter_path(Capability::CAP_EXEC, path.abs()) {
                 error!("ctx": "parse_elf",
                     "err": "malformed ELF file",
-                    "pid": process.pid.as_raw(), "path": path.abs());
+                    "pid": pid.as_raw(), "path": path.abs());
             }
             return Err(Errno::ENOEXEC);
         }
@@ -283,7 +288,7 @@ pub(crate) fn sysenter_exec(
         if !sandbox.filter_path(Capability::CAP_EXEC, path.abs()) {
             error!("ctx": "deny_script",
                 "err": "script execution denied",
-                "pid": process.pid.as_raw(), "path": path.abs(),
+                "pid": pid.as_raw(), "path": path.abs(),
                 "exe": format!("{exe}"));
         }
         return Err(Errno::EACCES);
@@ -302,7 +307,7 @@ pub(crate) fn sysenter_exec(
         if !sandbox.filter_path(Capability::CAP_EXEC, path.abs()) {
             error!("ctx": "check_elf",
                 "err": "ld.so exec-indirection",
-                "pid": process.pid.as_raw(), "path": path.abs(),
+                "pid": pid.as_raw(), "path": path.abs(),
                 "exe": format!("{exe}"));
         }
         return Err(Errno::EACCES);
@@ -311,7 +316,7 @@ pub(crate) fn sysenter_exec(
     if !is_script && restrict_pie && matches!(exe, ExecutableFile::Elf { pie: false, .. }) {
         if !sandbox.filter_path(Capability::CAP_EXEC, path.abs()) {
             error!("ctx": "check_elf", "err": "not PIE",
-                "pid": process.pid.as_raw(), "path": path.abs(),
+                "pid": pid.as_raw(), "path": path.abs(),
                 "tip": "configure `trace/allow_unsafe_nopie:1'",
                 "exe": format!("{exe}"));
         }
@@ -321,7 +326,7 @@ pub(crate) fn sysenter_exec(
     if !is_script && restrict_xs && matches!(exe, ExecutableFile::Elf { xs: true, .. }) {
         if !sandbox.filter_path(Capability::CAP_EXEC, path.abs()) {
             error!("ctx": "check_elf", "err": "execstack",
-                "pid": process.pid.as_raw(), "path": path.abs(),
+                "pid": pid.as_raw(), "path": path.abs(),
                 "tip": "configure `trace/allow_unsafe_stack:1'",
                 "exe": format!("{exe}"));
         }
@@ -340,7 +345,7 @@ pub(crate) fn sysenter_exec(
     {
         if !sandbox.filter_path(Capability::CAP_EXEC, path.abs()) {
             error!("ctx": "check_elf", "err": "32-bit",
-                "pid": process.pid.as_raw(), "path": path.abs(),
+                "pid": pid.as_raw(), "path": path.abs(),
                 "tip": "configure `trace/deny_elf32:0'",
                 "exe": format!("{exe}"));
         }
@@ -359,7 +364,7 @@ pub(crate) fn sysenter_exec(
     {
         if !sandbox.filter_path(Capability::CAP_EXEC, path.abs()) {
             error!("ctx": "check_elf", "err": "dynamic-link",
-                "pid": process.pid.as_raw(), "path": path.abs(),
+                "pid": pid.as_raw(), "path": path.abs(),
                 "tip": "configure `trace/deny_elf_dynamic:0'",
                 "exe": format!("{exe}"));
         }
@@ -378,7 +383,7 @@ pub(crate) fn sysenter_exec(
     {
         if !sandbox.filter_path(Capability::CAP_EXEC, path.abs()) {
             error!("ctx": "check_elf", "err": "static-link",
-                "pid": process.pid.as_raw(), "path": path,
+                "pid": pid.as_raw(), "path": path,
                 "tip": "configure `trace/deny_elf_static:0'",
                 "exe": format!("{exe}"));
         }
@@ -387,44 +392,51 @@ pub(crate) fn sysenter_exec(
 
     // Check for Force sandboxing.
     if sandbox.enabled(Capability::CAP_FORCE) {
-        match sandbox.check_force2(path.abs(), &mut file) {
+        // Calculate checksum and reset the file offset.
+        if offset != 0 {
+            file.rewind().or(Err(Errno::EACCES))?;
+        }
+        let result = sandbox.check_force2(path.abs(), &mut file);
+        file.seek(SeekFrom::Start(offset)).or(Err(Errno::EACCES))?;
+
+        match result {
             Ok(Action::Allow) => {}
             Ok(Action::Warn) => {
                 warn!("ctx": "verify_elf", "act": Action::Warn,
-                    "pid": process.pid.as_raw(), "path": path.abs(),
+                    "pid": pid.as_raw(), "path": path.abs(),
                     "tip": format!("configure `force+{path}:<checksum>'"));
             }
             Ok(Action::Filter) => return Err(Errno::EACCES),
             Ok(Action::Deny) => {
                 warn!("ctx": "verify_elf", "act": Action::Deny,
-                    "pid": process.pid.as_raw(), "path": path.abs(),
+                    "pid": pid.as_raw(), "path": path.abs(),
                     "tip": format!("configure `force+{path}:<checksum>'"));
                 return Err(Errno::EACCES);
             }
             Ok(Action::Stop) => {
                 warn!("ctx": "verify_elf", "act": Action::Stop,
-                    "pid": process.pid.as_raw(), "path": path.abs(),
+                    "pid": pid.as_raw(), "path": path.abs(),
                     "tip": format!("configure `force+{path}:<checksum>'"));
-                let _ = process.pidfd_kill(libc::SIGSTOP);
+                let _ = kill(pid, Some(Signal::SIGSTOP));
                 return Err(Errno::EACCES);
             }
             Ok(Action::Abort) => {
                 warn!("ctx": "verify_elf", "act": Action::Abort,
-                    "pid": process.pid.as_raw(), "path": path.abs(),
+                    "pid": pid.as_raw(), "path": path.abs(),
                     "tip": format!("configure `force+{path}:<checksum>'"));
-                let _ = process.pidfd_kill(libc::SIGABRT);
+                let _ = kill(pid, Some(Signal::SIGABRT));
                 return Err(Errno::EACCES);
             }
             Ok(Action::Kill) => {
                 warn!("ctx": "verify_elf", "act": Action::Kill,
-                    "pid": process.pid.as_raw(), "path": path.abs(),
+                    "pid": pid.as_raw(), "path": path.abs(),
                     "tip": format!("configure `force+{path}:<checksum>'"));
-                let _ = process.pidfd_kill(libc::SIGKILL);
+                let _ = kill(pid, Some(Signal::SIGKILL));
                 return Err(Errno::EACCES);
             }
             Ok(Action::Exit) => {
                 error!("ctx": "verify_elf", "act": Action::Exit,
-                    "pid": process.pid.as_raw(), "path": path.abs(),
+                    "pid": pid.as_raw(), "path": path.abs(),
                     "tip": format!("configure `force+{path}:<checksum>'"));
                 std::process::exit(libc::EACCES);
             }
@@ -432,7 +444,7 @@ pub(crate) fn sysenter_exec(
             Err(IntegrityError::Sys(errno)) => {
                 error!("ctx": "verify_elf",
                     "err": format!("system error during ELF checksum calculation: {errno}"),
-                    "pid": process.pid.as_raw(), "path": path.abs(),
+                    "pid": pid.as_raw(), "path": path.abs(),
                     "tip": format!("configure `force+{path}:<checksum>'"));
                 return Err(Errno::EACCES);
             }
@@ -444,18 +456,24 @@ pub(crate) fn sysenter_exec(
                 if !matches!(action, Action::Allow | Action::Filter) {
                     error!("ctx": "verify_elf", "act": action,
                         "err": format!("ELF checksum mismatch: {found} is not {expected}"),
-                        "pid": process.pid.as_raw(), "path": path.abs(),
+                        "pid": pid.as_raw(), "path": path.abs(),
                         "tip": format!("configure `force+{path}:<checksum>'"));
                 }
                 match action {
                     Action::Allow | Action::Warn => {}
                     Action::Filter | Action::Deny => return Err(Errno::EACCES),
                     Action::Stop | Action::Abort | Action::Kill => {
-                        let _ = process.pidfd_kill(
-                            action
-                                .signal()
-                                .map(|sig| sig as i32)
-                                .unwrap_or(libc::SIGKILL),
+                        let _ = kill(
+                            pid,
+                            Some(
+                                Signal::try_from(
+                                    action
+                                        .signal()
+                                        .map(|sig| sig as i32)
+                                        .unwrap_or(libc::SIGKILL),
+                                )
+                                .unwrap_or(Signal::SIGKILL),
+                            ),
                         );
                         return Err(Errno::EACCES);
                     }
@@ -473,7 +491,7 @@ pub(crate) fn sysenter_exec(
         return Err(Errno::ECANCELED);
     }
 
-    Ok(exe)
+    Ok((file, exe))
 }
 
 pub(crate) fn sys_execve(request: UNotifyEventRequest) -> ScmpNotifResp {
@@ -535,6 +553,7 @@ fn syscall_exec_handler(
         // and only the file executability check (aka AT_EXECVE_CHECK)
         // will be performed.
         let sandbox = request.get_sandbox();
+        let restrict_memfd = !sandbox.allow_unsafe_memfd();
 
         // Read remote path.
         let (path, _) = request.read_path(&sandbox, arg, false)?;
@@ -544,7 +563,7 @@ fn syscall_exec_handler(
             sandbox_path(
                 Some(&request),
                 &sandbox,
-                &RemoteProcess::new(request.scmpreq.pid()), // Unused when request.is_some()
+                request.scmpreq.pid(), // Unused when request.is_some()
                 path.abs(),
                 Capability::CAP_EXEC,
                 sandbox.enabled(Capability::CAP_STAT),
@@ -553,6 +572,29 @@ fn syscall_exec_handler(
         }
         drop(sandbox); // release the read-lock.
 
+        if !arg.fsflags.follow_last()
+            && path
+                .typ
+                .as_ref()
+                .map(|typ| typ.is_symlink() || typ.is_magic_link())
+                .unwrap_or(false)
+        {
+            // SAFETY: AT_SYMLINK_NOFOLLOW: If the file identified by dirfd
+            // and a non-NULL pathname is a symbolic link, then the call
+            // fails with the error ELOOP.
+            return Err(Errno::ELOOP);
+        }
+
+        // SAFETY: Return EACCES without any more processing
+        // if the file is not a regular file or a memory fd.
+        // Mfd check depends on trace/allow_unsafe_memfd option.
+        match path.typ.as_ref() {
+            Some(FileType::Reg) => {}
+            Some(FileType::Mfd) if !restrict_memfd => {}
+            Some(FileType::Dir) => return Err(Errno::EISDIR),
+            _ => return Err(Errno::EACCES),
+        };
+
         // SAFETY:
         // 1. Return EACCES without any more processing
         //    if the file is not executable.
