@@ -20,8 +20,8 @@ use crate::{
     config::{PAGE_SIZE, PROC_FILE},
     elf::ExecutableFile,
     error,
-    fs::{safe_open_magicsym, CanonicalPath},
-    hook::{RemoteProcess, UNotifyEventRequest},
+    fs::{is_writable_fd, safe_open_magicsym, CanonicalPath},
+    hook::UNotifyEventRequest,
     kernel::sandbox_path,
     path::XPathBuf,
     proc::{proc_mem_limit, proc_statm},
@@ -83,15 +83,22 @@ fn syscall_mem_handler(
     let mem_max = sandbox.mem_max;
     let mem_vm_max = sandbox.mem_vm_max;
     let mem_act = sandbox.default_action(Capability::CAP_MEM);
+    let restrict_memory = !sandbox.allow_unsafe_memory();
     let restrict_stack = !sandbox.allow_unsafe_stack();
 
-    if !exec && !force && !tpe && !restrict_stack && (!mem || (mem_max == 0 && mem_vm_max == 0)) {
+    if !exec
+        && !force
+        && !tpe
+        && !restrict_memory
+        && !restrict_stack
+        && (!mem || (mem_max == 0 && mem_vm_max == 0))
+    {
         // SAFETY: No pointer dereference in security check.
         // This is safe to continue.
         return unsafe { request.continue_syscall() };
     }
 
-    if (exec || force || tpe || restrict_stack)
+    if (exec || force || tpe || restrict_memory || restrict_stack)
         && is_prefix(name.as_bytes(), b"mmap")
         && req.data.args[2] & PROT_EXEC != 0
         && req.data.args[3] & MAP_ANONYMOUS == 0
@@ -110,20 +117,27 @@ fn syscall_mem_handler(
             Ok(fd) => fd,
             Err(_) => return request.fail_syscall(Errno::EBADF),
         };
+
+        // Step 1: Check if file is open for write,
+        // but set as PROT_READ|PROT_EXEC which breaks W^X!
+        // We do not need to check for PROT_WRITE here as
+        // this is already enforced at kernel-level when
+        // trace/allow_unsafe_memory:1 is not set at startup.
+        if restrict_memory && is_writable_fd(&fd).unwrap_or(true) {
+            return request.fail_syscall(Errno::EACCES);
+        }
+
         let mut path = match CanonicalPath::new_fd(fd.into(), req.pid(), remote_fd) {
             Ok(path) => path,
             Err(errno) => return request.fail_syscall(errno),
         };
 
-        // Step 1: Check for Exec sandboxing.
+        // Step 2: Check for Exec sandboxing.
         if exec {
-            // Unused when request.is_some()
-            let process = RemoteProcess::new(request.scmpreq.pid());
-
             if let Err(errno) = sandbox_path(
                 Some(&request),
                 &sandbox,
-                &process,
+                request.scmpreq.pid(), // Unused when request.is_some()
                 path.abs(),
                 Capability::CAP_EXEC,
                 false,
@@ -133,19 +147,23 @@ fn syscall_mem_handler(
             }
         }
 
-        // Step 2: Check for TPE sandboxing.
+        // Step 3: Check for TPE sandboxing.
         if tpe {
-            let action = sandbox.check_tpe(path.abs());
+            // MUST_PATH ensures path.dir is Some.
+            #[allow(clippy::disallowed_methods)]
+            let file = path.dir.as_ref().unwrap();
+            let (action, msg) = sandbox.check_tpe(file, path.abs());
             if !matches!(action, Action::Allow | Action::Filter) {
+                let msg = msg.as_deref().unwrap_or("?");
                 if verbose {
                     error!("ctx": "trusted_path_execution",
-                        "err": "library load from untrusted path blocked",
+                        "err": format!("library load from untrusted path blocked: {msg}"),
                         "sys": name, "path": &path,
                         "req": &request,
                         "tip": "move the library to a safe location or use `sandbox/tpe:off'");
                 } else {
                     error!("ctx": "trusted_path_execution",
-                        "err": "library load from untrusted path blocked",
+                        "err": format!("library load from untrusted path blocked: {msg}"),
                         "sys": name, "path": &path,
                         "pid": request.scmpreq.pid,
                         "tip": "move the library to a safe location or use `sandbox/tpe:off'");
@@ -167,13 +185,11 @@ fn syscall_mem_handler(
         if force || restrict_stack {
             // The following checks require the contents of the file.
             // SAFETY:
-            // 1. Reopen the file via `/proc/self/fd` to avoid sharing the file offset.
+            // 1. Reopen the file via `/proc/thread-self/fd` to avoid sharing the file offset.
             // 2. `path` is a remote-fd transfer which asserts `path.dir` is Some.
             #[allow(clippy::disallowed_methods)]
             let fd = path.dir.take().unwrap();
-
-            let mut pfd = XPathBuf::from("self/fd");
-            pfd.push_fd(fd.as_raw_fd());
+            let pfd = XPathBuf::from_self_fd(fd.as_raw_fd());
 
             let mut file =
                 match safe_open_magicsym(PROC_FILE(), &pfd, OFlag::O_RDONLY).map(File::from) {
@@ -184,14 +200,14 @@ fn syscall_mem_handler(
                 };
 
             if restrict_stack {
-                // Step 3: Check for non-executable stack.
+                // Step 4: Check for non-executable stack.
                 // An execstack library that is dlopened into an executable
                 // that is otherwise mapped no-execstack can change the
                 // stack permissions to executable! This has been
                 // (ab)used in at least one CVE:
                 // https://www.qualys.com/2023/07/19/cve-2023-38408/rce-openssh-forwarded-ssh-agent.txt
                 let result = (|file: &mut File| -> Result<(), Errno> {
-                    let exe = ExecutableFile::parse(&mut *file, true).or(Err(Errno::EBADF))?;
+                    let exe = ExecutableFile::parse(&mut *file, true).or(Err(Errno::EACCES))?;
                     if matches!(exe, ExecutableFile::Elf { xs: true, .. }) {
                         if !sandbox.filter_path(Capability::CAP_EXEC, path.abs()) {
                             if verbose {
@@ -222,7 +238,7 @@ fn syscall_mem_handler(
             }
 
             if force {
-                // Step 4: Check for Force sandboxing.
+                // Step 5: Check for Force sandboxing.
                 if restrict_stack && file.rewind().is_err() {
                     drop(sandbox); // release the read-lock.
                     return request.fail_syscall(Errno::EBADF);
