@@ -87,6 +87,9 @@ pub(crate) mod rename;
 /// Set UID/GID syscall handlers
 pub(crate) mod setid;
 
+/// Shared memory syscall handlers
+pub(crate) mod shm;
+
 /// {,rt_}sigaction(2) handler
 pub(crate) mod sigaction;
 
@@ -133,12 +136,16 @@ use memchr::memmem;
 use nix::{
     errno::Errno,
     fcntl::AtFlags,
-    sys::stat::{Mode, SFlag},
+    sys::{
+        signal::{kill, Signal},
+        stat::{Mode, SFlag},
+    },
+    unistd::Pid,
 };
 
 use crate::{
     fs::{to_valid_fd, CanonicalPath, FileInfo, FileType},
-    hook::{PathArgs, RemoteProcess, SysArg, UNotifyEventRequest},
+    hook::{PathArgs, SysArg, UNotifyEventRequest},
     notice,
     path::{XPath, XPathBuf},
     sandbox::{Action, Capability, SandboxGuard},
@@ -151,7 +158,7 @@ use crate::{
 pub(crate) fn sandbox_path(
     request: Option<&UNotifyEventRequest>,
     sandbox: &SandboxGuard,
-    process: &RemoteProcess,
+    pid: Pid,
     path: &XPath,
     caps: Capability,
     hide: bool,
@@ -179,7 +186,7 @@ pub(crate) fn sandbox_path(
     // Convert /proc/${pid} to /proc/self as necessary.
     let path = if let Some(p) = path.split_prefix(b"/proc") {
         let mut buf = itoa::Buffer::new();
-        let pid = buf.format(process.pid.as_raw());
+        let pid = buf.format(pid.as_raw());
         if let Some(p) = p.split_prefix(pid.as_bytes()) {
             let mut pdir = XPathBuf::from("/proc/self");
             pdir.push(p.as_bytes());
@@ -283,14 +290,14 @@ pub(crate) fn sandbox_path(
                 "tip": format!("configure `allow/{}+{}'",
                     caps.to_string().to_ascii_lowercase(),
                     path),
-                "pid": process.pid.as_raw());
+                "pid": pid.as_raw());
         } else {
             notice!("ctx": "access", "cap": caps, "act": action,
                 "sys": syscall_name, "path": &path,
                 "tip": format!("configure `allow/{}+{}'",
                     caps.to_string().to_ascii_lowercase(),
                     path),
-                "pid": process.pid.as_raw());
+                "pid": pid.as_raw());
         }
     }
 
@@ -311,7 +318,7 @@ pub(crate) fn sandbox_path(
             if let Some(request) = request {
                 let _ = request.pidfd_kill(libc::SIGSTOP);
             } else {
-                let _ = process.pidfd_kill(libc::SIGSTOP);
+                let _ = kill(pid, Some(Signal::SIGSTOP));
             }
             Err(deny_errno)
         }
@@ -319,7 +326,7 @@ pub(crate) fn sandbox_path(
             if let Some(request) = request {
                 let _ = request.pidfd_kill(libc::SIGABRT);
             } else {
-                let _ = process.pidfd_kill(libc::SIGABRT);
+                let _ = kill(pid, Some(Signal::SIGABRT));
             }
             Err(deny_errno)
         }
@@ -327,7 +334,7 @@ pub(crate) fn sandbox_path(
             if let Some(request) = request {
                 let _ = request.pidfd_kill(libc::SIGKILL);
             } else {
-                let _ = process.pidfd_kill(libc::SIGKILL);
+                let _ = kill(pid, Some(Signal::SIGKILL));
             }
             Err(deny_errno)
         }
@@ -427,7 +434,7 @@ where
                             #[allow(clippy::disallowed_methods)]
                             let files = request.crypt_map.as_ref().unwrap();
                             for (path, map) in
-                                &files.read().unwrap_or_else(|err| err.into_inner()).0
+                                files.read().unwrap_or_else(|err| err.into_inner()).iter()
                             {
                                 if info == map.info {
                                     found = Some(path.clone());
@@ -462,9 +469,6 @@ where
         }
 
         if !magic {
-            // Unused when request.is_some()
-            let process = RemoteProcess::new(request.scmpreq.pid());
-
             // Call sandbox access checker, skip magic paths.
             match (&paths[0], &paths[1]) {
                 (Some(path), None) => {
@@ -485,7 +489,7 @@ where
                     sandbox_path(
                         Some(&request),
                         &sandbox,
-                        &process,
+                        request.scmpreq.pid(), // Unused when request.is_some()
                         path.abs(),
                         caps,
                         hide,
@@ -500,7 +504,7 @@ where
                     sandbox_path(
                         Some(&request),
                         &sandbox,
-                        &process,
+                        request.scmpreq.pid(), // Unused when request.is_some()
                         path_0.abs(),
                         Capability::CAP_RENAME,
                         hide,
@@ -512,7 +516,7 @@ where
                         sandbox_path(
                             Some(&request),
                             &sandbox,
-                            &process,
+                            request.scmpreq.pid(), // Unused when request.is_some()
                             path_1.abs(),
                             Capability::CAP_CREATE,
                             hide,
