@@ -79,26 +79,16 @@ pub(crate) fn handle_accept(
     let cloexec = force_cloexec || flags.contains(SockFlag::SOCK_CLOEXEC);
     flags.insert(SockFlag::SOCK_CLOEXEC);
 
-    // SAFETY: Record blocking call so it can get invalidated.
-    let req = request.scmpreq;
-    let is_blocking = if !get_nonblock(&fd)? {
+    // Check whether we should block and ignore restarts.
+    let (is_blocking, ignore_restart) = if !get_nonblock(&fd)? {
         let ignore_restart = has_recv_timeout(&fd)?;
-
-        // Record the blocking call.
-        request.cache.add_sys_block(req, ignore_restart)?;
-
-        true
+        (true, ignore_restart)
     } else {
-        false
+        (false, false)
     };
 
     // Do the accept call.
-    let result = do_accept4(fd, flags);
-
-    // Remove invalidation record unless interrupted.
-    if is_blocking && !matches!(result, Err(Errno::EINTR)) {
-        request.cache.del_sys_block(req.id);
-    }
+    let result = do_accept4(fd, request, flags, is_blocking, ignore_restart);
 
     // Only now, bail if accept failed.
     let (fd, addr, addrlen_out) = result?;
@@ -145,7 +135,10 @@ pub(crate) fn handle_accept(
 
 fn do_accept4(
     fd: OwnedFd,
+    request: &UNotifyEventRequest,
     flags: SockFlag,
+    is_blocking: bool,
+    ignore_restart: bool,
 ) -> Result<(OwnedFd, SockaddrStorage, libc::socklen_t), Errno> {
     // Allocate storage for the address.
     let mut addr: [u8; std::mem::size_of::<SockaddrStorage>()] =
@@ -157,9 +150,25 @@ fn do_accept4(
     #[allow(clippy::cast_ptr_alignment)]
     let ptr = addr.as_mut_ptr() as *mut libc::sockaddr;
 
+    // SAFETY: Record blocking call so it can get invalidated.
+    if is_blocking {
+        request
+            .cache
+            .add_sys_block(request.scmpreq, ignore_restart)?;
+    };
+
     // SAFETY: In libc we trust.
-    let fd = Errno::result(unsafe { libc::accept4(fd.as_raw_fd(), ptr, &mut len, flags.bits()) })
-        .map(|fd| {
+    let result =
+        Errno::result(unsafe { libc::accept4(fd.as_raw_fd(), ptr, &mut len, flags.bits()) });
+
+    // Remove invalidation record unless interrupted.
+    if is_blocking {
+        request
+            .cache
+            .del_sys_block(request.scmpreq.id, matches!(result, Err(Errno::EINTR)))?;
+    }
+
+    let fd = result.map(|fd| {
         // SAFETY: accept4 returns a valid FD.
         unsafe { OwnedFd::from_raw_fd(fd) }
     })?;
