@@ -11,7 +11,7 @@ use std::{
     ffi::CString,
     net::IpAddr,
     os::{
-        fd::{AsRawFd, FromRawFd, OwnedFd, RawFd},
+        fd::{AsFd, AsRawFd, RawFd},
         unix::ffi::OsStrExt,
     },
 };
@@ -19,7 +19,10 @@ use std::{
 use libseccomp::ScmpNotifResp;
 use nix::{
     errno::Errno,
-    sys::socket::{AddressFamily, MsgFlags, SockFlag, SockType, SockaddrLike, SockaddrStorage},
+    fcntl::OFlag,
+    sys::socket::{
+        getsockname, AddressFamily, MsgFlags, SockFlag, SockType, SockaddrLike, SockaddrStorage,
+    },
     NixPath,
 };
 
@@ -27,8 +30,9 @@ use crate::{
     compat::{
         addr_family, fstatx, PF_ALG, PF_INET, PF_INET6, PF_NETLINK, PF_UNIX, PF_UNSPEC, STATX_INO,
     },
+    cookie::safe_socket,
     debug,
-    fs::{file_type, safe_canonicalize, CanonicalPath, FileType, FsFlags},
+    fs::{fd_status_flags, file_type, safe_canonicalize, CanonicalPath, FileType, FsFlags},
     hook::UNotifyEventRequest,
     kernel::net::{
         accept::handle_accept,
@@ -226,13 +230,7 @@ fn syscall_network_handler(request: UNotifyEventRequest, args: &[u64; 6], op: u8
                 let proto = libc::c_int::try_from(args[2]).or(Err(Errno::EAFNOSUPPORT))?;
                 let cloexec = force_cloexec || (stype & libc::SOCK_CLOEXEC != 0);
                 let stype = stype | libc::SOCK_CLOEXEC;
-
-                let fd =
-                    // SAFETY: We use libc version for convenience.
-                    Errno::result(unsafe { libc::socket(domain, stype, proto) }).map(|fd| {
-                        // SAFETY: socket returns a valid FD.
-                        unsafe { OwnedFd::from_raw_fd(fd) }
-                    })?;
+                let fd = safe_socket(domain, stype, proto)?;
 
                 if log_enabled!(LogLevel::Debug) {
                     let inode = fstatx(&fd, STATX_INO)
@@ -273,11 +271,12 @@ fn syscall_network_handler(request: UNotifyEventRequest, args: &[u64; 6], op: u8
         #[allow(clippy::cast_possible_truncation)]
         let fd = request.get_fd(args[0] as RawFd)?;
 
-        // SAFETY: Check if fd points to a socket, return ENOTSOCK otherwise.
-        // TODO: Return EBADF for O_PATH file descriptors:
-        //       LTP test accept03:
-        //       TFAIL: accept() on O_PATH file expected EBADF: ENOTSOCK (88)
-        if file_type(&fd, None, false)? != FileType::Sock {
+        // SAFETY:
+        // 1. Check if fd has O_PATH in status flags and return EBADF.
+        // 2. Check if fd points to a socket or return ENOTSOCK.
+        if fd_status_flags(&fd)?.contains(OFlag::O_PATH) {
+            return Err(Errno::EBADF);
+        } else if file_type(&fd, None, false)? != FileType::Sock {
             return Err(Errno::ENOTSOCK);
         }
 
@@ -375,8 +374,8 @@ fn syscall_network_handler(request: UNotifyEventRequest, args: &[u64; 6], op: u8
 
         // Emulate syscall.
         match op {
-            0x2 => handle_bind(fd, &addr, root.as_ref(), allow_safe_bind, &request),
-            0x3 => handle_connect(fd, &addr, &request),
+            0x2 => handle_bind(fd, &addr, root.as_ref(), &request, allow_safe_bind),
+            0x3 => handle_connect(fd, &addr, &request, allow_safe_bind),
             0xb => handle_sendto(fd, args, &request, Some(&addr)),
             _ => unreachable!(),
         }
@@ -456,12 +455,12 @@ fn canon_addr<'a>(
 
         let cstr = if path.base.is_empty() {
             // SAFETY: We open a FD to the path and then use the
-            // proc path /proc/self/fd/$fd in address' path argument
-            // to avoid symlink TOCTOU because connect and sendto
-            // follow symlinks in basename unlike bind.
+            // proc path /proc/thread-self/fd/$fd in address' path
+            // argument to avoid symlink TOCTOU because connect and
+            // sendto follow symlinks in basename unlike bind.
             #[allow(clippy::disallowed_methods)]
             let fd = path.dir.as_ref().unwrap();
-            let mut pfd = XPathBuf::from("/proc/self/fd");
+            let mut pfd = XPathBuf::from("/proc/thread-self/fd");
             pfd.push_fd(fd.as_raw_fd());
             CString::new(pfd.as_os_str().as_bytes())
         } else {
@@ -788,3 +787,52 @@ pub(crate) fn to_msgflags(arg: u64) -> Result<MsgFlags, Errno> {
     let flags = arg.try_into().or(Err(Errno::EINVAL))?;
     MsgFlags::from_bits(flags).ok_or(Errno::EINVAL)
 }
+
+// Handle allow_safe_bind for connect and sendto.
+// bind is handled privately in handle_bind.
+pub(crate) fn handle_safe_bind<Fd: AsFd>(
+    request: &UNotifyEventRequest,
+    fd: Fd,
+) -> Result<(), Errno> {
+    let addr = getsockname::<SockaddrStorage>(fd.as_fd().as_raw_fd())?;
+
+    let addr = if let Some(addr) = addr.as_sockaddr_in() {
+        let port = addr.port();
+        if port == 0 {
+            return Ok(());
+        }
+
+        // Allow implicit bind with safe_bind.
+        let addr = IpAddr::V4(addr.ip());
+        format!("{addr}!{port}")
+    } else if let Some(addr) = addr.as_sockaddr_in6() {
+        let port = addr.port();
+        if port == 0 {
+            return Ok(());
+        }
+
+        // Allow implicit bind with safe_bind.
+        let addr = IpAddr::V6(addr.ip());
+        format!("{addr}!{port}")
+    } else {
+        return Ok(());
+    };
+
+    // Configure sandbox, note we remove
+    // and readd the address so repeated
+    // binds to the same address cannot
+    // overflow the vector.
+    let config: &[String] = &[
+        format!("allow/net/connect-{addr}"),
+        format!("allow/net/connect+{addr}"),
+    ];
+
+    // TODO: Log errors!
+    let mut sandbox = request.get_mut_sandbox();
+    for cmd in config {
+        sandbox.config(cmd)?;
+    }
+    drop(sandbox);
+
+    Ok(())
+}
