@@ -7,7 +7,7 @@
 
 use std::{
     io::IoSlice,
-    os::fd::{AsRawFd, BorrowedFd, OwnedFd, RawFd},
+    os::fd::{AsRawFd, OwnedFd, RawFd},
 };
 
 use libseccomp::ScmpNotifResp;
@@ -24,7 +24,7 @@ use crate::{
         PF_NETLINK, PF_UNIX,
     },
     fs::{file_type, get_nonblock, has_recv_timeout, is_same_vm, FileType},
-    hook::{RemoteProcess, UNotifyEventRequest},
+    hook::UNotifyEventRequest,
     kernel::{
         net::{canon_addr, get_addr, sandbox_addr, to_msgflags},
         sandbox_path,
@@ -47,21 +47,16 @@ pub(crate) fn handle_sendmsg(
 
     let req = request.scmpreq;
     let is32 = scmp_arch_bits(req.data.arch) == 32;
-    let mut buf = if is32 {
-        // SAFETY: socketcall is the same number on all:
+    let size = if is32 {
+        // Note, socketcall is the same number on all:
         // x86, mips, mipsel, ppc, ppc64, ppc64le, s390 and s390x.
-        const SIZ: usize = std::mem::size_of::<msghdr32>();
-        let mut buf = Vec::new();
-        buf.try_reserve(SIZ).or(Err(Errno::ENOMEM))?;
-        buf.resize(SIZ, 0);
-        buf
+        std::mem::size_of::<msghdr32>()
     } else {
-        const SIZ: usize = std::mem::size_of::<msghdr>();
-        let mut buf = Vec::new();
-        buf.try_reserve(SIZ).or(Err(Errno::ENOMEM))?;
-        buf.resize(SIZ, 0);
-        buf
+        std::mem::size_of::<msghdr>()
     };
+    let mut buf = Vec::new();
+    buf.try_reserve(size).or(Err(Errno::ENOMEM))?;
+    buf.resize(size, 0);
     request.read_mem(&mut buf, args[1])?;
 
     let msg = if is32 {
@@ -142,8 +137,7 @@ pub(crate) fn handle_sendmsg(
     };
 
     if let Some((
-        ref _control_fds, // Keep the OwnedFd alive!
-        ref control_raw_fds,
+        ref control_fds,
         ref control_creds,
         ref control_ivs,
         ref control_ops,
@@ -156,7 +150,7 @@ pub(crate) fn handle_sendmsg(
     )) = control_data
     {
         // Check for sendfd access as necessary.
-        if !control_raw_fds.is_empty() {
+        if !control_fds.is_empty() {
             if let Some((ref addr, ref root)) = addr {
                 sandbox_addr(
                     request,
@@ -167,16 +161,13 @@ pub(crate) fn handle_sendmsg(
                     Capability::CAP_NET_SENDFD,
                 )?;
             } else {
-                // Unused when request.is_some()
-                let process = RemoteProcess::new(request.scmpreq.pid());
-
                 // SAFETY: For cases where address is not available, we
                 // perform an access check with a dummy path so as to
                 // enable user to practically confine this case.
                 sandbox_path(
                     Some(request),
                     &sandbox,
-                    &process,
+                    request.scmpreq.pid(), // Unused when request.is_some()
                     XPath::from_bytes(b"!unnamed"),
                     Capability::CAP_NET_SENDFD,
                     false,
@@ -185,7 +176,7 @@ pub(crate) fn handle_sendmsg(
             }
         }
 
-        for raw_fds in control_raw_fds {
+        for fds in control_fds {
             // SAFETY: Deny sending file descriptors referring to
             // 1. Block devices
             // 2. Directories
@@ -195,11 +186,7 @@ pub(crate) fn handle_sendmsg(
             // fds and event fds as some programs such as pipewire
             // depend on this. See test-pw-filter test of pipewire
             // for more information about this.
-            for raw_fd in raw_fds {
-                // SAFETY: raw_fd is a valid FD
-                // for the lifetime of this loop block.
-                let fd = unsafe { BorrowedFd::borrow_raw(*raw_fd) };
-
+            for fd in fds {
                 if matches!(
                     file_type(fd, None, false)?,
                     FileType::Blk | FileType::Dir | FileType::Lnk
@@ -209,6 +196,11 @@ pub(crate) fn handle_sendmsg(
                 }
             }
 
+            // SAFETY: OwnedFd is repr(transparent) over RawFd, so
+            // the pointer cast & the slice length are correct.
+            let raw_fds: &[RawFd] =
+                unsafe { std::slice::from_raw_parts(fds.as_ptr() as *const RawFd, fds.len()) };
+
             control_messages.push(ControlMessage::ScmRights(raw_fds));
         }
 
@@ -359,8 +351,10 @@ pub(crate) fn handle_sendmsg(
     drop(fd); // Close our copy of the socket.
 
     // Remove invalidation record unless interrupted.
-    if is_blocking && !matches!(result, Err(Errno::EINTR)) {
-        request.cache.del_sys_block(req.id);
+    if is_blocking {
+        request
+            .cache
+            .del_sys_block(req.id, matches!(result, Err(Errno::EINTR)))?;
     }
 
     result
@@ -493,8 +487,7 @@ pub(crate) fn handle_sendmmsg(
         let mut control_messages = Vec::new();
         for control_datum in &control_data {
             if let Some((
-                ref _control_fds, // Keep the OwnedFd alive!
-                ref control_raw_fds,
+                ref control_fds,
                 ref control_creds,
                 ref control_ivs,
                 ref control_ops,
@@ -507,7 +500,7 @@ pub(crate) fn handle_sendmmsg(
             )) = &control_datum
             {
                 // Check for sendfd access as necessary.
-                if !control_raw_fds.is_empty() {
+                if !control_fds.is_empty() {
                     if let Some((ref addr, ref root)) = addr_root {
                         sandbox_addr(
                             request,
@@ -518,16 +511,13 @@ pub(crate) fn handle_sendmmsg(
                             Capability::CAP_NET_SENDFD,
                         )?;
                     } else {
-                        // Unused when request.is_some()
-                        let process = RemoteProcess::new(request.scmpreq.pid());
-
                         // SAFETY: For cases where address is not available, we
                         // perform an access check with a dummy path so as to
                         // enable user to practically confine this case.
                         sandbox_path(
                             Some(request),
                             &sandbox,
-                            &process,
+                            request.scmpreq.pid(), // Unused when request.is_some()
                             XPath::from_bytes(b"!unnamed"),
                             Capability::CAP_NET_SENDFD,
                             false,
@@ -536,7 +526,7 @@ pub(crate) fn handle_sendmmsg(
                     }
                 }
 
-                for raw_fds in control_raw_fds {
+                for fds in control_fds {
                     // SAFETY: Deny sending file descriptors referring to
                     // 1. Block devices
                     // 2. Directories
@@ -546,11 +536,7 @@ pub(crate) fn handle_sendmmsg(
                     // fds and event fds as some programs such as pipewire
                     // depend on this. See test-pw-filter test of pipewire
                     // for more information about this.
-                    for raw_fd in raw_fds {
-                        // SAFETY: raw_fd is a valid FD
-                        // for the lifetime of this loop block.
-                        let fd = unsafe { BorrowedFd::borrow_raw(*raw_fd) };
-
+                    for fd in fds {
                         if matches!(
                             file_type(fd, None, false)?,
                             FileType::Blk | FileType::Dir | FileType::Lnk
@@ -560,6 +546,12 @@ pub(crate) fn handle_sendmmsg(
                         }
                     }
 
+                    // SAFETY: OwnedFd is repr(transparent) over RawFd, so
+                    // the pointer cast & the slice length are correct.
+                    let raw_fds: &[RawFd] = unsafe {
+                        std::slice::from_raw_parts(fds.as_ptr() as *const RawFd, fds.len())
+                    };
+
                     control_messages.push(ControlMessage::ScmRights(raw_fds));
                 }
 
@@ -693,8 +685,10 @@ pub(crate) fn handle_sendmmsg(
             };
 
             // Remove invalidation record unless interrupted.
-            if is_blocking && !matches!(result, Err(Errno::EINTR)) {
-                request.cache.del_sys_block(req.id);
+            if is_blocking {
+                request
+                    .cache
+                    .del_sys_block(req.id, matches!(result, Err(Errno::EINTR)))?;
             }
 
             mmsg.msg_len = result?.try_into().or(Err(Errno::EINVAL))?;
@@ -738,7 +732,6 @@ fn parse_control_messages(
 ) -> Result<
     (
         Vec<Vec<OwnedFd>>,
-        Vec<Vec<RawFd>>,
         Vec<UnixCredentials>,
         Vec<SecureBytes>, // ivs
         Vec<libc::c_int>, // ops
@@ -752,7 +745,6 @@ fn parse_control_messages(
     Errno,
 > {
     let mut control_fds = Vec::new();
-    let mut control_raw_fds = Vec::new();
     let mut control_creds = Vec::new();
     let mut control_ivs = Vec::new();
     let mut control_ops = Vec::new();
@@ -816,9 +808,7 @@ fn parse_control_messages(
                     .ok_or(Errno::EINVAL)?;
 
                 let mut fds = Vec::new();
-                let mut raw_fds = Vec::new();
                 fds.try_reserve(fd_count).or(Err(Errno::ENOMEM))?;
-                raw_fds.try_reserve(fd_count).or(Err(Errno::ENOMEM))?;
 
                 // SAFETY: Multiplying `i` by `std::mem::size_of::<RawFd>()` calculates the
                 // offset for each file descriptor in the control message data. This is safe
@@ -838,10 +828,8 @@ fn parse_control_messages(
                     // SAFETY: See the comment above.
                     let fd = unsafe { *(data[fd_offset..].as_ptr() as *const RawFd) };
                     let fd = request.get_fd(fd)?;
-                    raw_fds.push(fd.as_raw_fd());
-                    fds.push(fd); // Keep a ref to the OwnedFd.
+                    fds.push(fd);
                 }
-                control_raw_fds.push(raw_fds);
                 control_fds.push(fds);
             }
             (libc::SOL_SOCKET, libc::SCM_CREDENTIALS) => {
@@ -1063,7 +1051,6 @@ fn parse_control_messages(
 
     Ok((
         control_fds,
-        control_raw_fds,
         control_creds,
         control_ivs,
         control_ops,
