@@ -63,8 +63,10 @@ pub(crate) fn handle_sendto(
     .map(|n| request.return_syscall(n as i64));
 
     // Remove invalidation record unless interrupted.
-    if is_blocking && !matches!(result, Err(Errno::EINTR)) {
-        request.cache.del_sys_block(req.id);
+    if is_blocking {
+        request
+            .cache
+            .del_sys_block(req.id, matches!(result, Err(Errno::EINTR)))?;
     }
 
     result
