@@ -18,11 +18,13 @@ use nix::{
     errno::Errno,
     fcntl::{OFlag, OpenHow, ResolveFlag},
     sys::stat::{umask, Mode},
+    unistd::Uid,
     NixPath,
 };
 
 use crate::{
-    config::{AES_CYCLE_TIME, ESYD_SH, MAGIC_PREFIX, NULL_FD, PROC_FILE, ROOT_FILE, SYD_EL},
+    compat::{fstatx, STATX_MODE, STATX_UID},
+    config::{ESYD_SH, MAGIC_PREFIX, NULL_FD, PROC_FILE, ROOT_FILE, SYD_EL},
     cookie::safe_openat2,
     fs::{
         create_memfd, lock_fd, safe_open_how_magicsym, safe_open_magicsym, seal_memfd, unlock_fd,
@@ -64,12 +66,10 @@ pub(crate) fn sys_creat(request: UNotifyEventRequest) -> ScmpNotifResp {
         None => return request.fail_syscall(Errno::EINVAL),
     };
 
-    // SAFETY:
-    // 1. Omit MUST_PATH in fsflags as path may not exist yet.
-    // 2. Use WANT_BASE to split base which will be opened by the handler.
+    // Here we omit MUST_PATH in fsflags as path may not exist yet.
     let arg = SysArg {
         path: Some(0),
-        fsflags: FsFlags::WANT_BASE,
+        fsflags: FsFlags::empty(),
         ..Default::default()
     };
     let flags = OFlag::O_CREAT | OFlag::O_WRONLY | OFlag::O_TRUNC;
@@ -326,7 +326,6 @@ fn syscall_open_handler(
         let (is_dir, is_blocking, may_crypt) = if let Some(ref file_type) = path.typ {
             if o_creat && !o_tmpfl {
                 // Creating open with existing file:
-                //
                 // Remove Create from capabilities, and add back
                 // Truncate capability as necessary to improve
                 // usability.
@@ -395,6 +394,12 @@ fn syscall_open_handler(
                         false,
                     )
                 }
+                FileType::Mfd => {
+                    // SAFETY:
+                    // 1. We do not encrypt memory fds.
+                    // 2. We do not support interruptions on memory fd blocks.
+                    (false, false, false)
+                }
             }
         } else if !o_creat {
             // Non-creating open on non-existing file.
@@ -431,19 +436,19 @@ fn syscall_open_handler(
 
             true // TTY.
         } else {
-            // SAFETY: Ensure the sandbox process cannot
-            // acquire a new controlling terminal other
-            // than what they already have.
-            // Note, O_NOCTTY is invalid with O_PATH,
-            // but we don't care for now since we'll change it
-            // to O_RDONLY later as a mitigation due to our
-            // inability to emulate them.
-            // See: https://bugzilla.kernel.org/show_bug.cgi?id=218501
-            flags.insert(OFlag::O_NOCTTY);
-
             false // Not a TTY.
         };
 
+        // SAFETY: Ensure the sandbox process cannot
+        // acquire a new controlling terminal other
+        // than what they already have.
+        // Note, O_NOCTTY is invalid with O_PATH,
+        // but we don't care for now since we'll change it
+        // to O_RDONLY later as a mitigation due to our
+        // inability to emulate them.
+        // See: https://bugzilla.kernel.org/show_bug.cgi?id=218501
+        flags.insert(OFlag::O_NOCTTY);
+
         // Sandboxing.
         let orig_caps = caps;
         let caps = sandbox.getcaps(orig_caps); // Get enabled caps.
@@ -567,17 +572,14 @@ fn syscall_open_handler(
         let verbose = sandbox.verbose;
         let unsafe_open_path = sandbox.allow_unsafe_open_path();
         let unsafe_open_cdev = sandbox.allow_unsafe_open_cdev();
+        let restrict_create = !sandbox.allow_unsafe_create();
         let restrict_memfd = !sandbox.allow_unsafe_memfd();
         if sandbox.force_cloexec() {
             o_cloexec = true;
         }
         let o_rand_fd = sandbox.force_rand_fd();
         let setup_fds = sandbox.crypt_setup();
-        let crypt_tmp = if crypted {
-            sandbox.crypt_tmp.as_ref().map(|fd| fd.as_raw_fd())
-        } else {
-            None
-        };
+        let crypt_tmp = if crypted { sandbox.crypt_tmp } else { None };
         drop(sandbox); // release the read-lock.
 
         // Perform action: allow->emulate, deny->log.
@@ -658,6 +660,39 @@ fn syscall_open_handler(
             }
         }
 
+        // SAFETY: Restrictions a la protected_{fifos,regular}
+        // and CONFIG_GRKERNSEC_FIFO. Here, though, we take
+        // a stricter approach and block creates on non-owned
+        // files or fifos regardless of the owned directory.
+        //
+        // This is done after O_CREAT|O_EXCL -> EEXIST check.
+        if restrict_create && o_creat && matches!(path.typ, Some(FileType::Reg | FileType::Fifo)) {
+            let fd = path.dir.as_ref().map(|fd| fd.as_fd()).ok_or(Errno::EBADF)?;
+            // assert fd refers to file.
+            assert!(path.base.is_empty());
+
+            let statx = fstatx(fd, STATX_MODE | STATX_UID)?;
+            let mode = libc::mode_t::from(statx.stx_mode);
+
+            // Return EACCES if:
+            // 1. File is group or world-writable.
+            // 2. File is not owned by caller.
+            if mode & libc::S_IWOTH != 0 {
+                // File is world-writable.
+                return Err(Errno::EACCES);
+            }
+
+            if mode & libc::S_IWGRP != 0 {
+                // File is group-writable.
+                return Err(Errno::EACCES);
+            }
+
+            if statx.stx_uid != Uid::effective().as_raw() {
+                // File is not owned by caller.
+                return Err(Errno::EACCES);
+            }
+        }
+
         match action {
             Action::Allow | Action::Warn => {
                 // The system call is allowed.
@@ -797,11 +832,11 @@ fn syscall_open_handler(
                     // reopen, because the fd we're going to send will
                     // share the same open file description and we do
                     // not want to mess up file offsets for everyone.
-                    let mut pfd = XPathBuf::from("self/fd");
-
+                    //
                     // SAFETY: ^^ empty base asserts dir is Some.
                     #[allow(clippy::disallowed_methods)]
-                    pfd.push_fd(path.dir.as_ref().map(|fd| fd.as_raw_fd()).unwrap());
+                    let fd = path.dir.as_ref().map(|fd| fd.as_raw_fd()).unwrap();
+                    let pfd = XPathBuf::from_self_fd(fd);
 
                     // Reopen the `O_PATH` path fd with the requested flags.
                     flags.remove(OFlag::O_NOFOLLOW);
@@ -815,8 +850,10 @@ fn syscall_open_handler(
                     let result = safe_openat2(PROC_FILE(), &pfd, how);
 
                     // Remove invalidation record unless interrupted.
-                    if is_blocking && !matches!(result, Err(Errno::EINTR)) {
-                        request.cache.del_sys_block(req.id);
+                    if is_blocking {
+                        request
+                            .cache
+                            .del_sys_block(req.id, matches!(result, Err(Errno::EINTR)))?;
                     }
 
                     result?
@@ -843,8 +880,10 @@ fn syscall_open_handler(
                     let result = safe_openat2(fd, path.base, how);
 
                     // Remove invalidation record unless interrupted.
-                    if is_blocking && !matches!(result, Err(Errno::EINTR)) {
-                        request.cache.del_sys_block(req.id);
+                    if is_blocking {
+                        request
+                            .cache
+                            .del_sys_block(req.id, matches!(result, Err(Errno::EINTR)))?;
                     }
 
                     result?
@@ -870,11 +909,11 @@ fn syscall_open_handler(
 fn do_open(path: &CanonicalPath, mut safe_flags: OFlag) -> Result<OwnedFd, Errno> {
     if path.base.is_empty() {
         // Existing path: Construct path to proc magic symlink.
-        let mut pfd = XPathBuf::from("self/fd");
-
+        //
         // SAFETY: ^^ empty base asserts dir is Some.
         #[allow(clippy::disallowed_methods)]
-        pfd.push_fd(path.dir.as_ref().map(|fd| fd.as_raw_fd()).unwrap());
+        let fd = path.dir.as_ref().map(|fd| fd.as_raw_fd()).unwrap();
+        let pfd = XPathBuf::from_self_fd(fd);
 
         // Reopen the `O_PATH` path fd with the requested flags.
         safe_flags.remove(OFlag::O_NOFOLLOW);
@@ -920,45 +959,49 @@ fn handle_crypt(
     o_rand_fd: bool,
     restrict_memfd: bool,
 ) -> Result<Option<OwnedFd>, Errno> {
-    let my_mode = AesMod::from(safe_flags);
     #[allow(clippy::disallowed_methods)]
     let files = request.crypt_map.as_ref().unwrap();
-    let maybe = {
-        files
-            .read()
-            .unwrap_or_else(|err| err.into_inner())
-            .0
+    let my_mode = AesMod::from(safe_flags);
+
+    // Check for concurrent read access to the same crypt fd.
+    if my_mode == AesMod::Read {
+        // Lock AesMap for read:
+        // We keep the read lock until we open the file,
+        // to ensure no races occur with a subsequent Syd thread.
+        let files = files.read().unwrap_or_else(|err| err.into_inner());
+
+        // Concurrent access is only allowed if both modes are Read.
+        let entry = files
             .get(path.abs())
-            .map(|v| (v.crypt_fd, v.mode))
-    };
-    if let Some((fd, file_mode)) = maybe {
-        // Open a new file description.
-        // Lock it for read and pass to sandbox process.
-        let mut flags = safe_flags;
-        flags.remove(OFlag::O_NOFOLLOW);
-        if flags.contains(OFlag::O_WRONLY) {
-            // Promote O_WRONLY to O_RDWR,
-            // so that we can hold an OFD read-lock.
-            flags.remove(OFlag::O_WRONLY);
-            flags.insert(OFlag::O_RDWR);
-        }
+            .filter(|&entry| entry.mode == AesMod::Read);
 
-        let mut pfd = XPathBuf::from("self/fd");
-        pfd.push_fd(fd.as_raw_fd());
+        if let Some(entry) = entry {
+            // Open a new file description.
+            // Lock it for read and pass to sandbox process.
+            let mut flags = safe_flags;
+            flags.remove(OFlag::O_NOFOLLOW);
 
-        if let Ok(fd) = safe_open_magicsym(PROC_FILE(), &pfd, flags) {
-            // SAFETY: If our attempt to lock fails,
-            // aes thread has already taken over!
-            if lock_fd(&fd, false, false).is_ok() {
-                if my_mode > file_mode {
-                    // File mode upgraded, update AesMap.
-                    let mut files = files.write().unwrap_or_else(|err| err.into_inner());
-                    if let Some(entry) = files.0.get_mut(path.abs()) {
-                        entry.mode = my_mode;
-                    }
+            let pfd = XPathBuf::from_self_fd(entry.crypt_fd);
+            match safe_open_magicsym(PROC_FILE(), &pfd, flags) {
+                Ok(fd) if lock_fd(&fd, false, false).is_ok() => {
+                    // Return shared fd.
+                    return Ok(Some(fd));
                 }
-                return Ok(Some(fd));
-            }
+                Ok(fd) => {
+                    // SAFETY: If our attempt to lock fails,
+                    // syd_aes thread has already taken over.
+                    // We can not use the crypt fd any longer!
+                    drop(fd);
+                    /* fall through */
+                }
+                Err(Errno::ENOENT) => {
+                    // SAFETY: If our attempt to open via procfs fails,
+                    // syd_aes thread had already closed the crypt fd.
+                    // We can not use the crypt fd any longer!
+                    /* fall through */
+                }
+                Err(errno) => return Err(errno),
+            };
         }
     }
 
@@ -985,35 +1028,16 @@ fn handle_crypt(
     }?;
 
     // Hold a write-lock to ensure no concurrent Syd aes writes.
+    //
     // SAFETY: Block until we can acquire the lock to ensure safe
     // concurrent access _unless_ the open was a non-blocking open
     // in which case we happily return EAGAIN so the caller can
-    // retry. Similarly, `aes_ctr_tmp` respects non-blocking
-    // opens too and returns a non-blocking fd as necessary.
+    // retry.
+    //
+    // Similarly, `aes_ctr_tmp` respects non-blocking opens too,
+    // and returns a non-blocking fd as necessary.
     let wait = !(safe_flags.contains(OFlag::O_NONBLOCK) || safe_flags.contains(OFlag::O_NDELAY));
-    match lock_fd(&enc_fd, true, create || wait) {
-        Ok(()) => {}
-        Err(Errno::EAGAIN) if !wait => return Err(Errno::EAGAIN),
-        Err(Errno::EAGAIN) => {
-            // Wait a bit and try again to let the
-            // writer thread finish so that we can
-            // hopefully join in with the shared-fd.
-            std::thread::sleep(AES_CYCLE_TIME);
-            return handle_crypt(
-                setup_fds,
-                request,
-                path,
-                tmpdir,
-                safe_flags,
-                mode,
-                force_umask,
-                o_cloexec,
-                o_rand_fd,
-                restrict_memfd,
-            );
-        }
-        Err(errno) => return Err(errno),
-    };
+    lock_fd(&enc_fd, true, create || wait)?;
 
     // Acquire encryption sockets.
     let setup_fds = setup_fds?;
@@ -1039,8 +1063,7 @@ fn handle_crypt(
 
     // Reopen a new instance to pass to the sandbox process.
     // This instance points to a different open file description!
-    let mut pfd = XPathBuf::from("self/fd");
-    pfd.push_fd(fd.as_raw_fd());
+    let pfd = XPathBuf::from_self_fd(fd.as_raw_fd());
 
     let mut flags = safe_flags;
     flags.remove(
@@ -1062,21 +1085,17 @@ fn handle_crypt(
 
     // Record encryption process information.
     {
-        files
-            .write()
-            .unwrap_or_else(|err| err.into_inner())
-            .0
-            .insert(
-                path.abs().to_owned(),
-                AesVal {
-                    enc_fd,
-                    info,
-                    iv,
-                    crypt_fd: fd.into_raw_fd(),
-                    mode: my_mode,
-                    spawned: false,
-                },
-            );
+        files.write().unwrap_or_else(|err| err.into_inner()).insert(
+            path.abs().to_owned(),
+            AesVal {
+                info,
+                iv: Some(iv),
+                enc_fd: Some(enc_fd),
+                crypt_fd: fd.into_raw_fd(),
+                mode: my_mode,
+                spawned: false,
+            },
+        );
     }
 
     Ok(None)
