@@ -23,7 +23,7 @@ use crate::{
     compat::{fstatat64, statx, STATX_BASIC_STATS, STATX_MODE, STATX_TYPE},
     config::{MAGIC_PREFIX, MMAP_MIN_ADDR},
     fs::{is_sidechannel_device, parse_fd, CanonicalPath, FileInfo, FileType, FsFlags},
-    hook::{RemoteProcess, SysArg, SysFlags, UNotifyEventRequest},
+    hook::{SysArg, SysFlags, UNotifyEventRequest},
     kernel::sandbox_path,
     path::XPath,
     sandbox::Capability,
@@ -168,6 +168,7 @@ fn syscall_stat_handler(
         // We'll either upgrade it or downgrade it based on magic lock.
         // Exception: Sandbox lock had been set and there's no turning back.
         let req = request.scmpreq;
+        let is_fd = arg.path.is_none();
         let sandbox = request.get_sandbox();
         let is_lock = sandbox.locked_for(req.pid());
 
@@ -176,7 +177,7 @@ fn syscall_stat_handler(
 
         // Check for chroot.
         if sandbox.is_chroot() {
-            return Err(Errno::ENOENT);
+            return Err(if is_fd { Errno::EACCES } else { Errno::ENOENT });
         }
 
         // Read the remote path.
@@ -268,14 +269,16 @@ fn syscall_stat_handler(
         } else {
             // Handle fstat for files with encryption in progress.
             let mut crypt_stat = false;
-            if is_crypt && arg.path.is_none() {
+            if is_crypt && is_fd {
                 // SAFETY: SysArg.path is None asserting dirfd is Some fd!=AT_FDCWD.
                 #[allow(clippy::disallowed_methods)]
                 let fd = path.dir.as_ref().unwrap();
                 if let Ok(info) = FileInfo::from_fd(fd) {
                     #[allow(clippy::disallowed_methods)]
                     let files = request.crypt_map.as_ref().unwrap();
-                    for (enc_path, map) in &files.read().unwrap_or_else(|err| err.into_inner()).0 {
+                    for (enc_path, map) in
+                        files.read().unwrap_or_else(|err| err.into_inner()).iter()
+                    {
                         if info == map.info {
                             // Found underlying encrypted file for the memory fd.
                             // Note, we only ever attempt to encrypt regular files.
@@ -299,13 +302,10 @@ fn syscall_stat_handler(
                 && arg.path.is_some()
                 && !path.abs().starts_with(b"/memfd:syd-")
             {
-                // Unused when request.is_some()
-                let process = RemoteProcess::new(request.scmpreq.pid());
-
                 sandbox_path(
                     Some(&request),
                     &sandbox,
-                    &process,
+                    request.scmpreq.pid(), // Unused when request.is_some()
                     path.abs(),
                     Capability::CAP_STAT,
                     false,
