@@ -6,14 +6,15 @@
 //
 // SPDX-License-Identifier: GPL-3.0
 
-use std::os::fd::{AsRawFd, RawFd};
+use std::os::fd::AsRawFd;
 
 use libseccomp::{ScmpArch, ScmpNotifResp};
-use nix::{errno::Errno, fcntl::FallocateFlags, NixPath};
+use nix::{errno::Errno, fcntl::FallocateFlags};
 
 use crate::{
-    compat::{ftruncate64, truncate64},
-    hook::{PathArgs, SysArg, UNotifyEventRequest},
+    cookie::{safe_ftruncate, safe_ftruncate64, safe_truncate, safe_truncate64},
+    fs::is_valid_fd,
+    hook::{SysArg, UNotifyEventRequest},
     kernel::syscall_path_handler,
     path::XPathBuf,
     scmp_arch_bits, scmp_big_endian,
@@ -46,47 +47,38 @@ pub(crate) fn sys_truncate(request: UNotifyEventRequest) -> ScmpNotifResp {
         ..Default::default()
     }];
 
-    syscall_path_handler(
-        request,
-        "truncate",
-        argv,
-        |path_args: PathArgs, request, sandbox| {
-            drop(sandbox); // release the read-lock.
+    syscall_path_handler(request, "truncate", argv, |path_args, request, sandbox| {
+        drop(sandbox); // release the read-lock.
 
-            // SAFETY:
-            // 1. SysArg has one element.
-            // 2. `/` is not a regular file -> EINVAL.
-            #[allow(clippy::disallowed_methods)]
-            let fd = path_args
-                .0
-                .as_ref()
-                .map(|path| path.dir.as_ref().unwrap().as_raw_fd())
-                .ok_or(Errno::EINVAL)?;
+        // SAFETY:
+        // 1. SysArg has one element.
+        // 2. `/` is not a regular file -> EINVAL.
+        #[allow(clippy::disallowed_methods)]
+        let fd = path_args
+            .0
+            .as_ref()
+            .map(|path| path.dir.as_ref().unwrap().as_raw_fd())
+            .ok_or(Errno::EINVAL)?;
 
-            // SAFETY: ftruncate(2) requires write fds.
-            // We fallback to `/proc` indirection to avoid opening the file as write.
-            //
-            // path to fd is open already!
-            let mut pfd = XPathBuf::from("/proc/self/fd");
-            pfd.push_fd(fd);
+        // SAFETY: ftruncate(2) requires write fds.
+        // We fallback to `/proc` indirection to avoid opening the file as write.
+        //
+        // path to fd is open already!
+        let mut pfd = XPathBuf::from("/proc/thread-self/fd");
+        pfd.push_fd(fd);
 
-            // SAFETY: Record blocking call so it can get invalidated.
-            request.cache.add_sys_block(req, false)?;
+        // SAFETY: Record blocking call so it can get invalidated.
+        request.cache.add_sys_block(req, false)?;
 
-            let result = pfd
-                .with_nix_path(|cstr| {
-                    // SAFETY: libc version may call truncate64 behind our back!
-                    Errno::result(unsafe { libc::syscall(libc::SYS_truncate, cstr.as_ptr(), len) })
-                })?
-                .map(|_| request.return_syscall(0));
+        let result = safe_truncate(&pfd, len).map(|_| request.return_syscall(0));
 
-            if !matches!(result, Err(Errno::EINTR)) {
-                request.cache.del_sys_block(req.id);
-            }
+        // Remove invalidation record unless interrupted.
+        request
+            .cache
+            .del_sys_block(req.id, matches!(result, Err(Errno::EINTR)))?;
 
-            result
-        },
-    )
+        result
+    })
 }
 
 pub(crate) fn sys_truncate64(request: UNotifyEventRequest) -> ScmpNotifResp {
@@ -140,7 +132,7 @@ pub(crate) fn sys_truncate64(request: UNotifyEventRequest) -> ScmpNotifResp {
         request,
         "truncate64",
         argv,
-        |path_args: PathArgs, request, sandbox| {
+        |path_args, request, sandbox| {
             drop(sandbox); // release the read-lock.
 
             // SAFETY:
@@ -159,17 +151,18 @@ pub(crate) fn sys_truncate64(request: UNotifyEventRequest) -> ScmpNotifResp {
             // We fallback to `/proc` indirection to avoid opening the file as write.
             //
             // path to fd is open already!
-            let mut pfd = XPathBuf::from("/proc/self/fd");
+            let mut pfd = XPathBuf::from("/proc/thread-self/fd");
             pfd.push_fd(fd.as_raw_fd());
 
             // SAFETY: Record blocking call so it can get invalidated.
             request.cache.add_sys_block(req, false)?;
 
-            let result = truncate64(&pfd, len).map(|_| request.return_syscall(0));
+            let result = safe_truncate64(&pfd, len).map(|_| request.return_syscall(0));
 
-            if !matches!(result, Err(Errno::EINTR)) {
-                request.cache.del_sys_block(req.id);
-            }
+            // Remove invalidation record unless interrupted.
+            request
+                .cache
+                .del_sys_block(req.id, matches!(result, Err(Errno::EINTR)))?;
 
             result
         },
@@ -182,7 +175,7 @@ pub(crate) fn sys_ftruncate(request: UNotifyEventRequest) -> ScmpNotifResp {
     // SAFETY: FD-only call:
     // 1. Assert valid fd before other arguments.
     // 2. AT_FDCWD is an invalid fd argument.
-    if RawFd::try_from(req.data.args[0]).is_err() {
+    if !is_valid_fd(req.data.args[0]) {
         return request.fail_syscall(Errno::EBADF);
     }
 
@@ -207,37 +200,31 @@ pub(crate) fn sys_ftruncate(request: UNotifyEventRequest) -> ScmpNotifResp {
         ..Default::default()
     }];
 
-    syscall_path_handler(
-        request,
-        "ftruncate",
-        argv,
-        |path_args: PathArgs, request, sandbox| {
-            drop(sandbox); // release the read-lock.
+    syscall_path_handler(request, "ftruncate", argv, |path_args, request, sandbox| {
+        drop(sandbox); // release the read-lock.
 
-            // SAFETY:
-            // 1. SysArg has one element.
-            // 2. SysArg.path is None asserting dir is Some.
-            #[allow(clippy::disallowed_methods)]
-            let fd = path_args
-                .0
-                .as_ref()
-                .map(|path| path.dir.as_ref().unwrap().as_raw_fd())
-                .ok_or(Errno::EINVAL)?;
+        // SAFETY:
+        // 1. SysArg has one element.
+        // 2. SysArg.path is None asserting dir is Some.
+        #[allow(clippy::disallowed_methods)]
+        let fd = path_args
+            .0
+            .as_ref()
+            .map(|path| path.dir.as_ref().unwrap())
+            .ok_or(Errno::EINVAL)?;
 
-            // SAFETY: Record blocking call so it can get invalidated.
-            request.cache.add_sys_block(req, false)?;
+        // SAFETY: Record blocking call so it can get invalidated.
+        request.cache.add_sys_block(req, false)?;
 
-            // SAFETY: Libc version may call ftruncate64 behind our back.
-            let result = Errno::result(unsafe { libc::syscall(libc::SYS_ftruncate, fd, len) })
-                .map(|_| request.return_syscall(0));
+        let result = safe_ftruncate(fd, len).map(|_| request.return_syscall(0));
 
-            if !matches!(result, Err(Errno::EINTR)) {
-                request.cache.del_sys_block(req.id);
-            }
+        // Remove invalidation record unless interrupted.
+        request
+            .cache
+            .del_sys_block(req.id, matches!(result, Err(Errno::EINTR)))?;
 
-            result
-        },
-    )
+        result
+    })
 }
 
 pub(crate) fn sys_ftruncate64(request: UNotifyEventRequest) -> ScmpNotifResp {
@@ -246,7 +233,7 @@ pub(crate) fn sys_ftruncate64(request: UNotifyEventRequest) -> ScmpNotifResp {
     // SAFETY: FD-only call:
     // 1. Assert valid fd before other arguments.
     // 2. AT_FDCWD is an invalid fd argument.
-    if RawFd::try_from(req.data.args[0]).is_err() {
+    if !is_valid_fd(req.data.args[0]) {
         return request.fail_syscall(Errno::EBADF);
     }
 
@@ -273,13 +260,7 @@ pub(crate) fn sys_ftruncate64(request: UNotifyEventRequest) -> ScmpNotifResp {
             (libc::off64_t::from(len_high) << 32) | libc::off64_t::from(len_low)
         }
     } else {
-        // Align arg_no to the next even number for specific 64-bit architectures.
-        let arg_no = match req.data.arch {
-            ScmpArch::Aarch64 | ScmpArch::Ppc64 | ScmpArch::Ppc64Le => 2,
-            _ => 1,
-        };
-
-        req.data.args[arg_no] as libc::off64_t
+        req.data.args[1] as libc::off64_t
     };
 
     if len < 0 {
@@ -295,7 +276,7 @@ pub(crate) fn sys_ftruncate64(request: UNotifyEventRequest) -> ScmpNotifResp {
         request,
         "ftruncate64",
         argv,
-        |path_args: PathArgs, request, sandbox| {
+        |path_args, request, sandbox| {
             drop(sandbox); // release the read-lock.
 
             // SAFETY:
@@ -307,11 +288,12 @@ pub(crate) fn sys_ftruncate64(request: UNotifyEventRequest) -> ScmpNotifResp {
             // SAFETY: Record blocking call so it can get invalidated.
             request.cache.add_sys_block(req, false)?;
 
-            let result = ftruncate64(fd.as_raw_fd(), len).map(|_| request.return_syscall(0));
+            let result = safe_ftruncate64(fd, len).map(|_| request.return_syscall(0));
 
-            if !matches!(result, Err(Errno::EINTR)) {
-                request.cache.del_sys_block(req.id);
-            }
+            // Remove invalidation record unless interrupted.
+            request
+                .cache
+                .del_sys_block(req.id, matches!(result, Err(Errno::EINTR)))?;
 
             result
         },
@@ -324,7 +306,7 @@ pub(crate) fn sys_fallocate(request: UNotifyEventRequest) -> ScmpNotifResp {
     // SAFETY: FD-only call:
     // 1. Assert valid fd before other arguments.
     // 2. AT_FDCWD is an invalid fd argument.
-    if RawFd::try_from(req.data.args[0]).is_err() {
+    if !is_valid_fd(req.data.args[0]) {
         return request.fail_syscall(Errno::EBADF);
     }
 
@@ -378,13 +360,7 @@ pub(crate) fn sys_fallocate(request: UNotifyEventRequest) -> ScmpNotifResp {
             (libc::off64_t::from(len_high) << 32) | libc::off64_t::from(len_low)
         }
     } else {
-        // Align arg_no to the next even number for specific 64-bit architectures.
-        let arg_no = match req.data.arch {
-            ScmpArch::Aarch64 | ScmpArch::Ppc64 | ScmpArch::Ppc64Le => 4,
-            _ => 3,
-        };
-
-        req.data.args[arg_no] as libc::off64_t
+        req.data.args[3] as libc::off64_t
     };
 
     if len <= 0 {
@@ -395,32 +371,28 @@ pub(crate) fn sys_fallocate(request: UNotifyEventRequest) -> ScmpNotifResp {
         dirfd: Some(0),
         ..Default::default()
     }];
-    syscall_path_handler(
-        request,
-        "fallocate",
-        argv,
-        |path_args: PathArgs, request, sandbox| {
-            drop(sandbox); // release the read-lock.
+    syscall_path_handler(request, "fallocate", argv, |path_args, request, sandbox| {
+        drop(sandbox); // release the read-lock.
 
-            // SAFETY:
-            // 1. SysArg has one element.
-            // 2. SysArg.path is None asserting dir is Some.
-            #[allow(clippy::disallowed_methods)]
-            let fd = path_args.0.as_ref().unwrap().dir.as_ref().unwrap();
+        // SAFETY:
+        // 1. SysArg has one element.
+        // 2. SysArg.path is None asserting dir is Some.
+        #[allow(clippy::disallowed_methods)]
+        let fd = path_args.0.as_ref().unwrap().dir.as_ref().unwrap();
 
-            // SAFETY: Record blocking call so it can get invalidated.
-            request.cache.add_sys_block(req, false)?;
+        // SAFETY: Record blocking call so it can get invalidated.
+        request.cache.add_sys_block(req, false)?;
 
-            let result =
+        let result =
                 // SAFETY: nix does not have an interface for fallocate64.
                 Errno::result(unsafe { libc::fallocate64(fd.as_raw_fd(), mode.bits(), off, len) })
                     .map(|_| request.return_syscall(0));
 
-            if !matches!(result, Err(Errno::EINTR)) {
-                request.cache.del_sys_block(req.id);
-            }
+        // Remove invalidation record unless interrupted.
+        request
+            .cache
+            .del_sys_block(req.id, matches!(result, Err(Errno::EINTR)))?;
 
-            result
-        },
-    )
+        result
+    })
 }
