@@ -7,12 +7,14 @@
 // SPDX-License-Identifier: GPL-3.0
 
 use std::{
+    borrow::Cow,
     ffi::CStr,
     mem::MaybeUninit,
     os::fd::{AsFd, AsRawFd},
 };
 
 use libseccomp::ScmpNotifResp;
+use memchr::memchr;
 use nix::{errno::Errno, fcntl::AtFlags, unistd::fchdir, NixPath};
 
 use crate::{
@@ -25,6 +27,17 @@ use crate::{
     sandbox::SandboxGuard,
 };
 
+/*
+ * Constants from <linux/limits.h> not defined by libc yet.
+ */
+
+// # chars in an extended attribute name.
+const XATTR_NAME_MAX: usize = 255;
+// size of an extended attribute value (64k).
+const XATTR_SIZE_MAX: usize = 1 << 16;
+// size of extended attribute namelist (64k).
+const XATTR_LIST_MAX: usize = 1 << 16;
+
 pub(crate) fn sys_getxattr(request: UNotifyEventRequest) -> ScmpNotifResp {
     // SAFETY: We set WANT_BASE because fgetxattr requires a read-only
     // fd but we may not have access to open the file! Note, getxattr is
@@ -74,24 +87,21 @@ pub(crate) fn sys_fgetxattr(request: UNotifyEventRequest) -> ScmpNotifResp {
 
         let req = request.scmpreq;
         let name = if req.data.args[1] != 0 {
-            const SIZ: usize = libc::PATH_MAX as usize;
             let mut buf = Vec::new();
-            buf.try_reserve(SIZ).or(Err(Errno::ENOMEM))?;
-            buf.resize(SIZ, 0);
+            buf.try_reserve(XATTR_NAME_MAX).or(Err(Errno::ENOMEM))?;
+            buf.resize(XATTR_NAME_MAX, 0);
             request.read_mem(&mut buf, req.data.args[1])?;
             Some(buf)
         } else {
             None
         };
         let name = if let Some(ref name) = name {
-            CStr::from_bytes_until_nul(name)
-                .or(Err(Errno::E2BIG))?
-                .as_ptr()
+            to_name(name)?.as_ptr()
         } else {
             std::ptr::null()
         };
 
-        if sandbox.locked_for(req.pid()) {
+        if !sandbox.allow_unsafe_xattr() && sandbox.locked_for(req.pid()) {
             // SAFETY: Deny user.syd* extended attributes. name is either
             // NULL or a valid nul-terminated C-String.
             // SAFETY: Deny with ENODATA for stealth.
@@ -102,8 +112,7 @@ pub(crate) fn sys_fgetxattr(request: UNotifyEventRequest) -> ScmpNotifResp {
         // SAFETY: The size argument to the getxattr call
         // must not be fully trusted, it can be overly large,
         // and allocating a Vector of that capacity may overflow.
-        let len = usize::try_from(req.data.args[3]).or(Err(Errno::E2BIG))?;
-        let len = len.min(libc::PATH_MAX as usize); // Cap count at PATH_MAX
+        let len = to_len_cap(req.data.args[3], XATTR_SIZE_MAX)?;
         let mut buf = if len > 0 {
             let mut buf = Vec::new();
             buf.try_reserve(len).or(Err(Errno::ENOMEM))?;
@@ -116,24 +125,26 @@ pub(crate) fn sys_fgetxattr(request: UNotifyEventRequest) -> ScmpNotifResp {
             Some(b) => b.as_mut_ptr(),
             None => std::ptr::null_mut(),
         };
-        // SAFETY: In libc we trust.
-        let n = unsafe {
-            libc::fgetxattr(
-                fd.as_raw_fd(),
-                name,
-                ptr as *mut _ as *mut libc::c_void,
-                len,
-            )
-        };
+
         #[allow(clippy::cast_sign_loss)]
-        let n = if n == -1 {
-            return Err(Errno::last());
-        } else {
-            n as usize
+        // SAFETY: In libc we trust.
+        let n = match Errno::result(unsafe {
+            libc::fgetxattr(fd.as_raw_fd(), name, ptr.cast(), len)
+        }) {
+            Ok(n) => n as usize,
+            Err(Errno::ERANGE) if len == XATTR_SIZE_MAX => {
+                // SAFETY: Avoid a well-behaving process from
+                // repeating calls to potentially exhaust memory.
+                // See tar's tests for an example.
+                return Err(Errno::E2BIG);
+            }
+            Err(errno) => return Err(errno),
         };
+
         if let Some(buf) = buf {
             request.write_mem(&buf[..n], req.data.args[2])?;
         }
+
         #[allow(clippy::cast_possible_wrap)]
         Ok(request.return_syscall(n as i64))
     })
@@ -229,24 +240,21 @@ pub(crate) fn sys_lsetxattr(request: UNotifyEventRequest) -> ScmpNotifResp {
         let req = request.scmpreq;
 
         let name = if req.data.args[1] != 0 {
-            const SIZ: usize = libc::PATH_MAX as usize;
             let mut buf = Vec::new();
-            buf.try_reserve(SIZ).or(Err(Errno::ENOMEM))?;
-            buf.resize(SIZ, 0);
+            buf.try_reserve(XATTR_NAME_MAX).or(Err(Errno::ENOMEM))?;
+            buf.resize(XATTR_NAME_MAX, 0);
             request.read_mem(&mut buf, req.data.args[1])?;
             Some(buf)
         } else {
             None
         };
         let name = if let Some(ref name) = name {
-            CStr::from_bytes_until_nul(name)
-                .or(Err(Errno::E2BIG))?
-                .as_ptr()
+            to_name(name)?.as_ptr()
         } else {
             std::ptr::null()
         };
 
-        if sandbox.locked_for(req.pid()) {
+        if !sandbox.allow_unsafe_xattr() && sandbox.locked_for(req.pid()) {
             // SAFETY: Deny user.syd* extended attributes. name is either
             // NULL or a valid nul-terminated C-String.
             // SAFETY: Deny with ENODATA for stealth.
@@ -260,8 +268,7 @@ pub(crate) fn sys_lsetxattr(request: UNotifyEventRequest) -> ScmpNotifResp {
         let (buf, len) = if req.data.args[3] == 0 {
             (None, 0)
         } else {
-            let len = usize::try_from(req.data.args[3]).or(Err(Errno::E2BIG))?;
-            let len = len.min(libc::PATH_MAX as usize); // Cap count at PATH_MAX.
+            let len = to_len_val(req.data.args[3], XATTR_SIZE_MAX)?;
             let mut buf = Vec::new();
             buf.try_reserve(len).or(Err(Errno::ENOMEM))?;
             buf.resize(len, 0);
@@ -358,8 +365,7 @@ pub(crate) fn sys_flistxattr(request: UNotifyEventRequest) -> ScmpNotifResp {
             // SAFETY: The size argument to the flistxattr call
             // must not be fully trusted, it can be overly large,
             // and allocating a Vector of that capacity may overflow.
-            let len = usize::try_from(req.data.args[2]).or(Err(Errno::E2BIG))?;
-            let len = len.min(10240); // Cap count at 10240.
+            let len = to_len_cap(req.data.args[2], XATTR_LIST_MAX)?;
             let mut buf = if len > 0 {
                 let mut buf = Vec::new();
                 buf.try_reserve(len).or(Err(Errno::ENOMEM))?;
@@ -373,21 +379,26 @@ pub(crate) fn sys_flistxattr(request: UNotifyEventRequest) -> ScmpNotifResp {
                 .map_or(std::ptr::null_mut(), |b| b.as_mut_ptr())
                 as *mut libc::c_char;
 
-            // SAFETY: In libc we trust.
-            let n = unsafe { libc::flistxattr(fd.as_raw_fd(), ptr, len) };
             #[allow(clippy::cast_sign_loss)]
-            let n = if n == -1 {
-                return Err(Errno::last());
-            } else {
-                n as usize
+            // SAFETY: In libc we trust.
+            let n = match Errno::result(unsafe { libc::flistxattr(fd.as_raw_fd(), ptr, len) }) {
+                Ok(n) => n as usize,
+                Err(Errno::ERANGE) if len == XATTR_LIST_MAX => {
+                    // SAFETY: Avoid a well-behaving process from
+                    // repeating calls to potentially exhaust memory.
+                    // See tar's tests for an example.
+                    return Err(Errno::E2BIG);
+                }
+                Err(errno) => return Err(errno),
             };
+
             let n = if let Some(buf) = buf {
                 // SAFETY: Filter out attributes that start with "user.syd".
                 // SAFETY: Deny only if the Sandbox is locked for the process.
-                let buf = if sandbox.locked_for(req.pid()) {
-                    filterxattr(&buf[..n], n)?
+                let buf = if !sandbox.allow_unsafe_xattr() && sandbox.locked_for(req.pid()) {
+                    Cow::Owned(filterxattr(&buf[..n], n)?)
                 } else {
-                    buf
+                    Cow::Borrowed(&buf[..n])
                 };
 
                 request.write_mem(&buf, req.data.args[1])?;
@@ -395,6 +406,7 @@ pub(crate) fn sys_flistxattr(request: UNotifyEventRequest) -> ScmpNotifResp {
             } else {
                 n
             };
+
             #[allow(clippy::cast_possible_wrap)]
             Ok(request.return_syscall(n as i64))
         },
@@ -537,24 +549,21 @@ pub(crate) fn sys_lremovexattr(request: UNotifyEventRequest) -> ScmpNotifResp {
             let req = request.scmpreq;
 
             let name = if req.data.args[1] != 0 {
-                const SIZ: usize = libc::PATH_MAX as usize;
                 let mut buf = Vec::new();
-                buf.try_reserve(SIZ).or(Err(Errno::ENOMEM))?;
-                buf.resize(SIZ, 0);
+                buf.try_reserve(XATTR_NAME_MAX).or(Err(Errno::ENOMEM))?;
+                buf.resize(XATTR_NAME_MAX, 0);
                 request.read_mem(&mut buf, req.data.args[1])?;
                 Some(buf)
             } else {
                 None
             };
             let name = if let Some(ref name) = name {
-                CStr::from_bytes_until_nul(name)
-                    .or(Err(Errno::E2BIG))?
-                    .as_ptr()
+                to_name(name)?.as_ptr()
             } else {
                 std::ptr::null()
             };
 
-            if sandbox.locked_for(req.pid()) {
+            if !sandbox.allow_unsafe_xattr() && sandbox.locked_for(req.pid()) {
                 // SAFETY: Deny user.syd* extended attributes.
                 // name is either NULL or a valid nul-terminated C-String.
                 // SAFETY: Deny with ENODATA for stealth.
@@ -642,24 +651,21 @@ fn syscall_getxattr_handler(
     };
 
     let name = if req.data.args[1] != 0 {
-        const SIZ: usize = libc::PATH_MAX as usize;
         let mut buf = Vec::new();
-        buf.try_reserve(SIZ).or(Err(Errno::ENOMEM))?;
-        buf.resize(SIZ, 0);
+        buf.try_reserve(XATTR_NAME_MAX).or(Err(Errno::ENOMEM))?;
+        buf.resize(XATTR_NAME_MAX, 0);
         request.read_mem(&mut buf, req.data.args[1])?;
         Some(buf)
     } else {
         None
     };
     let name = if let Some(ref name) = name {
-        CStr::from_bytes_until_nul(name)
-            .or(Err(Errno::E2BIG))?
-            .as_ptr()
+        to_name(name)?.as_ptr()
     } else {
         std::ptr::null()
     };
 
-    if sandbox.locked_for(req.pid()) {
+    if !sandbox.allow_unsafe_xattr() && sandbox.locked_for(req.pid()) {
         // SAFETY: Deny user.syd* extended attributes. name is either
         // NULL or a valid nul-terminated C-String.
         // SAFETY: Deny with ENODATA for stealth.
@@ -670,10 +676,7 @@ fn syscall_getxattr_handler(
     // SAFETY: The size argument to the getxattr call
     // must not be fully trusted, it can be overly large,
     // and allocating a Vector of that capacity may overflow.
-    #[allow(clippy::cast_possible_truncation)]
-    let len = req.data.args[3] as usize;
-    let len = len.min(libc::PATH_MAX as usize); // Cap count at PATH_MAX.
-
+    let len = to_len_cap(req.data.args[3], XATTR_SIZE_MAX)?;
     let mut buf = if len > 0 {
         let mut buf = Vec::new();
         buf.try_reserve(len).or(Err(Errno::ENOMEM))?;
@@ -701,7 +704,16 @@ fn syscall_getxattr_handler(
         .with_nix_path(|cstr| unsafe { libc::lgetxattr(cstr.as_ptr(), name, ptr.cast(), len) })?;
 
     #[allow(clippy::cast_sign_loss)]
-    let n = Errno::result(res)? as usize;
+    let n = match Errno::result(res) {
+        Ok(n) => n as usize,
+        Err(Errno::ERANGE) if len == XATTR_SIZE_MAX => {
+            // SAFETY: Avoid a well-behaving process from
+            // repeating calls to potentially exhaust memory.
+            // See tar's tests for an example.
+            return Err(Errno::E2BIG);
+        }
+        Err(errno) => return Err(errno),
+    };
 
     if let Some(buf) = buf {
         request.write_mem(&buf[..n], req.data.args[2])?;
@@ -759,24 +771,21 @@ fn syscall_getxattrat_handler(
     }
 
     let name = if req.data.args[3] != 0 {
-        const SIZ: usize = libc::PATH_MAX as usize;
         let mut buf = Vec::new();
-        buf.try_reserve(SIZ).or(Err(Errno::ENOMEM))?;
-        buf.resize(SIZ, 0);
+        buf.try_reserve(XATTR_NAME_MAX).or(Err(Errno::ENOMEM))?;
+        buf.resize(XATTR_NAME_MAX, 0);
         request.read_mem(&mut buf, req.data.args[3])?;
         Some(buf)
     } else {
         None
     };
     let name = if let Some(ref name) = name {
-        CStr::from_bytes_until_nul(name)
-            .or(Err(Errno::E2BIG))?
-            .as_ptr()
+        to_name(name)?.as_ptr()
     } else {
         std::ptr::null()
     };
 
-    if sandbox.locked_for(req.pid()) {
+    if !sandbox.allow_unsafe_xattr() && sandbox.locked_for(req.pid()) {
         // SAFETY: Deny user.syd* extended attributes. name is either
         // NULL or a valid nul-terminated C-String.
         // SAFETY: Deny with ENODATA for stealth.
@@ -787,10 +796,7 @@ fn syscall_getxattrat_handler(
     // SAFETY: The size element of the struct xattr_args
     // must not be fully trusted, it can be overly large,
     // and allocating a Vector of that capacity may overflow.
-    #[allow(clippy::cast_possible_truncation)]
-    let len = args.size as usize;
-    let len = len.min(libc::PATH_MAX as usize); // Cap count at PATH_MAX.
-
+    let len = to_len_cap(args.size.into(), XATTR_SIZE_MAX)?;
     let mut buf = if len > 0 {
         let mut buf = Vec::new();
         buf.try_reserve(len).or(Err(Errno::ENOMEM))?;
@@ -818,7 +824,17 @@ fn syscall_getxattrat_handler(
     // SAFETY:
     // 1. `name` is a valid raw pointer (may be NULL)!
     // 2. We do not resolve symbolic links here!
-    let n = unsafe { getxattrat(fd, base, name, &mut my_args, AtFlags::AT_SYMLINK_NOFOLLOW) }?;
+    let n = match unsafe { getxattrat(fd, base, name, &mut my_args, AtFlags::AT_SYMLINK_NOFOLLOW) }
+    {
+        Ok(n) => n,
+        Err(Errno::ERANGE) if len == XATTR_SIZE_MAX => {
+            // SAFETY: Avoid a well-behaving process from
+            // repeating calls to potentially exhaust memory.
+            // See tar's tests for an example.
+            return Err(Errno::E2BIG);
+        }
+        Err(errno) => return Err(errno),
+    };
 
     if let Some(buf) = buf {
         request.write_mem(&buf[..n], args.value)?;
@@ -842,24 +858,21 @@ fn syscall_setxattr_handler(
 
     let req = request.scmpreq;
     let name = if req.data.args[1] != 0 {
-        const SIZ: usize = libc::PATH_MAX as usize;
         let mut buf = Vec::new();
-        buf.try_reserve(SIZ).or(Err(Errno::ENOMEM))?;
-        buf.resize(SIZ, 0);
+        buf.try_reserve(XATTR_NAME_MAX).or(Err(Errno::ENOMEM))?;
+        buf.resize(XATTR_NAME_MAX, 0);
         request.read_mem(&mut buf, req.data.args[1])?;
         Some(buf)
     } else {
         None
     };
     let name = if let Some(ref name) = name {
-        CStr::from_bytes_until_nul(name)
-            .or(Err(Errno::E2BIG))?
-            .as_ptr()
+        to_name(name)?.as_ptr()
     } else {
         std::ptr::null()
     };
 
-    if sandbox.locked_for(req.pid()) {
+    if !sandbox.allow_unsafe_xattr() && sandbox.locked_for(req.pid()) {
         // SAFETY: Deny user.syd* extended attributes. name is either
         // NULL or a valid nul-terminated C-String.
         // SAFETY: Deny with EACCES to denote access violation.
@@ -873,8 +886,7 @@ fn syscall_setxattr_handler(
     let (buf, len) = if req.data.args[3] == 0 {
         (None, 0)
     } else {
-        let len = usize::try_from(req.data.args[3]).or(Err(Errno::E2BIG))?;
-        let len = len.min(libc::PATH_MAX as usize); // Cap count at PATH_MAX.
+        let len = to_len_val(req.data.args[3], XATTR_SIZE_MAX)?;
         let mut buf = Vec::new();
         buf.try_reserve(len).or(Err(Errno::ENOMEM))?;
         buf.resize(len, 0);
@@ -887,11 +899,8 @@ fn syscall_setxattr_handler(
     let flags = req.data.args[4] as libc::c_int;
 
     // SAFETY: In libc we trust.
-    if unsafe { libc::fsetxattr(fd.as_raw_fd(), name, buf, len, flags) } == 0 {
-        Ok(request.return_syscall(0))
-    } else {
-        Err(Errno::last())
-    }
+    Errno::result(unsafe { libc::fsetxattr(fd.as_raw_fd(), name, buf, len, flags) })
+        .map(|_| request.return_syscall(0))
 }
 
 /// A helper function to handle setxattrat syscall.
@@ -937,24 +946,21 @@ fn syscall_setxattrat_handler(
     let args = unsafe { args.assume_init() };
 
     let name = if req.data.args[3] != 0 {
-        const SIZ: usize = libc::PATH_MAX as usize;
         let mut buf = Vec::new();
-        buf.try_reserve(SIZ).or(Err(Errno::ENOMEM))?;
-        buf.resize(SIZ, 0);
+        buf.try_reserve(XATTR_NAME_MAX).or(Err(Errno::ENOMEM))?;
+        buf.resize(XATTR_NAME_MAX, 0);
         request.read_mem(&mut buf, req.data.args[3])?;
         Some(buf)
     } else {
         None
     };
     let name = if let Some(ref name) = name {
-        CStr::from_bytes_until_nul(name)
-            .or(Err(Errno::E2BIG))?
-            .as_ptr()
+        to_name(name)?.as_ptr()
     } else {
         std::ptr::null()
     };
 
-    if sandbox.locked_for(req.pid()) {
+    if !sandbox.allow_unsafe_xattr() && sandbox.locked_for(req.pid()) {
         // SAFETY: Deny user.syd* extended attributes. name is either
         // NULL or a valid nul-terminated C-String.
         // SAFETY: Deny with EACCES to denote access violation.
@@ -968,8 +974,7 @@ fn syscall_setxattrat_handler(
     let (buf, len) = if args.size == 0 {
         (None, 0)
     } else {
-        let len = usize::try_from(args.size).or(Err(Errno::E2BIG))?;
-        let len = len.min(libc::PATH_MAX as usize); // Cap count at PATH_MAX.
+        let len = to_len_val(args.size.into(), XATTR_SIZE_MAX)?;
         let mut buf = Vec::new();
         buf.try_reserve(len).or(Err(Errno::ENOMEM))?;
         buf.resize(len, 0);
@@ -1018,10 +1023,7 @@ fn syscall_listxattr_handler(
     // SAFETY: The size argument to the llistxattr call
     // must not be fully trusted, it can be overly large,
     // and allocating a Vector of that capacity may overflow.
-    let len = usize::try_from(req.data.args[2])
-        .or(Err(Errno::E2BIG))?
-        .min(10240); // Cap count at 10240.
-
+    let len = to_len_cap(req.data.args[2], XATTR_LIST_MAX)?;
     let mut buf = if len > 0 {
         let mut buf = Vec::new();
         buf.try_reserve(len).or(Err(Errno::ENOMEM))?;
@@ -1048,15 +1050,24 @@ fn syscall_listxattr_handler(
         .with_nix_path(|cstr| unsafe { libc::llistxattr(cstr.as_ptr(), ptr, len) })?;
 
     #[allow(clippy::cast_sign_loss)]
-    let mut n = Errno::result(res)? as usize;
+    let mut n = match Errno::result(res) {
+        Ok(n) => n as usize,
+        Err(Errno::ERANGE) if len == XATTR_LIST_MAX => {
+            // SAFETY: Avoid a well-behaving process from
+            // repeating calls to potentially exhaust memory.
+            // See tar's tests for an example.
+            return Err(Errno::E2BIG);
+        }
+        Err(errno) => return Err(errno),
+    };
 
     if let Some(buf) = buf {
         // SAFETY: Filter out attributes that start with "user.syd".
         // SAFETY: Deny only if the Sandbox is locked for the process.
-        let buf = if sandbox.locked_for(req.pid()) {
-            filterxattr(&buf[..n], n)?
+        let buf = if !sandbox.allow_unsafe_xattr() && sandbox.locked_for(req.pid()) {
+            Cow::Owned(filterxattr(&buf[..n], n)?)
         } else {
-            buf[..n].to_vec()
+            Cow::Borrowed(&buf[..n])
         };
 
         request.write_mem(&buf, req.data.args[1])?;
@@ -1088,10 +1099,7 @@ fn syscall_listxattrat_handler(
     // SAFETY: The size argument to the llistxattr call
     // must not be fully trusted, it can be overly large,
     // and allocating a Vector of that capacity may overflow.
-    let len = usize::try_from(req.data.args[4])
-        .or(Err(Errno::E2BIG))?
-        .min(10240); // Cap count at 10240.
-
+    let len = to_len_cap(req.data.args[4], XATTR_LIST_MAX)?;
     let mut buf = if len > 0 {
         let mut buf = Vec::new();
         buf.try_reserve(len).or(Err(Errno::ENOMEM))?;
@@ -1113,15 +1121,24 @@ fn syscall_listxattrat_handler(
     // SAFETY:
     // 1. `ptr` is a valid raw pointer (may be NULL)!
     // 2. We do not resolve symbolic links here!
-    let mut n = unsafe { listxattrat(fd, base, AtFlags::AT_SYMLINK_NOFOLLOW, ptr, len) }?;
+    let mut n = match unsafe { listxattrat(fd, base, AtFlags::AT_SYMLINK_NOFOLLOW, ptr, len) } {
+        Ok(n) => n,
+        Err(Errno::ERANGE) if len == XATTR_LIST_MAX => {
+            // SAFETY: Avoid a well-behaving process from
+            // repeating calls to potentially exhaust memory.
+            // See tar's tests for an example.
+            return Err(Errno::E2BIG);
+        }
+        Err(errno) => return Err(errno),
+    };
 
     if let Some(buf) = buf {
         // SAFETY: Filter out attributes that start with "user.syd".
         // SAFETY: Deny only if the Sandbox is locked for the process.
-        let buf = if sandbox.locked_for(req.pid()) {
-            filterxattr(&buf[..n], n)?
+        let buf = if !sandbox.allow_unsafe_xattr() && sandbox.locked_for(req.pid()) {
+            Cow::Owned(filterxattr(&buf[..n], n)?)
         } else {
-            buf[..n].to_vec()
+            Cow::Borrowed(&buf[..n])
         };
 
         request.write_mem(&buf, req.data.args[3])?;
@@ -1147,24 +1164,21 @@ fn syscall_removexattr_handler(
     let req = request.scmpreq;
 
     let name = if req.data.args[1] != 0 {
-        const SIZ: usize = libc::PATH_MAX as usize;
         let mut buf = Vec::new();
-        buf.try_reserve(SIZ).or(Err(Errno::ENOMEM))?;
-        buf.resize(SIZ, 0);
+        buf.try_reserve(XATTR_NAME_MAX).or(Err(Errno::ENOMEM))?;
+        buf.resize(XATTR_NAME_MAX, 0);
         request.read_mem(&mut buf, req.data.args[1])?;
         Some(buf)
     } else {
         None
     };
     let name = if let Some(ref name) = name {
-        CStr::from_bytes_until_nul(name)
-            .or(Err(Errno::E2BIG))?
-            .as_ptr()
+        to_name(name)?.as_ptr()
     } else {
         std::ptr::null()
     };
 
-    if sandbox.locked_for(req.pid()) {
+    if !sandbox.allow_unsafe_xattr() && sandbox.locked_for(req.pid()) {
         // SAFETY: Deny user.syd* extended attributes.
         // name is either NULL or a valid nul-terminated C-String.
         // SAFETY: Deny with ENODATA for stealth.
@@ -1173,11 +1187,8 @@ fn syscall_removexattr_handler(
     }
 
     // SAFETY: In libc we trust.
-    if unsafe { libc::fremovexattr(fd.as_raw_fd(), name) } == 0 {
-        Ok(request.return_syscall(0))
-    } else {
-        Err(Errno::last())
-    }
+    Errno::result(unsafe { libc::fremovexattr(fd.as_raw_fd(), name) })
+        .map(|_| request.return_syscall(0))
 }
 
 /// A helper function to handle removexattrat syscall.
@@ -1199,24 +1210,21 @@ fn syscall_removexattrat_handler(
     };
 
     let name = if req.data.args[3] != 0 {
-        const SIZ: usize = libc::PATH_MAX as usize;
         let mut buf = Vec::new();
-        buf.try_reserve(SIZ).or(Err(Errno::ENOMEM))?;
-        buf.resize(SIZ, 0);
+        buf.try_reserve(XATTR_NAME_MAX).or(Err(Errno::ENOMEM))?;
+        buf.resize(XATTR_NAME_MAX, 0);
         request.read_mem(&mut buf, req.data.args[3])?;
         Some(buf)
     } else {
         None
     };
     let name = if let Some(ref name) = name {
-        CStr::from_bytes_until_nul(name)
-            .or(Err(Errno::E2BIG))?
-            .as_ptr()
+        to_name(name)?.as_ptr()
     } else {
         std::ptr::null()
     };
 
-    if sandbox.locked_for(req.pid()) {
+    if !sandbox.allow_unsafe_xattr() && sandbox.locked_for(req.pid()) {
         // SAFETY: Deny user.syd* extended attributes.
         // name is either NULL or a valid nul-terminated C-String.
         // SAFETY: Deny with ENODATA for stealth.
@@ -1235,3 +1243,43 @@ fn syscall_removexattrat_handler(
     unsafe { removexattrat(fd, base, name, AtFlags::AT_SYMLINK_NOFOLLOW) }
         .map(|_| request.return_syscall(0))
 }
+
+// Capping length converter, used by *{get,list}xattr*
+#[inline]
+fn to_len_cap(arg: u64, max: usize) -> Result<usize, Errno> {
+    Ok(usize::try_from(arg).or(Err(Errno::E2BIG))?.min(max))
+}
+
+// Validating length converter, used by *setxattr*
+#[inline]
+fn to_len_val(arg: u64, max: usize) -> Result<usize, Errno> {
+    match usize::try_from(arg).or(Err(Errno::ERANGE)) {
+        Ok(len) if len <= max => Ok(len),
+        _ => Err(Errno::ERANGE),
+    }
+}
+
+// Validate extended attribute name.
+#[inline]
+fn to_name(name: &[u8]) -> Result<&CStr, Errno> {
+    // Note, the ERANGE error return here is misleading:
+    // https://bugzilla.kernel.org/show_bug.cgi?id=220374
+    let name = CStr::from_bytes_until_nul(name).or(Err(Errno::ERANGE))?;
+
+    // Check for non-empty name.
+    let buf = name.to_bytes();
+    let len = buf.len();
+    if len == 0 {
+        return Err(Errno::ERANGE);
+    }
+
+    // Check for qualified name in namespace.attribute form.
+    // EINVAL is expected here by sys-apps/attr's tests.
+    #[allow(clippy::arithmetic_side_effects)]
+    match memchr(b'.', buf) {
+        None => Err(Errno::EOPNOTSUPP),
+        Some(0) => Err(Errno::EINVAL),
+        Some(n) if n == len - 1 => Err(Errno::EINVAL),
+        _ => Ok(name),
+    }
+}
