@@ -109,6 +109,8 @@
 // TODO: remove the allow-lint below when libseccomp is back to upstream.
 #![allow(macro_expanded_macro_exports_accessed_by_absolute_paths)]
 
+/// JSON serializers for syd(2) API
+pub mod api;
 /// System call argument cookies
 pub(crate) mod cookie;
 /// System call handlers
@@ -251,6 +253,34 @@ use crate::{
     spec::{speculation_get, SpeculationFeature},
 };
 
+/* Macros */
+
+/// Convenience macro to define a main function with correct errno return.
+#[macro_export]
+macro_rules! main {
+    { $($body:tt)* } => {
+        fn main() -> std::process::ExitCode {
+            match (|| -> syd::err::SydResult<std::process::ExitCode> { $($body)* })() {
+                Ok(code) => code,
+                Err(err) => {
+                    use std::io::Write;
+
+                    let desc = format!("Error: {err}\n");
+                    let _ = std::io::stderr().write_all(desc.as_bytes());
+
+                    u8::try_from(
+                        err.errno()
+                           .map(|e| e as i32)
+                           .unwrap_or(128)
+                    )
+                    .map(std::process::ExitCode::from)
+                    .unwrap_or(std::process::ExitCode::FAILURE)
+                }
+            }
+        }
+    };
+}
+
 /* Data structures */
 
 /// Simple wrapper over ScmpSyscall and ScmpArch to provide Display.
@@ -290,6 +320,14 @@ impl Serialize for Sydcall {
 
 pub(crate) struct SydArch(ScmpArch);
 
+impl std::fmt::Display for SydArch {
+    fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
+        let arch = format!("{:?}", self.0).to_ascii_lowercase();
+        let arch = if arch == { "x8664" } { "x86_64" } else { &arch };
+        write!(f, "{arch}")
+    }
+}
+
 impl Serialize for SydArch {
     fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
     where
@@ -603,11 +641,26 @@ impl SydMsFlags {
 
 /* Utilities */
 
+/// Check for CONFIG_CROSS_MEMORY_ATTACH support.
+pub fn check_cross_memory_attach() -> bool {
+    // SAFETY: We explicitly call the system call with
+    // invalid arguments just to check for host Linux
+    // kernel support.
+    !matches!(
+        Errno::result(unsafe {
+            libc::process_vm_readv(0, std::ptr::null(), 0, std::ptr::null(), 0, 0)
+        }),
+        Err(Errno::ENOSYS)
+    )
+}
+
 /// Print Syd version information,
 /// and information about the system to
 /// standard output.
 #[allow(clippy::cognitive_complexity)]
 pub fn syd_info() {
+    use crate::config::*;
+
     #[allow(clippy::disallowed_methods)]
     let major = env!("CARGO_PKG_VERSION_MAJOR")
         .parse::<u64>()
@@ -635,7 +688,7 @@ pub fn syd_info() {
         .join(" ");
 
     println!("syd {} ({})", *crate::config::VERSION, code_name);
-    println!("Author: Ali Polatel");
+    println!("Author: Ali Polatel <alip@chesswob.org>");
     println!("License: GPL-3.0");
 
     let feat = [
@@ -707,15 +760,7 @@ pub fn syd_info() {
     );
 
     // Check CONFIG_CROSS_MEMORY_ATTACH.
-    let cfg_cma = !matches!(
-        // SAFETY: We explicitly call the system call with
-        // invalid arguments just to check for host Linux
-        // kernel support.
-        Errno::result(unsafe {
-            libc::process_vm_readv(0, std::ptr::null(), 0, std::ptr::null(), 0, 0)
-        }),
-        Err(Errno::ENOSYS)
-    );
+    let cfg_cma = check_cross_memory_attach();
     println!(
         "Cross memory attach is {}supported{}",
         if cfg_cma { "" } else { "not " },
@@ -752,17 +797,9 @@ pub fn syd_info() {
     print!("Open file limits: {nofile_soft} soft, {nofile_hard} hard, ");
 
     // Lookup system-wide open file limits.
-    match crate::config::proc_init() {
-        Ok(_) => {
-            let file_max = proc_fs_file_max().unwrap_or(0);
-            let nr_open = proc_fs_nr_open().unwrap_or(0);
-            crate::config::proc_close();
-            println!("{nr_open} nr_open, {file_max} file-max");
-        }
-        Err(errno) => {
-            println!("? nr_open, ? file-max (error: {errno})");
-        }
-    }
+    let file_max = proc_fs_file_max().unwrap_or(0);
+    let nr_open = proc_fs_nr_open().unwrap_or(0);
+    println!("{nr_open} nr_open, {file_max} file-max");
 
     let uname = match uname() {
         Ok(info) => OsStr::to_str(info.release()).unwrap_or("?").to_string(),
@@ -774,6 +811,34 @@ pub fn syd_info() {
         seccomp_arch_native_name().unwrap_or("?")
     );
 
+    // Print detected host Linux kernel version and related features.
+    print!("Host Linux: {}.{} with", KERNEL_VERSION.0, KERNEL_VERSION.1);
+    print!(
+        " {}pidfd_thread",
+        if *HAVE_PIDFD_THREAD { "+" } else { "-" }
+    );
+    print!(
+        ", {}procmap_query",
+        if *HAVE_PROCMAP_QUERY { "+" } else { "-" }
+    );
+    print!(
+        ", {}seccomp_user_notif_fd_sync_wake_up",
+        if *HAVE_SECCOMP_USER_NOTIF_FD_SYNC_WAKE_UP {
+            "+"
+        } else {
+            "-"
+        }
+    );
+    print!(
+        ", {}statx_mnt_id_unique",
+        if *HAVE_STATX_MNT_ID_UNIQUE { "+" } else { "-" }
+    );
+    print!(
+        ", {}at_execve_check",
+        if *HAVE_AT_EXECVE_CHECK { "+" } else { "-" }
+    );
+    println!(", mmap_min_addr={}", *MMAP_MIN_ADDR);
+
     // SAFETY: In libc we trust.
     let pers = match unsafe { libc::personality(0xFFFFFFFF) } {
         n if n < 0 => format!("? (error: {})", Errno::last()),
@@ -1438,13 +1503,15 @@ pub(crate) const SCMP_ARCH: &[ScmpArch] = &[ScmpArch::Mips];
 #[cfg(all(target_arch = "mips32r6", target_endian = "little"))]
 pub(crate) const SCMP_ARCH: &[ScmpArch] = &[ScmpArch::Mipsel];
 #[cfg(all(target_arch = "mips64", target_endian = "big"))]
-pub(crate) const SCMP_ARCH: &[ScmpArch] = &[ScmpArch::Mips64, ScmpArch::Mips64N32];
+pub(crate) const SCMP_ARCH: &[ScmpArch] = &[ScmpArch::Mips64, ScmpArch::Mips64N32, ScmpArch::Mips];
 #[cfg(all(target_arch = "mips64", target_endian = "little"))]
-pub(crate) const SCMP_ARCH: &[ScmpArch] = &[ScmpArch::Mipsel64, ScmpArch::Mipsel64N32];
+pub(crate) const SCMP_ARCH: &[ScmpArch] =
+    &[ScmpArch::Mipsel64, ScmpArch::Mipsel64N32, ScmpArch::Mipsel];
 #[cfg(all(target_arch = "mips64r6", target_endian = "big"))]
-pub(crate) const SCMP_ARCH: &[ScmpArch] = &[ScmpArch::Mips64, ScmpArch::Mips64N32];
+pub(crate) const SCMP_ARCH: &[ScmpArch] = &[ScmpArch::Mips64, ScmpArch::Mips64N32, ScmpArch::Mips];
 #[cfg(all(target_arch = "mips64r6", target_endian = "little"))]
-pub(crate) const SCMP_ARCH: &[ScmpArch] = &[ScmpArch::Mipsel64, ScmpArch::Mipsel64N32];
+pub(crate) const SCMP_ARCH: &[ScmpArch] =
+    &[ScmpArch::Mipsel64, ScmpArch::Mipsel64N32, ScmpArch::Mipsel];
 #[cfg(target_arch = "powerpc")]
 pub(crate) const SCMP_ARCH: &[ScmpArch] = &[ScmpArch::Ppc];
 #[cfg(all(target_arch = "powerpc64", target_endian = "big"))]
@@ -1462,7 +1529,30 @@ pub(crate) const SCMP_ARCH: &[ScmpArch] = &[ScmpArch::S390X, ScmpArch::S390];
 #[cfg(target_arch = "loongarch64")]
 pub(crate) const SCMP_ARCH: &[ScmpArch] = &[ScmpArch::Loongarch64];
 
-/// Return true if native architecture has the multiplexed socketcall system call.
+/// List of architectures with the ipc(2) multiplexer system call.
+pub(crate) const IPC_ARCH: &[ScmpArch] = &[
+    ScmpArch::X86,
+    ScmpArch::M68k,
+    ScmpArch::Mips,
+    ScmpArch::Mipsel,
+    ScmpArch::Ppc,
+    ScmpArch::Ppc64,
+    ScmpArch::Ppc64Le,
+    ScmpArch::S390X,
+    ScmpArch::S390,
+    ScmpArch::Sheb,
+    ScmpArch::Sh,
+];
+
+/// Return true if native architecture has the multiplexed ipc(2) system call.
+///
+/// Panics if it cannot determine the native architecture.
+pub fn seccomp_native_has_ipc() -> bool {
+    IPC_ARCH.contains(&ScmpArch::native())
+}
+
+/// Return true if native architecture has the multiplexed socketcall(2) system call.
+///
 /// Panics if it cannot determine the native architecture.
 pub fn seccomp_native_has_socketcall() -> bool {
     matches!(
@@ -1726,13 +1816,17 @@ pub fn human_size(bytes: usize) -> String {
     format!("{:.2}{}", size, SIZES[i])
 }
 
+#[allow(clippy::unnecessary_cast)]
+const SIOCGIFINDEX: u64 = libc::SIOCGIFINDEX as u64;
 #[allow(clippy::unnecessary_cast)]
 const SIOCGIFFLAGS: u64 = libc::SIOCGIFFLAGS as u64;
 #[allow(clippy::unnecessary_cast)]
 const SIOCSIFFLAGS: u64 = libc::SIOCSIFFLAGS as u64;
 
 /// Functionally equivalent to "ifconfig lo up".
-pub fn bring_up_loopback() -> Result<(), Errno> {
+///
+/// Returns loopback interface index.
+pub fn loopback_set_up() -> Result<i32, Errno> {
     // Create a socket
     let sock = socket(
         AddressFamily::Inet,
@@ -1766,6 +1860,38 @@ pub fn bring_up_loopback() -> Result<(), Errno> {
         ifr_ifru: unsafe { std::mem::zeroed() },
     };
 
+    // SAFETY: Request loopback network device index.
+    let loindex: i32 = unsafe {
+        let mut ifr_index: libc::ifreq = std::mem::zeroed();
+        ifr_index.ifr_name = ifreq.ifr_name;
+        Errno::result(libc::syscall(
+            libc::SYS_ioctl,
+            sock.as_raw_fd(),
+            SIOCGIFINDEX as libc::c_ulong,
+            &mut ifr_index,
+        ))?;
+        // HACK: ifr_ifru is a union but libc crate does not define ifru_ivalue,
+        // which is a libc::c_int, so here we refer to it with ifru_mtu which
+        // is the same type.
+        ifr_index.ifr_ifru.ifru_mtu
+    };
+
+    // Set BIGTCP to LOOPBACK_BIGTCP_MAX if available.
+    // Note, we _must_ do this before setting up the network device.
+    use crate::config::LOOPBACK_BIGTCP_MAX;
+    match loopback_set_bigtcp(loindex, LOOPBACK_BIGTCP_MAX) {
+        Ok(_) => {
+            info!("ctx": "loopback_set_bigtcp",
+                "msg": "loopback network device has BIGTCP set",
+                "max": LOOPBACK_BIGTCP_MAX);
+        }
+        Err(errno) => {
+            info!("ctx": "loopback_set_bigtcp",
+                "msg": format!("set BIGTCP for loopback network device error: {errno}"),
+                "err": errno as i32);
+        }
+    };
+
     // SAFETY: Get the current flags.
     Errno::result(unsafe {
         libc::syscall(
@@ -1792,8 +1918,183 @@ pub fn bring_up_loopback() -> Result<(), Errno> {
             SIOCSIFFLAGS as libc::c_ulong,
             &mut ifreq,
         )
-    })
-    .map(drop)
+    })?;
+
+    Ok(loindex)
+}
+
+// libc crate does not define struct nl from linux/rtnetlink.h.
+#[repr(C)]
+#[derive(Debug, Clone, Copy)]
+struct nlattr {
+    nla_len: u16,
+    nla_type: u16,
+}
+
+// libc crate does not define struct nlmsg from linux/rtnetlink.h.
+#[repr(C)]
+#[derive(Debug, Clone, Copy)]
+struct nlmsg {
+    hdr: libc::nlmsghdr,
+    info: ifinfomsg,
+    attrs: [u8; 64],
+}
+
+// libc crate does not define struct ifinfomsg from linux/rtnetlink.h yet.
+#[repr(C)]
+#[derive(Debug, Copy, Clone)]
+struct ifinfomsg {
+    family: u8,
+    pad: u8,
+    ifi_type: u16, // ARPHRD_*
+    index: i32,    // Interface index
+    flags: u32,    // IFF_* flags
+    change: u32,   // IFF_* change mask
+}
+
+// These values are based on the Linux kernel headers.
+const IFLA_GRO_IPV4_MAX_SIZE: libc::c_ushort = 0x40;
+const IFLA_GRO_MAX_SIZE: libc::c_ushort = 0x3a;
+const IFLA_GSO_IPV4_MAX_SIZE: libc::c_ushort = 0x3f;
+const IFLA_GSO_MAX_SIZE: libc::c_ushort = 0x29;
+
+/// Functionally equivalent to "ip link set dev $ifindex g{r,s}o_max_size $max_size",
+/// which sets BIGTCP if available, see: https://lwn.net/Articles/884104/
+///
+/// Requires loopback interface index as argument.
+pub fn loopback_set_bigtcp(ifindex: i32, max_size: u32) -> Result<(), Errno> {
+    // Set BIGTCP to max_size if available.
+    use netlink_sys::{constants::*, Socket, SocketAddr};
+
+    // SAFETY: create netlink socket using netlink_sys for NETLINK_ROUTE.
+    let mut sock = Socket::new(NETLINK_ROUTE)
+        .map_err(|e| Errno::from_raw(e.raw_os_error().unwrap_or(libc::ENOSYS)))?;
+    sock.bind(&SocketAddr::new(0, 0))
+        .map_err(|e| Errno::from_raw(e.raw_os_error().unwrap_or(libc::ENOSYS)))?;
+
+    // SAFETY: Zero initialize. netlink message.
+    let mut msg: nlmsg = unsafe { std::mem::zeroed() };
+
+    // SAFETY: Set up netlink header.
+    let nl_hdr = &mut msg.hdr;
+    #[allow(clippy::arithmetic_side_effects)]
+    #[allow(clippy::cast_possible_truncation)]
+    {
+        nl_hdr.nlmsg_len =
+            (std::mem::size_of::<libc::nlmsghdr>() + std::mem::size_of::<ifinfomsg>()) as u32;
+        nl_hdr.nlmsg_type = libc::RTM_NEWLINK;
+        nl_hdr.nlmsg_flags = (libc::NLM_F_REQUEST | libc::NLM_F_ACK) as u16;
+        nl_hdr.nlmsg_seq = 1;
+        nl_hdr.nlmsg_pid = 0;
+    }
+
+    // SAFETY: Populate ifinfomsg.
+    let info = &mut msg.info;
+    #[allow(clippy::cast_possible_truncation)]
+    {
+        info.family = libc::AF_UNSPEC as u8;
+        info.index = ifindex;
+        info.change = u32::MAX;
+    }
+
+    let mut offset = 0;
+    for &kind in &[
+        IFLA_GRO_IPV4_MAX_SIZE,
+        IFLA_GRO_MAX_SIZE,
+        IFLA_GSO_IPV4_MAX_SIZE,
+        IFLA_GSO_MAX_SIZE,
+    ] {
+        // SAFETY: Set attribute header.
+        #[allow(clippy::cast_ptr_alignment)]
+        let attr_ptr = unsafe { msg.attrs.as_mut_ptr().add(offset) as *mut nlattr };
+
+        // SAFETY: Write attribute metadata.
+        #[allow(clippy::arithmetic_side_effects)]
+        #[allow(clippy::cast_possible_truncation)]
+        unsafe {
+            (*attr_ptr).nla_type = kind;
+            (*attr_ptr).nla_len =
+                (std::mem::size_of::<nlattr>() + std::mem::size_of::<u32>()) as u16;
+        }
+
+        // SAFETY: Write u32 payload.
+        #[allow(clippy::cast_ptr_alignment)]
+        unsafe {
+            let ptr = (attr_ptr as *mut u8).add(std::mem::size_of::<nlattr>()) as *mut u32;
+            *ptr = max_size;
+        }
+
+        #[allow(clippy::arithmetic_side_effects)]
+        #[allow(clippy::cast_lossless)]
+        #[allow(clippy::cast_sign_loss)]
+        {
+            // SAFETY: NLA_ALIGN ensures proper alignment for netlink attributes
+            // as required by Linux kernel ABI.
+            offset += unsafe { libc::NLA_ALIGN((*attr_ptr).nla_len as libc::c_int) } as usize;
+        }
+    }
+
+    #[allow(clippy::arithmetic_side_effects)]
+    #[allow(clippy::cast_possible_truncation)]
+    {
+        msg.hdr.nlmsg_len += offset as u32;
+    }
+
+    // SAFETY: Cast to byte slice for send.
+    let buf = unsafe {
+        std::slice::from_raw_parts(
+            std::ptr::addr_of!(msg) as *const u8,
+            msg.hdr.nlmsg_len as usize,
+        )
+    };
+    sock.send(buf, 0)
+        .map_err(|e| Errno::from_raw(e.raw_os_error().unwrap_or(libc::ENOSYS)))?;
+
+    // Receive response.
+    let (buf, _) = sock
+        .recv_from_full()
+        .map_err(|e| Errno::from_raw(e.raw_os_error().unwrap_or(libc::ENOSYS)))?;
+
+    // Check response messages for error.
+    let mut offset = 0;
+    #[allow(clippy::arithmetic_side_effects)]
+    #[allow(clippy::cast_ptr_alignment)]
+    while offset + std::mem::size_of::<libc::nlmsghdr>() <= buf.len() {
+        // SAFETY: We're reading a netlink message header from validated bounds.
+        let hdr = unsafe { &*(buf.as_ptr().add(offset) as *const libc::nlmsghdr) };
+
+        let len = hdr.nlmsg_len as usize;
+        if len < std::mem::size_of::<libc::nlmsghdr>() || offset + len > buf.len() {
+            return Err(Errno::EINVAL);
+        }
+
+        // Check for error message.
+        #[allow(clippy::cast_possible_truncation)]
+        if hdr.nlmsg_type == libc::NLMSG_ERROR as libc::c_ushort
+            && len >= std::mem::size_of::<libc::nlmsghdr>() + std::mem::size_of::<libc::nlmsgerr>()
+        {
+            // SAFETY: Enough data to safely parse nlmsgerr.
+            let err = unsafe {
+                &*(buf
+                    .as_ptr()
+                    .add(offset + std::mem::size_of::<libc::nlmsghdr>())
+                    as *const libc::nlmsgerr)
+            };
+            if err.error != 0 {
+                return Err(Errno::from_raw(-err.error));
+            }
+        }
+
+        #[allow(clippy::cast_possible_truncation)]
+        #[allow(clippy::cast_possible_wrap)]
+        #[allow(clippy::cast_sign_loss)]
+        {
+            // SAFETY: nlmsg_len is kernel-aligned; advance to next message.
+            offset += unsafe { libc::NLA_ALIGN(len as i32) as usize };
+        }
+    }
+
+    Ok(())
 }
 
 const DEV_NULL_MAJOR: u32 = 1;
