@@ -12,7 +12,7 @@ use std::{
     time::{SystemTime, UNIX_EPOCH},
 };
 
-use hex::DisplayHex;
+use data_encoding::HEXLOWER;
 use nix::{
     errno::Errno,
     unistd::{fdatasync, write, Pid, Uid},
@@ -270,10 +270,11 @@ pub fn log_init(default_level: LogLevel, default_log_fd: Option<RawFd>) -> Resul
     // Decide if we want a TTY-like console.
     let mut tty = std::env::var_os(ENV_FORCE_TTY).is_some();
     if !tty {
-        if fd == Some(libc::STDERR_FILENO) && std::env::var_os(ENV_QUIET_TTY).is_none() {
-            // check if we have a TTY
-            tty = nix::unistd::isatty(std::io::stdin()).unwrap_or(false)
-                && nix::unistd::isatty(std::io::stderr()).unwrap_or(false);
+        if std::env::var_os(ENV_QUIET_TTY).is_none() {
+            let fd = fd.unwrap_or(libc::STDERR_FILENO);
+            // Check if log FD is a TTY.
+            // SAFETY: In libc we trust.
+            tty = unsafe { libc::isatty(fd) } == 1;
         } else {
             tty = false;
         }
@@ -338,10 +339,10 @@ pub fn log_init_simple(default_level: LogLevel) -> Result<(), Errno> {
     // Decide if we want a TTY-like console.
     let mut tty = std::env::var_os(ENV_FORCE_TTY).is_some();
     if !tty {
-        if fd == libc::STDERR_FILENO && std::env::var_os(ENV_QUIET_TTY).is_none() {
-            // check if we have a TTY
-            tty = nix::unistd::isatty(std::io::stdin()).unwrap_or(false)
-                && nix::unistd::isatty(std::io::stderr()).unwrap_or(false);
+        if std::env::var_os(ENV_QUIET_TTY).is_none() {
+            // Check if log FD is a TTY.
+            // SAFETY: In libc we trust.
+            tty = unsafe { libc::isatty(fd) } == 1;
         } else {
             tty = false;
         }
@@ -350,18 +351,6 @@ pub fn log_init_simple(default_level: LogLevel) -> Result<(), Errno> {
     // Record TTY information to an atomic for easy access.
     LOG_TTY.store(tty, std::sync::atomic::Ordering::Relaxed);
 
-    // We'll do the same FD detection.
-    let mut tty = std::env::var_os(ENV_FORCE_TTY).is_some();
-    if !tty {
-        if std::env::var_os(ENV_QUIET_TTY).is_none() {
-            tty = nix::unistd::isatty(std::io::stdin()).unwrap_or(false)
-                && nix::unistd::isatty(std::io::stderr()).unwrap_or(false);
-        } else {
-            tty = false;
-        }
-    }
-    LOG_TTY.store(tty, std::sync::atomic::Ordering::Relaxed);
-
     // Create a global syslog with ring.
     init_global_syslog(0, Some(fd), level, host_syslog, true)?;
 
@@ -460,16 +449,30 @@ pub fn log(level: crate::syslog::LogLevel, timestamp: u64, mut msg: Map<String,
 /// Initializes Syslog global object, reading environment variables.
 #[cfg(not(feature = "log"))]
 #[inline(always)]
+#[allow(clippy::cognitive_complexity)]
 pub fn log_init(_default_level: LogLevel, default_log_fd: Option<RawFd>) -> Result<(), Errno> {
     use std::os::unix::ffi::OsStrExt;
 
+    // Determine the main FD for logging:
+    // By default, we use stderr. If ENV_LOG_FD is set, we parse it.
+    let fd = match std::env::var_os(ENV_LOG_FD) {
+        None => default_log_fd,
+        Some(val) => {
+            Some(btoi::btoi::<RawFd>(val.as_os_str().as_bytes()).map_err(|_| Errno::EBADF)?)
+        }
+    };
+    if let Some(fd) = fd {
+        LOG_FD.store(fd, std::sync::atomic::Ordering::Relaxed);
+    }
+
     // Decide if we want a TTY-like console.
     let mut tty = std::env::var_os(ENV_FORCE_TTY).is_some();
     if !tty {
-        if std::env::var_os(ENV_LOG_FD).is_none() && std::env::var_os(ENV_QUIET_TTY).is_none() {
-            // check if we have a TTY
-            tty = nix::unistd::isatty(std::io::stdin()).unwrap_or(false)
-                && nix::unistd::isatty(std::io::stderr()).unwrap_or(false);
+        if std::env::var_os(ENV_QUIET_TTY).is_none() {
+            let fd = fd.unwrap_or(libc::STDERR_FILENO);
+            // Check if log FD is a TTY.
+            // SAFETY: In libc we trust.
+            tty = unsafe { libc::isatty(fd) } == 1;
         } else {
             tty = false;
         }
@@ -508,9 +511,9 @@ pub fn log_init_simple(default_level: LogLevel) -> Result<(), Errno> {
 #[cfg(not(feature = "log"))]
 #[allow(clippy::cognitive_complexity)]
 pub fn log(level: crate::syslog::LogLevel, timestamp: u64, mut msg: Map<String, Value>) {
-    // For "higher" severities, we add more contextual fields, and honour TTY.
+    // For "higher" severities, we add more contextual fields.
     let add_context = level.as_u8() <= crate::syslog::LogLevel::Notice.as_u8();
-    let tty = add_context && LOG_TTY.load(std::sync::atomic::Ordering::Relaxed);
+    let tty = LOG_TTY.load(std::sync::atomic::Ordering::Relaxed);
 
     // If there's a "pid", we might add cmd/cwd, etc.
     // We remove and readd to reorder for better visibility.
@@ -645,11 +648,11 @@ fn format_iso8601(timestamp: u64) -> String {
 /// Returns a boolean in addition to the String which is true if String is hex-encoded.
 pub fn log_untrusted_buf(buf: &[u8]) -> (String, bool) {
     if contains_ascii_unprintable(buf) {
-        (buf.to_lower_hex_string(), true)
+        (HEXLOWER.encode(buf), true)
     } else if let Ok(s) = std::str::from_utf8(buf) {
         (s.to_string(), false)
     } else {
-        (buf.to_lower_hex_string(), true)
+        (HEXLOWER.encode(buf), true)
     }
 }
 
