@@ -1077,6 +1077,18 @@ impl XPathBuf {
         buf.format(fd).as_bytes().into()
     }
 
+    /// Create a path for the given self-FD.
+    ///
+    /// Used for _procfs_(5) indirection.
+    pub fn from_self_fd(fd: RawFd) -> Self {
+        // SAFETY:
+        // Use /proc/thread-self rather than /proc/self
+        // because CLONE_FILES may be in effect!
+        let mut pfd = Self::from("thread-self/fd");
+        pfd.push_fd(fd);
+        pfd
+    }
+
     /// Append the formatted FD as a new component.
     pub fn push_pid(&mut self, pid: Pid) {
         let mut buf = itoa::Buffer::new();
@@ -1232,6 +1244,41 @@ impl XPathBuf {
     pub fn is_file(&self) -> bool {
         self.as_path().is_file()
     }
+
+    /// Returns a slice containing the entire path buffer.
+    #[inline]
+    pub fn as_slice(&self) -> &[u8] {
+        self.0.as_slice()
+    }
+
+    /// Returns a mutable slice containing the entire path buffer.
+    #[inline]
+    pub fn as_mut_slice(&mut self) -> &mut [u8] {
+        self.0.as_mut_slice()
+    }
+
+    /// Returns a pointer to the internal `Vec`.
+    #[inline]
+    pub fn as_ptr(&self) -> *const u8 {
+        self.0.as_ptr()
+    }
+
+    /// Returns a mutable pointer to the internal `Vec`.
+    #[inline]
+    pub fn as_mut_ptr(&mut self) -> *mut u8 {
+        self.0.as_mut_ptr()
+    }
+
+    /// Forces the length of the internal `Vec`.
+    ///
+    /// # Safety
+    ///
+    /// - `new_len` must be less than or equal to `capacity()`.
+    /// - The elements at `old_len..new_len` must be initialized.
+    #[inline]
+    pub unsafe fn set_len(&mut self, new_len: usize) {
+        self.0.set_len(new_len)
+    }
 }
 
 /// Logs an untrusted Path, escaping it as hex if it contains control
