@@ -17,6 +17,7 @@ use std::{
     collections::HashMap,
     env,
     fs::OpenOptions,
+    io::Write,
     option::Option,
     os::{
         fd::{AsRawFd, BorrowedFd, RawFd},
@@ -27,34 +28,36 @@ use std::{
         Arc, RwLock,
     },
     thread,
-    thread::JoinHandle,
 };
 
 use nix::{
     errno::Errno,
-    sched::{sched_setaffinity, CpuSet},
+    fcntl::OFlag,
+    sched::{unshare, CloneFlags},
     sys::{
         epoll::Epoll,
         signal::{sigaction, SaFlags, SigAction, SigHandler, SigSet, Signal},
     },
-    unistd::{getpid, Pid},
+    unistd::{getpid, pipe2, read},
 };
 
 use crate::{
+    alert,
     compat::epoll_ctl_safe,
     config::*,
-    err::{err2no, SydResult},
+    crit,
+    err::{err2no, SydJoinHandle, SydResult},
     error,
-    fs::{retry_on_eintr, seccomp_export_pfc},
+    fs::{block_signal, closeexcept, retry_on_eintr, retry_on_intr, seccomp_export_pfc},
     hook::HandlerMap,
-    info,
+    info, rwrite, rwriteln,
     sandbox::{Capability, Flags, Sandbox},
     workers::{
         aes::{AesMap, AesWorker},
         emu::Worker,
         int::Interrupter,
         ipc::IpcWorker,
-        BindMap, PidFdMap, WorkerCache, WorkerData,
+        BindMap, WorkerCache, WorkerData,
     },
     ExportMode,
 };
@@ -86,7 +89,7 @@ pub(crate) struct ThreadPool {
     core_size: usize,
     keep_alive: u16,
     flags: Flags,
-    fd: RawFd,
+    seccomp_fd: RawFd,
     pub(crate) epoll: Arc<Epoll>,
     pub(crate) cache: Arc<WorkerCache<'static>>,
     sandbox: Arc<RwLock<Sandbox>>,
@@ -113,27 +116,23 @@ impl ThreadPool {
     #[allow(clippy::too_many_arguments)]
     pub(crate) fn new(
         epoll: Epoll,
-        fd: RawFd,
+        seccomp_fd: RawFd,
         flags: Flags,
         core_size: usize,
         keep_alive: u16,
         sandbox: Arc<RwLock<Sandbox>>,
         handlers: Arc<HandlerMap>,
+        should_exit: Arc<AtomicBool>,
         crypt_map: Option<AesMap>,
     ) -> Result<Self, Errno> {
         // SAFETY: Borrow FD to make I/O safe API hippie.
-        let seccomp_fd = unsafe { BorrowedFd::borrow_raw(fd) };
+        let seccomp_fd = unsafe { BorrowedFd::borrow_raw(seccomp_fd) };
         epoll_ctl_safe(&epoll.0, seccomp_fd.as_raw_fd(), Some(SCMP_EPOLL_EVENT))?;
 
         let epoll = Arc::new(epoll);
-        let cache = Arc::new(WorkerCache::new(Arc::clone(&epoll), fd));
-
-        // Create pidfd map.
-        let pidfd_map = PidFdMap::new(Arc::clone(&cache));
-        PIDFD_MAP.set(pidfd_map).or(Err(Errno::EAGAIN))?;
+        let cache = Arc::new(WorkerCache::new());
 
         Ok(Self {
-            fd,
             cache,
             sandbox,
             crypt_map,
@@ -142,7 +141,8 @@ impl ThreadPool {
             keep_alive,
             epoll,
             flags,
-            should_exit: Arc::new(AtomicBool::new(false)),
+            should_exit,
+            seccomp_fd: seccomp_fd.as_raw_fd(),
             worker_data: Arc::new(WorkerData::default()),
             bind_map: Arc::new(RwLock::new(HashMap::default())),
         })
@@ -150,7 +150,7 @@ impl ThreadPool {
 
     /// Boot the thread pool. This is the main entry point.
     #[allow(clippy::cognitive_complexity)]
-    pub(crate) fn boot(self) -> SydResult<Option<JoinHandle<()>>> {
+    pub(crate) fn boot(self) -> SydResult<SydJoinHandle<()>> {
         // Export seccomp rules if requested.
         // We have to prepare the filter twice if exporting,
         // as we cannot move it safely between threads...
@@ -158,7 +158,13 @@ impl ThreadPool {
         match ExportMode::from_env() {
             Some(ExportMode::BerkeleyPacketFilter) => {
                 // Worker rules
-                let ctx = Worker::prepare_confine(self.fd, self.epoll.0.as_raw_fd(), self.flags)?;
+                let is_crypt = self.crypt_map.is_some();
+                let ctx = Worker::prepare_confine(
+                    self.seccomp_fd,
+                    self.epoll.0.as_raw_fd(),
+                    self.flags,
+                    is_crypt,
+                )?;
                 let file = OpenOptions::new()
                     .write(true)
                     .create_new(true)
@@ -168,7 +174,8 @@ impl ThreadPool {
 
                 // Interrupter rules
                 // We pass dry_run=true to avoid Landlock confinement.
-                let ctx = Interrupter::prepare_confine(self.fd, getpid(), self.flags, true)?;
+                let ctx =
+                    Interrupter::prepare_confine(self.seccomp_fd, getpid(), self.flags, true)?;
                 let file = OpenOptions::new()
                     .write(true)
                     .create_new(true)
@@ -197,96 +204,61 @@ impl ThreadPool {
                 ctx.export_bpf(file)?;
             }
             Some(ExportMode::PseudoFiltercode) => {
-                println!("# Syd monitor rules with seccomp fd {}", self.fd);
-                let ctx = Worker::prepare_confine(self.fd, self.epoll.0.as_raw_fd(), self.flags)?;
-                print!("{}", seccomp_export_pfc(&ctx)?);
+                // Lock stdout to prevent concurrent access.
+                let mut stdout = std::io::stdout().lock();
+
+                rwriteln!(
+                    stdout,
+                    "# Syd monitor rules with seccomp fd {}",
+                    self.seccomp_fd
+                )?;
+                let is_crypt = self.crypt_map.is_some();
+                let ctx = Worker::prepare_confine(
+                    self.seccomp_fd,
+                    self.epoll.0.as_raw_fd(),
+                    self.flags,
+                    is_crypt,
+                )?;
+                rwrite!(stdout, "{}", seccomp_export_pfc(&ctx)?)?;
 
                 // We pass dry_run=true to avoid Landlock confinement.
-                println!("# Syd interrupter rules with seccomp fd {}", self.fd);
-                let ctx = Interrupter::prepare_confine(self.fd, getpid(), self.flags, true)?;
-                print!("{}", seccomp_export_pfc(&ctx)?);
+                rwriteln!(
+                    stdout,
+                    "# Syd interrupter rules with seccomp fd {}",
+                    self.seccomp_fd
+                )?;
+                let ctx =
+                    Interrupter::prepare_confine(self.seccomp_fd, getpid(), self.flags, true)?;
+                rwrite!(stdout, "{}", seccomp_export_pfc(&ctx)?)?;
 
                 // We pass dummy RawFd=2525 for epoll FD.
                 // We pass dry_run=true to avoid Landlock confinement.
-                println!("# Syd ipc rules");
+                rwriteln!(stdout, "# Syd ipc rules")?;
                 let ctx = IpcWorker::prepare_confine(2525, self.flags, true)?;
-                print!("{}", seccomp_export_pfc(&ctx)?);
+                rwrite!(stdout, "{}", seccomp_export_pfc(&ctx)?)?;
 
-                println!("# Syd encryptor rules");
+                rwriteln!(stdout, "# Syd encryptor rules")?;
                 let ctx = AesWorker::prepare_confine(self.flags)?;
-                print!("{}", seccomp_export_pfc(&ctx)?);
+                rwrite!(stdout, "{}", seccomp_export_pfc(&ctx)?)?;
             }
             _ => {}
         }
 
-        // Spawn the AES thread if encryption is on.
-        let crypt = {
-            let sandbox = self.sandbox.read().unwrap_or_else(|err| err.into_inner());
-            #[allow(clippy::disallowed_methods)]
-            if sandbox.enabled(Capability::CAP_CRYPT) {
-                let crypt_fds = sandbox.crypt_setup().unwrap();
-                let is_mem_fd = sandbox.crypt_tmp.is_none();
-                Some((crypt_fds, is_mem_fd))
-            } else {
-                None
-            }
-        };
-
-        // Note, we spawn the AES thread before CPU pinning intentionally,
-        // so they get to run on whichever CPU.
-        #[allow(clippy::disallowed_methods)]
-        let crypt_handle = if let Some((crypt_fds, is_mem_fd)) = crypt {
-            let crypt_map = self.crypt_map.as_ref().map(Arc::clone).unwrap();
-            Some(
-                self.try_spawn_aes(crypt_fds, crypt_map, is_mem_fd)
-                    .expect("spawn AES encryption thread"),
-            )
-        } else {
-            None
-        };
-
         // Ensure the lazy num_cpus::get is called before
-        // the CPU pinning below as subsequent invocations
-        // is going to return 1.
+        // the CPU pinning otherwise it may report incorrect
+        // value.
         let nproc = *NPROC;
-        info!("ctx": "boot", "op": "pin_main_thread",
+        info!("ctx": "boot", "op": "check_num_cpus",
             "msg": format!("detected {nproc} CPUs on the system"),
             "num_cpus": nproc);
 
-        // Attempt to set thread's CPU affinity mask to 0.
-        // We pin the main, init and monitor threads to CPU:0.
-        // Emulator threads are pinned according to num-cpus.
-        let cpu_id = 0;
-        let mut cpu_set = CpuSet::new();
-        if cpu_set.set(cpu_id).is_ok() {
-            match sched_setaffinity(Pid::from_raw(0), &cpu_set) {
-                Ok(_) => {
-                    info!("ctx": "boot", "op": "pin_main_thread",
-                        "msg": format!("pinned main thread to CPU:{cpu_id}"),
-                        "cpu": cpu_id);
-                }
-                Err(errno) => {
-                    error!("ctx": "boot", "op": "pin_main_thread",
-                        "msg": format!("failed to pin main thread to CPU:{cpu_id}: {errno}"),
-                        "err": errno as i32,
-                        "cpu": cpu_id);
-                }
-            }
-        }
-
-        // Spawn the interrupt thread which will confine itself.
-        self.try_spawn_interrupt()?;
-
         // Spawn the monitor thread which may confine itself, and spawn
         // emulator threads. Note, this will panic if it cannot spawn
         // the initial emulator thread which is going to tear everything
-        // down.
-        self.monitor()?;
-
-        // Return join handle of the encryption thread,
-        // so we can wait for ongoing encryption processes
-        // before exiting the sandbox.
-        Ok(crypt_handle)
+        // down. Return a join handle to the main thread so it can wait
+        // for the monitor thread to gracefully exit which in turn is
+        // going to wait for the AES threads to gracefully exit.
+        self.monitor()
     }
 
     /// Spawn a monitor thread that watches the worker pool busy count,
@@ -295,11 +267,82 @@ impl ThreadPool {
     /// workers by e.g. opening the read end of a FIFO over and over
     /// again.
     #[allow(clippy::cognitive_complexity)]
-    pub(crate) fn monitor(self) -> Result<(), Errno> {
+    pub(crate) fn monitor(self) -> SydResult<SydJoinHandle<()>> {
         thread::Builder::new()
             .name("syd_mon".to_string())
             .stack_size(MON_STACK_SIZE)
             .spawn(move || {
+                // SAFETY: Default panic hook wont work with stdio(3) closed.
+                std::panic::set_hook(Box::new(|info| {
+                    let err = match info.payload().downcast_ref::<&'static str>() {
+                        Some(s) => *s,
+                        None => match info.payload().downcast_ref::<String>() {
+                            Some(s) => &**s,
+                            None => "?",
+                        },
+                    };
+
+                    let file = info.location().map(|l| l.file());
+                    let line = info.location().map(|l| l.line());
+
+                    crit!("ctx": "panic", "op": "panic_monitor_thread",
+                        "msg": err, "file": file, "line": line);
+                }));
+
+                // Get information on encryption from the sandbox.
+                let (crypt, fpid) = {
+                    let sandbox = self.sandbox.read().unwrap_or_else(|err| err.into_inner());
+                    if sandbox.enabled(Capability::CAP_CRYPT) {
+                        (Some((sandbox.crypt_setup()?, sandbox.crypt_tmp)), sandbox.fpid)
+                    } else {
+                        (None, sandbox.fpid)
+                    }
+                };
+
+                // SAFETY: We use exit_group(2) here to bail,
+                // because this unsharing is a critical safety feature.
+                if let Err(errno) = unshare(CloneFlags::CLONE_FS | CloneFlags::CLONE_FILES) {
+                    alert!("ctx": "boot", "op": "unshare_monitor_thread",
+                        "msg": format!("failed to unshare(CLONE_FS|CLONE_FILES): {errno}"),
+                        "err": errno as i32);
+                    std::process::exit(101);
+                }
+
+                // SAFETY: The monitor thread needs to inherit FDs.
+                // We have to sort the set as the FDs are randomized.
+                #[allow(clippy::cast_sign_loss)]
+                let mut set = vec![
+                    ROOT_FD() as libc::c_uint,
+                    PROC_FD() as libc::c_uint,
+                    NULL_FD() as libc::c_uint,
+                    fpid as libc::c_uint,
+                    self.seccomp_fd as libc::c_uint,
+                    self.epoll.0.as_raw_fd() as libc::c_uint,
+                    crate::log::LOG_FD.load(Ordering::Relaxed) as libc::c_uint,
+                ];
+                #[allow(clippy::cast_sign_loss)]
+                if let Some((crypt_fds, crypt_tmp)) = crypt {
+                    set.push(crypt_fds.0 as libc::c_uint);
+                    set.push(crypt_fds.1 as libc::c_uint);
+                    if let Some(crypt_tmp) = crypt_tmp {
+                        set.push(crypt_tmp as libc::c_uint);
+                    }
+                }
+                set.sort_unstable();
+                closeexcept(&set)?;
+
+                // Spawn the interrupt thread which will confine itself.
+                self.try_spawn_interrupt()?;
+
+                // Spawn the AES thread if encryption is on.
+                let crypt_handle = if let Some((fds, tmp)) = crypt {
+                    let map = self.crypt_map.as_ref().map(Arc::clone).ok_or(Errno::ENOKEY)?;
+                    let should_exit = Arc::clone(&self.should_exit);
+                    Some(self.try_spawn_aes(fds, map, tmp.is_none(), should_exit)?)
+                } else {
+                    None
+                };
+
                 info!("ctx": "boot", "op": "start_monitor_thread",
                     "msg": format!("started monitor thread with pool size set to {} threads and keep alive set to {} seconds",
                         self.core_size,
@@ -312,16 +355,15 @@ impl ThreadPool {
                 // spawned threads don't need to reapply the same filter
                 // as it is inherited.
                 let dry_run = env::var_os(ENV_SKIP_SCMP).is_some() || ExportMode::from_env().is_some();
+                let is_crypt = self.crypt_map.is_some();
                 let safe_setid = self.flags.intersects(Flags::FL_ALLOW_SAFE_SETUID | Flags::FL_ALLOW_SAFE_SETGID);
-                #[allow(clippy::disallowed_methods)]
                 let mut confined = if Sandbox::locked_once() {
-                    let ctx = Worker::prepare_confine(self.fd, self.epoll.0.as_raw_fd(), self.flags)
-                        .expect("prepare monitor thread confinement");
+                    let ctx = Worker::prepare_confine(self.seccomp_fd, self.epoll.0.as_raw_fd(), self.flags, is_crypt)?;
                     if !dry_run {
                         info!("ctx": "confine", "op": "confine_monitor_thread",
                             "msg": format!("monitor thread confined with{} SROP mitigation",
                                 if safe_setid { "out" } else { "" }));
-                        ctx.load().expect("confine monitor thread");
+                        ctx.load()?;
                     } else {
                         error!("ctx": "confine", "op": "confine_monitor_thread",
                             "msg": "monitor thread is running unconfined in debug mode");
@@ -340,29 +382,23 @@ impl ThreadPool {
                     "core_size": self.core_size,
                     "keep_alive": self.keep_alive);
 
-                // SAFETY: Panic if we cannot spawn the initial core thread.
-                // There's little sense in continuing in this case.
-                #[allow(clippy::disallowed_methods)]
-                self.try_spawn(confined)
-                    .expect("spawn core emulator thread")
-                    .map(drop)
-                    .unwrap();
+                // Spawn the initial core emulator thread.
+                self.try_spawn(confined).map(drop)?;
 
-                // SAFETY: Wait for grace period to give the initial core emulator
-                // thread a chance to spawn itself.
+                // Wait for grace period to give the initial
+                // core emulator thread a chance to spawn itself.
                 std::thread::sleep(MON_GRACE_TIME);
 
                 loop {
                     // Confine as necessary.
                     #[allow(clippy::disallowed_methods)]
                     if !confined && Sandbox::locked_once() {
-                        let ctx = Worker::prepare_confine(self.fd, self.epoll.0.as_raw_fd(), self.flags)
-                            .expect("prepare monitor thread confinement");
+                        let ctx = Worker::prepare_confine(self.seccomp_fd, self.epoll.0.as_raw_fd(), self.flags, is_crypt)?;
                         if !dry_run {
+                            ctx.load()?;
                             info!("ctx": "confine", "op": "confine_monitor_thread",
                                 "msg": format!("monitor thread confined with{} SROP mitigation",
                                     if safe_setid { "out" } else { "" }));
-                            ctx.load().expect("confine monitor thread");
                         } else {
                             error!("ctx": "confine", "op": "confine_monitor_thread",
                                 "msg": "monitor thread is running unconfined in debug mode");
@@ -394,15 +430,24 @@ impl ThreadPool {
                         }
                     }
                 }
+
+                // Wait for AES threads to gracefully exit.
+                if let Some(crypt_handle) = crypt_handle {
+                    crypt_handle.join().or(Err(Errno::EAGAIN))??;
+                }
+
+                Ok(())
             })
-            .map(drop)
-            .map_err(|err| err2no(&err))
+            .map_err(|err| err2no(&err).into())
     }
 
     /// Spawn an interrupt handler thread to unblock Syd syscall
     /// handler threads when the respective sandbox process
     /// receives a non-restarting signal.
-    pub(crate) fn try_spawn_interrupt(&self) -> Result<JoinHandle<()>, Errno> {
+    pub(crate) fn try_spawn_interrupt(&self) -> SydResult<SydJoinHandle<()>> {
+        // Block SIGALRM, this mask will be inherited by emulators.
+        block_signal(Signal::SIGALRM)?;
+
         // Set up the signal handler for SIGALRM.
         let sig_action = SigAction::new(
             SigHandler::Handler(handle_sigalrm),
@@ -411,17 +456,39 @@ impl ThreadPool {
         );
 
         // SAFETY: Register the handler for SIGALRM.
+        // This handler is per-process.
         unsafe { sigaction(Signal::SIGALRM, &sig_action) }?;
 
-        retry_on_eintr(|| {
+        // Set up a notification pipe and wait for
+        // the interrupt worker to start and unshare CLONE_F{ILE,}S.
+        let (pipe_rd, pipe_wr) = pipe2(OFlag::O_CLOEXEC)?;
+
+        let handle = retry_on_intr(|| {
             Interrupter::new(
-                self.fd,
+                self.seccomp_fd,
                 self.flags,
                 Arc::clone(&self.should_exit),
                 Arc::clone(&self.cache),
             )
-            .try_spawn()
-        })
+            .try_spawn((pipe_rd.as_raw_fd(), pipe_wr.as_raw_fd()))
+        })?;
+
+        // Wait for startup notification.
+        let mut buf = [0u8; 1];
+        match retry_on_eintr(|| read(&pipe_rd, &mut buf[..]))? {
+            0 => {
+                // Interrupt thread died before unshare.
+                // This should ideally never happen.
+                return Err(Errno::EIO.into());
+            }
+            1 if buf[0] == 42 => {
+                // Interrupt thread unshared successfully.
+                // We can go ahead and close our copies now.
+            }
+            _ => unreachable!("BUG: The meaning of life is not {:#x}!", buf[0]),
+        }
+
+        Ok(handle)
     }
 
     /// Try to create a new encryption thread.
@@ -430,24 +497,22 @@ impl ThreadPool {
         fdalg: (RawFd, RawFd),
         files: AesMap,
         memfd: bool,
-    ) -> Result<JoinHandle<()>, Errno> {
-        let worker = AesWorker::new(fdalg, files, self.flags, memfd);
+        should_exit: Arc<AtomicBool>,
+    ) -> Result<SydJoinHandle<()>, Errno> {
+        let worker = AesWorker::new(fdalg, files, self.flags, memfd, should_exit);
 
         // AesWorker has only RawFds as Fds which
         // we do _not_ want to duplicate on clone,
         // so we can get away with a clone here...
-        retry_on_eintr(|| worker.clone().try_spawn())
+        retry_on_intr(|| worker.clone().try_spawn())
     }
 
     /// Try to create a new worker thread as needed.
-    /// Returns Ok(Some((JoinHandle, bool))) if spawn succeeded, Ok(None) if no spawn was needed.
+    /// Returns Ok(Some((SydJoinHandle, bool))) if spawn succeeded, Ok(None) if no spawn was needed.
     /// The boolean in the success case is true if the thread we spawned was a core thread.
     #[allow(clippy::cognitive_complexity)]
     #[allow(clippy::type_complexity)]
-    pub(crate) fn try_spawn(
-        &self,
-        confined: bool,
-    ) -> Result<Option<(JoinHandle<()>, bool)>, Errno> {
+    pub(crate) fn try_spawn(&self, confined: bool) -> SydResult<Option<(SydJoinHandle<()>, bool)>> {
         // Create a new worker if there are no idle threads and the
         // current worker count is lower than the max pool size.
         let worker_count_val = self.worker_data.0.load(Ordering::Relaxed);
@@ -492,9 +557,9 @@ impl ThreadPool {
 
         // Try to spawn a new worker.
         Ok(Some((
-            retry_on_eintr(|| {
+            retry_on_intr(|| {
                 Worker::new(
-                    self.fd,
+                    self.seccomp_fd,
                     Arc::clone(&self.epoll),
                     Arc::clone(&self.cache),
                     Arc::clone(&self.sandbox),
