@@ -87,9 +87,9 @@ use syd::{
     caps,
     compat::{waitid, WaitStatus},
     config::*,
-    err::{err2no, SydResult},
+    err::err2no,
     error,
-    fs::{pidfd_open, retry_on_eintr, set_cloexec},
+    fs::{closeexcept, pidfd_open, retry_on_eintr, set_cloexec},
     hash::SydRandomState,
     hook::Supervisor,
     info,
@@ -108,8 +108,7 @@ use syd::{
 #[global_allocator]
 static GLOBAL: tcmalloc::TCMalloc = tcmalloc::TCMalloc;
 
-#[allow(clippy::cognitive_complexity)]
-fn main() -> SydResult<ExitCode> {
+syd::main! {
     use lexopt::prelude::*;
 
     // Set SIGPIPE handler to default.
@@ -875,20 +874,18 @@ fn main() -> SydResult<ExitCode> {
             drop(cmd_envp);
             drop(pty_child);
             drop(sandbox);
-            // SAFETY: nix does not have a close_range wrapper yet.
-            let _ = unsafe {
-                libc::syscall(
-                    libc::SYS_close_range,
-                    libc::STDERR_FILENO + 1,
-                    libc::c_int::MAX,
-                    0,
-                )
-            };
+
             child
         }
     };
 
     // Only parent ends up here and `child` is child pid.
+    //
+    // Unshare CLONE_FS|CLONE_FILES for isolation.
+    // Close all file descriptors but standard error.
+    unshare(CloneFlags::CLONE_FS | CloneFlags::CLONE_FILES)?;
+    closeexcept(&[libc::STDERR_FILENO as libc::c_uint])?;
+
     // SAFETY: Set up a Landlock sandbox to disallow all access.
     let abi = syd::landlock::ABI::new_current();
     let policy = LandlockPolicy {
