@@ -18,7 +18,8 @@ use std::{
 
 use nix::{
     errno::Errno,
-    sys::stat::Mode,
+    fcntl::AT_FDCWD,
+    sys::stat::{fchmodat, umask, FchmodatFlags, Mode},
     unistd::{chdir, mkdir, mkdtemp},
 };
 use syd::{err::SydResult, syslog::LogLevel, wildmatch::inamematch};
@@ -114,6 +115,9 @@ fn main() -> SydResult<ExitCode> {
     // Initialize logging.
     syd::log::log_init_simple(LogLevel::Warn)?;
 
+    // Set a sane umask.
+    umask(Mode::from_bits_truncate(0o077));
+
     println!("# syd-test: Welcome to the Machine!");
     println!("# usage: syd-test [-hlq] [<name-regex>|<number>|<number>..<number>]..");
 
@@ -158,18 +162,24 @@ fn main() -> SydResult<ExitCode> {
             env::var("SYD_TEST_TMPDIR").unwrap_or(".".to_string())
         );
         match mkdtemp(Path::new(&tmp)) {
-            Ok(path) => match chdir(&path) {
-                Ok(_) => {
-                    println!("# Running tests under '{}'.", path.display());
-                    println!("# Use SYD_TEST_TMPDIR to override.");
-                    env::set_var("HOME", &path);
-                    Some(TempDir::new(path))
+            Ok(path) => {
+                // HOME must be canonicalized!
+                let path = path.canonicalize()?;
+
+                match chdir(&path) {
+                    Ok(_) => {
+                        println!("# Running tests under '{}'.", path.display());
+                        println!("# Use SYD_TEST_TMPDIR to override.");
+                        //Setting HOME here breaks podman!
+                        //env::set_var("HOME", &path);
+                        Some(TempDir::new(path))
+                    }
+                    Err(error) => {
+                        println!("# chdir failed: {error}.");
+                        None
+                    }
                 }
-                Err(error) => {
-                    println!("# chdir failed: {error}.");
-                    None
-                }
-            },
+            }
             Err(errno) => {
                 println!("# mkdtemp failed: {errno}.");
                 None
@@ -328,10 +338,23 @@ fn main() -> SydResult<ExitCode> {
         }
         println!("\x1b[92m*** {status} ***\x1b[0m");
 
-        #[allow(clippy::disallowed_methods)]
+        // Note, the extra chmod is needed to
+        // drop the setgid bits we inherit from
+        // the parent directory, otherwise some
+        // tests might fail, e.g:
+        // https://builds.sr.ht/~alip/job/1543770
         mkdir(name, Mode::from_bits_truncate(0o700))?;
-        #[allow(clippy::disallowed_methods)]
+        fchmodat(
+            AT_FDCWD,
+            name,
+            Mode::from_bits_truncate(0o700),
+            FchmodatFlags::FollowSymlink,
+        )?;
         chdir(name)?;
+        let cwd = std::env::current_dir()?.canonicalize()?;
+        std::env::set_var("PWD", &cwd);
+        std::env::set_var("TMP", &cwd);
+        std::env::set_var("TMPDIR", &cwd);
         let result = test();
         #[allow(clippy::disallowed_methods)]
         chdir("..")?;
