@@ -326,6 +326,19 @@ macro_rules! skip_unless_stdin_is_a_tty {
     }};
 }
 
+#[macro_export]
+macro_rules! skip_unless_stdout_is_a_tty {
+    () => {{
+        if !nix::unistd::isatty(std::io::stdout()).unwrap_or(false) {
+            eprintln!("Test requires STDOUT to be a TTY, skipping!");
+            std::env::set_var("SYD_TEST_SOFT_FAIL", "1");
+            return Ok(());
+        } else {
+            eprintln!("STDOUT is a TTY, proceeding with test...");
+        }
+    }};
+}
+
 #[macro_export]
 macro_rules! skip_unless_kernel_crypto_is_supported {
     () => {{
@@ -640,6 +653,12 @@ pub static SYD_DO: Lazy<String> = Lazy::new(|| match env::var("CARGO_BIN_EXE_syd
 pub static SYD_X: Lazy<String> =
     Lazy::new(|| env::var("CARGO_BIN_EXE_syd-x").unwrap_or("syd-x".to_string()));
 
+#[cfg(feature = "oci")]
+/// Holds a `String` to run `syd-oci`.
+/// Honours CARGO_BIN_EXE_syd-oci environment variable.
+pub static SYD_OCI: Lazy<String> =
+    Lazy::new(|| env::var("CARGO_BIN_EXE_syd-oci").unwrap_or("syd-oci".to_string()));
+
 /// A boolean which specifies if we're running under SourceHut CI.
 pub static CI_BUILD: Lazy<bool> = Lazy::new(|| env::var("JOB_ID").ok().is_some());
 
