@@ -77,6 +77,9 @@ pub extern "C" fn child_after_clone(arg: *mut libc::c_void) -> libc::c_int {
 
     // Restriction 0: Change controlling terminal to PTY as necessary.
     if let Some(pty_fd) = child.pty_fd.take() {
+        // SAFETY: pty_fd is a valid FD.
+        let pty_fd = unsafe { OwnedFd::from_raw_fd(pty_fd) };
+
         // Become session leader so we can take a controlling TTY.
         if let Err(errno) = setsid() {
             fail_errno_safe!(child, Err::SetSid, errno as i32);
@@ -150,11 +153,12 @@ pub extern "C" fn child_after_clone(arg: *mut libc::c_void) -> libc::c_int {
 
             seccomp_add_architectures(&mut ctx)?;
 
+            #[allow(clippy::useless_conversion)]
             for request in SECCOMP_IOCTL_NOTIF_LIST {
                 ctx.add_rule_conditional(
                     ScmpAction::KillProcess,
                     syscall,
-                    &[scmp_cmp!($arg1 == *request)],
+                    &[scmp_cmp!($arg1 == (*request).into())],
                 )?;
             }
 
@@ -192,7 +196,9 @@ pub extern "C" fn child_after_clone(arg: *mut libc::c_void) -> libc::c_int {
     // SAFETY: Clean Syd environment variables from process environment.
     // Note, we have just used ENV_LOG_FD above and do not need it anymore.
     for (key, _) in env::vars_os() {
-        if key.as_bytes().starts_with(b"SYD_") && !key.as_bytes().starts_with(b"SYD_TEST_") {
+        if key.as_bytes().starts_with(b"CARGO_BIN_EXE_syd")
+            || (key.as_bytes().starts_with(b"SYD_") && !key.as_bytes().starts_with(b"SYD_TEST_"))
+        {
             env::remove_var(key);
         }
     }
