@@ -1,4 +1,4 @@
-use std::{ffi::OsStr, os::fd::OwnedFd};
+use std::{ffi::OsStr, os::fd::RawFd};
 
 use nix::sys::signal::Signal;
 
@@ -105,7 +105,7 @@ impl Command {
     }
 
     /// Specify pseudoterminal file descriptor.
-    pub fn pty(&mut self, fd: Option<OwnedFd>) -> &mut Command {
+    pub fn pty(&mut self, fd: Option<RawFd>) -> &mut Command {
         self.pty_fd = fd;
         self
     }
