@@ -42,7 +42,7 @@ pub struct ChildInfo<'a> {
     pub exe: Exe<'a>,
     pub cfg: Config,
     pub pre_exec: Option<ChildPreExecFunc>,
-    pub pty_fd: Option<OwnedFd>,
+    pub pty_fd: Option<RawFd>,
     pub seccomp_filter: Option<ScmpFilterContext>,
     pub seccomp_pipefd: PipePair,
 }
@@ -130,7 +130,7 @@ impl Command {
             Ordering::Greater => {
                 let child = Pid::from_raw(child);
 
-                let seccomp_fd = match self.after_start(child, &pid_fd) {
+                let seccomp_fd = match self.after_start(child, pid_fd) {
                     Ok(seccomp_fd) => seccomp_fd,
                     Err(e) => loop {
                         match waitid(Id::PIDFd(pid_fd.as_fd()), WaitPidFlag::WEXITED) {
@@ -155,7 +155,7 @@ impl Command {
     }
 
     #[allow(clippy::cognitive_complexity)]
-    fn after_start<F: AsRawFd>(mut self, pid: Pid, pid_fd: &F) -> Result<RawFd, Errno> {
+    fn after_start<Fd: AsFd>(mut self, pid: Pid, pid_fd: Fd) -> Result<RawFd, Errno> {
         if self.config.stop {
             // Seize the process for tracing.
             // This must happen before reading the seccomp fd.
@@ -169,16 +169,12 @@ impl Command {
                 | Options::PTRACE_O_TRACESYSGOOD  // ditto.
                 | Options::PTRACE_O_EXITKILL; // we also set PDEATHSIG so this is the second layer.
 
-            // SAFETY: Prefer to use the PIDFd rather than the PID
-            // for waitid(2) calls. This gives us safety against
-            // e.g. PID recycling and ensures a secure attach process.
-            let pid_fd = unsafe { BorrowedFd::borrow_raw(pid_fd.as_raw_fd()) };
             // Step 1: Wait for the process to stop itself.
             // Note, we also wait for EXITED so that if the process is
             // interrupted, and the wait will fall through to the assert
             // to fail.
             let status = waitid(
-                Id::PIDFd(pid_fd),
+                Id::PIDFd(pid_fd.as_fd()),
                 WaitPidFlag::WEXITED | WaitPidFlag::WSTOPPED | WaitPidFlag::__WNOTHREAD,
             )?;
             assert_eq!(status, WaitStatus::Stopped(pid, libc::SIGSTOP));
@@ -190,7 +186,7 @@ impl Command {
             // e.g. trivial ptrace detectors.
             seize(pid, ptrace_options)?;
             let status = waitid(
-                Id::PIDFd(pid_fd),
+                Id::PIDFd(pid_fd.as_fd()),
                 WaitPidFlag::WEXITED | WaitPidFlag::WSTOPPED | WaitPidFlag::__WNOTHREAD,
             )?;
             assert_eq!(
@@ -204,7 +200,7 @@ impl Command {
             // it's done once and it's worth the trouble.
             kill(pid, Signal::SIGCONT)?;
             let status = waitid(
-                Id::PIDFd(pid_fd),
+                Id::PIDFd(pid_fd.as_fd()),
                 WaitPidFlag::WEXITED | WaitPidFlag::WSTOPPED | WaitPidFlag::__WNOTHREAD,
             )?;
             assert_eq!(
@@ -213,7 +209,7 @@ impl Command {
             );
             cont(pid, None)?;
             let status = waitid(
-                Id::PIDFd(pid_fd),
+                Id::PIDFd(pid_fd.as_fd()),
                 WaitPidFlag::WEXITED | WaitPidFlag::WSTOPPED | WaitPidFlag::__WNOTHREAD,
             )?;
             assert_eq!(status, WaitStatus::PtraceEvent(pid, libc::SIGCONT, 0));
@@ -233,21 +229,10 @@ impl Command {
             }
         }
 
-        // SAFETY:
-        // 1. Make Syd process the leader of a new process group
-        //    to isolate it from the sandbox process. This way signals
-        //    sent to sandbox process' process group will not affect
-        //    the Syd process.
-        // 2. This is best-effort because we may not always have the
-        //    permission to do this, e.g. with syd-oci(1), youki
-        //    container runtime invokes the same setpgid(0,0) before us,
-        //    so here we get EPERM.
-        if setpgid(Pid::from_raw(0), Pid::from_raw(0)).is_ok() && stdin_isatty {
-            // SAFETY: Release controlling terminal to ensure
-            // a compromised Syd cannot be used to inject
-            // characters into the controlling terminal.
-            Errno::result(unsafe { libc::ioctl(libc::STDIN_FILENO, libc::TIOCNOTTY) })?;
-        }
+        // SAFETY: Note we don't want to put Syd into a new process
+        // group here so as to defend the parent process against
+        // signals in cases when Landlock protection is not available.
+        // Landlock signal scopes are new in ABI 6, new in Linux>=6.12.
 
         if let Some(ref mut callback) = self.before_unfreeze {
             #[allow(clippy::cast_sign_loss)]
@@ -304,7 +289,7 @@ impl Command {
         // Get the seccomp notify fd using pidfd_getfd(2).
         // The child is waiting on the read end of the pipe,
         // for us to safely transfer the file descriptor.
-        let seccomp_fd = pidfd_getfd(pid_fd.as_raw_fd(), remote_seccomp_fd)?;
+        let seccomp_fd = pidfd_getfd(pid_fd, remote_seccomp_fd)?;
 
         // Unblock the child to safely continue and close
         // their copy of the seccomp notify file descriptor.
