@@ -8,17 +8,17 @@
 
 #![allow(non_snake_case)]
 
-use std::{fs::File, io::BufReader};
+use std::{fs::File, io::BufReader, process::ExitCode};
 
 use nix::errno::Errno;
 use syd::{
-    config::API_VERSION,
+    config::API_MAJOR_VERSION,
     err::SydResult,
     path::{XPath, XPathBuf},
     sandbox::Sandbox,
 };
 
-fn main() -> SydResult<()> {
+syd::main! {
     use lexopt::prelude::*;
 
     syd::set_sigpipe_dfl()?;
@@ -42,7 +42,7 @@ fn main() -> SydResult<()> {
         match arg {
             Short('h') => {
                 help();
-                return Ok(());
+                return Ok(ExitCode::SUCCESS);
             }
             Short('j') => optj = true,
             Short('J') => optJ = true,
@@ -77,7 +77,7 @@ fn main() -> SydResult<()> {
         } else {
             dump(&name)?;
         }
-        return Ok(());
+        return Ok(ExitCode::SUCCESS);
     }
 
     for path in paths {
@@ -87,7 +87,7 @@ fn main() -> SydResult<()> {
             return Err(Errno::EOPNOTSUPP.into());
         };
 
-        let syd_ext = XPathBuf::from(format!("syd-{API_VERSION}"));
+        let syd_ext = XPathBuf::from(format!("syd-{API_MAJOR_VERSION}"));
         let ips_ext = XPath::from_bytes(b"ipset");
         let net_ext = XPath::from_bytes(b"netset");
 
@@ -122,7 +122,7 @@ fn main() -> SydResult<()> {
         print!("{syd}");
     }
 
-    Ok(())
+    Ok(ExitCode::SUCCESS)
 }
 
 fn help() {
@@ -138,7 +138,7 @@ fn help() {
     println!("Supported configuration file extensions:");
     println!("    - ipset");
     println!("    - netset");
-    println!("    - syd-{API_VERSION}");
+    println!("    - syd-{API_MAJOR_VERSION}");
 }
 
 fn list() {
@@ -158,6 +158,7 @@ fn list() {
     println!("oci");
     println!("paludis");
     println!("quiet");
+    println!("rand");
     println!("readonly");
     println!("tty");
     println!("user");
@@ -330,6 +331,15 @@ fn dump(name: &str) -> SydResult<()> {
                 println!("{command}");
             }
         }
+        "rand" => {
+            println!("# Syd profile: Rand");
+            println!("# Number of rules: {}", syd::config::PROFILE_RAND.len());
+            println!("# Copyright (c) 2025 Ali Polatel <alip@chesswob.org>");
+            println!("# SPDX-License-Identifier: GPL-3.0");
+            for command in syd::config::PROFILE_RAND {
+                println!("{command}");
+            }
+        }
         "ro" | "readonly" => {
             println!("# Syd profile: Read Only");
             println!("# Number of rules: {}", syd::config::PROFILE_READONLY.len());
