@@ -8,9 +8,7 @@
 
 use std::process::ExitCode;
 
-use syd::err::SydResult;
-
-fn main() -> SydResult<ExitCode> {
+syd::main! {
     use lexopt::prelude::*;
 
     syd::set_sigpipe_dfl()?;
