@@ -16,7 +16,6 @@ use std::{
 use memchr::arch::all::is_equal;
 use nix::errno::Errno;
 use syd::{
-    err::SydResult,
     hash::SydRandomState,
     landlock::{AccessFs, AccessNet, CompatLevel, RulesetStatus, ABI},
     landlock_policy::LandlockPolicy,
@@ -25,7 +24,7 @@ use syd::{
     path::{XPath, XPathBuf},
 };
 
-fn main() -> SydResult<ExitCode> {
+syd::main! {
     use lexopt::prelude::*;
 
     syd::set_sigpipe_dfl()?;
