@@ -1,7 +1,8 @@
 //
 // Syd: rock-solid application kernel
 // src/syd-ls.rs: Print the names of the system calls which belong to the given set and exit
-//                  If set is prctl, print the list of allowed prctl options
+//                If set is prctl, print the list of allowed prctl options
+//                If set is personality, print the list of allowed personalities.
 //
 // Copyright (c) 2024, 2025 Ali Polatel <alip@chesswob.org>
 //
@@ -17,7 +18,7 @@ use std::{
     time::Instant,
 };
 
-use hex::DisplayHex;
+use data_encoding::HEXLOWER;
 use libseccomp::ScmpSyscall;
 use nix::{errno::Errno, unistd::isatty};
 use syd::{
@@ -25,7 +26,7 @@ use syd::{
     path::mask_path,
 };
 
-fn main() -> SydResult<ExitCode> {
+syd::main! {
     syd::set_sigpipe_dfl()?;
 
     let mut args = std::env::args();
@@ -42,9 +43,9 @@ fn main() -> SydResult<ExitCode> {
             println!("If set is drop, print the list of capabilities that are dropped at startup.");
             println!("If set is env, print the list of unsafe environment variables.");
             println!("If set is prctl, print the list of allowed prctl options.");
+            println!("If set is personality, print the list of allowed personalities.");
             println!("Available sets are:");
             println!("- cpu");
-            println!("- dead");
             println!("- deny");
             println!("- ebpf");
             println!("- futex");
@@ -67,11 +68,6 @@ fn main() -> SydResult<ExitCode> {
             println!("File names are printed hex-encoded, delimited by newline, use syd-hex(1) to decode.");
             println!("See EXAMPLES section in syd-ls(1) manual page.");
         }
-        Some("dead") => {
-            for name in syd::config::DEAD_SYSCALLS {
-                println!("{name}");
-            }
-        }
         Some("deny") => {
             let mut syscall_set: HashSet<_, SydRandomState> = syd::config::SAFE_SYSCALLS
                 .iter()
@@ -180,8 +176,13 @@ fn main() -> SydResult<ExitCode> {
                 println!("{env}");
             }
         }
+        Some("personality") => {
+            for (name, _) in syd::config::ALLOWLIST_PERSONALITY {
+                println!("{name}");
+            }
+        }
         Some("prctl") => {
-            for (_, name) in syd::config::ALLOWLIST_PRCTL {
+            for (name, _) in syd::config::ALLOWLIST_PRCTL {
                 println!("{name}");
             }
         }
@@ -219,7 +220,7 @@ fn readdir_cwd() -> SydResult<()> {
 
         for entry in &mut entries {
             // SAFETY: Hex-encode filename to mitigate terminal vulnerabilities.
-            let name = entry.name_bytes().to_lower_hex_string();
+            let name = HEXLOWER.encode(entry.name_bytes());
             println!("{name}");
 
             if report_progress {
