@@ -39,7 +39,10 @@ use libcontainer::{
     process,
     process::{args::ContainerArgs, intel_rdt::delete_resctrl_subdirectory},
     signal::Signal,
-    syscall::{linux::MountOption, syscall::SyscallType},
+    syscall::{
+        linux::MountOption,
+        syscall::{create_syscall, SyscallType},
+    },
     tty,
     user_ns::UserNamespaceConfig,
     utils,
@@ -63,13 +66,13 @@ use nix::{
     unistd::{pipe2, read, Gid, Pid, Uid},
 };
 use oci_spec::runtime::{
-    version, ApparmorBuilder, Capabilities as SpecCapabilities, Capability, CgroupBuilder,
-    FeaturesBuilder, IDMapBuilder, IntelRdtBuilder, LinuxBuilder, LinuxCapabilities,
-    LinuxCapabilitiesBuilder, LinuxFeatureBuilder, LinuxIdMappingBuilder, LinuxNamespace,
-    LinuxNamespaceBuilder, LinuxNamespaceType, LinuxPidsBuilder, LinuxResources,
-    LinuxResourcesBuilder, LinuxSchedulerPolicy, LinuxSeccompAction, LinuxSeccompBuilder,
+    ApparmorBuilder, Capabilities as SpecCapabilities, Capability, CgroupBuilder, FeaturesBuilder,
+    IDMapBuilder, IntelRdtBuilder, LinuxBuilder, LinuxCapabilities, LinuxCapabilitiesBuilder,
+    LinuxFeatureBuilder, LinuxIdMappingBuilder, LinuxNamespace, LinuxNamespaceBuilder,
+    LinuxNamespaceType, LinuxPidsBuilder, LinuxResources, LinuxResourcesBuilder,
+    LinuxSchedulerPolicy, LinuxSeccompAction, LinuxSeccompBuilder, LinuxSyscall,
     LinuxSyscallBuilder, Mount, MountExtensionsBuilder, Process, ProcessBuilder, SelinuxBuilder,
-    Spec,
+    Spec, VERSION,
 };
 use procfs::process::Namespace;
 use serde_json::to_writer_pretty;
@@ -1056,9 +1059,13 @@ impl SydInitContainerBuilder {
             return Err(ErrInvalidSpec::UnsupportedVersion)?;
         };
 
-        let is_rootless_required = rootless_required().map_err(LibcontainerError::OtherIO)?;
+        let syscall = create_syscall();
+        let is_rootless_required =
+            rootless_required(&*syscall).map_err(LibcontainerError::OtherIO)?;
         #[allow(clippy::disallowed_methods)]
-        let syd_dir = if !is_rootless_required {
+        let syd_dir = if env::var_os(syd::config::ENV_OCI_NO_CONFIG).is_some() {
+            None
+        } else if !is_rootless_required {
             Some(String::from("/etc/syd/oci"))
         } else if let Ok(path) = env::var("XDG_CONFIG_HOME") {
             Some(format!("{path}/syd/oci"))
@@ -1468,7 +1475,7 @@ struct Opts {
     subcmd: SubCommand,
 }
 
-fn main() -> SydResult<ExitCode> {
+syd::main! {
     // Guard against CVE-2019-5736:
     // Copy /proc/self/exe in an anonymous fd (created via memfd_create), seal it and re-execute it.
     // See:
@@ -1647,7 +1654,7 @@ fn cmd_features(_opt: GlobalOpts, _args: Features) -> SydResult<ExitCode> {
         .unwrap();
 
     let features = FeaturesBuilder::default()
-        .oci_version_max(version())
+        .oci_version_max(VERSION)
         .oci_version_min(String::from("1.0.0"))
         .hooks(known_hooks())
         .mount_options(MountOption::known_options())
@@ -1976,24 +1983,45 @@ fn syd_spec_load<P: AsRef<Path>>(config: P) -> Result<Spec, LibcontainerError> {
                 serde_json::to_string(&seccomp).unwrap_or("?".to_string())
             );
 
+            let mut syscalls = if let Some(syscalls) = seccomp.syscalls() {
+                syscalls
+                    .iter()
+                    .cloned()
+                    .map(|mut entry| {
+                        // Strip out our OCI_SYSCALLS from non-allow entries.
+                        if entry.action() != LinuxSeccompAction::ScmpActAllow
+                            || entry.errno_ret().is_some()
+                        {
+                            let filtered = entry
+                                .names()
+                                .iter()
+                                .filter(|n| {
+                                    syd::config::OCI_SYSCALLS
+                                        .binary_search(&n.as_str())
+                                        .is_err()
+                                })
+                                .cloned()
+                                .collect::<Vec<String>>();
+                            entry.set_names(filtered);
+                        }
+                        entry
+                    })
+                    .collect::<Vec<LinuxSyscall>>()
+            } else {
+                Vec::new()
+            };
+
+            // Explicitly allow the Syd OCI syscall list.
             let sydallowlist = LinuxSyscallBuilder::default()
                 .action(LinuxSeccompAction::ScmpActAllow)
-                .names(vec![
-                    "pidfd_getfd".to_string(),
-                    "process_vm_readv".to_string(),
-                    "process_vm_writev".to_string(),
-                    "ptrace".to_string(),
-                    "syslog".to_string(),
-                    "unshare".to_string(),
-                ])
+                .names(
+                    syd::config::OCI_SYSCALLS
+                        .iter()
+                        .map(ToString::to_string)
+                        .collect::<Vec<_>>(),
+                )
                 .build()?;
-            let syscalls = if let Some(syscalls) = seccomp.syscalls() {
-                let mut syscalls = syscalls.clone();
-                syscalls.push(sydallowlist);
-                syscalls
-            } else {
-                vec![sydallowlist]
-            };
+            syscalls.push(sydallowlist);
 
             let mut builder = LinuxSeccompBuilder::default()
                 .default_action(seccomp.default_action())
@@ -2028,46 +2056,62 @@ fn syd_spec_load<P: AsRef<Path>>(config: P) -> Result<Spec, LibcontainerError> {
 
     if let Some(process) = spec.process() {
         // Syd does not require CAP_SYS_PTRACE for rootless containers.
-        if rootless_required().map_err(LibcontainerError::OtherIO)? {
+        let syscall = create_syscall();
+        if rootless_required(&*syscall).map_err(LibcontainerError::OtherIO)? {
             return Ok(spec);
         }
 
         let mut p = process.clone();
         if let Some(capabilities) = process.capabilities() {
-            let bounding = if let Some(caps) = capabilities.bounding() {
-                let mut caps = caps.clone();
-                caps.insert(Capability::SysPtrace);
-                caps
+            let mut caps = LinuxCapabilitiesBuilder::default();
+
+            if let Some(c) = capabilities.bounding() {
+                let mut c = c.clone();
+                c.insert(Capability::SysPtrace);
+                caps = caps.bounding(c);
             } else {
-                HashSet::from([Capability::SysPtrace])
-            };
-            let effective = if let Some(caps) = capabilities.effective() {
-                let mut caps = caps.clone();
-                caps.insert(Capability::SysPtrace);
-                caps
+                caps = caps.bounding(HashSet::from([Capability::SysPtrace]));
+            }
+
+            if let Some(c) = capabilities.effective() {
+                let mut c = c.clone();
+                c.insert(Capability::SysPtrace);
+                caps = caps.effective(c);
             } else {
-                HashSet::from([Capability::SysPtrace])
-            };
-            let permitted = if let Some(caps) = capabilities.permitted() {
-                let mut caps = caps.clone();
-                caps.insert(Capability::SysPtrace);
-                caps
+                caps = caps.effective(HashSet::from([Capability::SysPtrace]));
+            }
+
+            if let Some(c) = capabilities.permitted() {
+                let mut c = c.clone();
+                c.insert(Capability::SysPtrace);
+                caps = caps.permitted(c);
             } else {
-                HashSet::from([Capability::SysPtrace])
-            };
+                caps = caps.permitted(HashSet::from([Capability::SysPtrace]));
+            }
+
             // SAFETY: Inheritable and Ambient are not necessary.
-            let caps = LinuxCapabilitiesBuilder::default()
-                .bounding(bounding)
-                .effective(effective)
-                .permitted(permitted)
-                .build()?;
+            if let Some(c) = capabilities.inheritable() {
+                caps = caps.inheritable(c.clone());
+            } else {
+                caps = caps.inheritable(HashSet::new());
+            }
+
+            if let Some(c) = capabilities.ambient() {
+                caps = caps.ambient(c.clone());
+            } else {
+                caps = caps.ambient(HashSet::new());
+            }
+
+            let caps = caps.build()?;
             p.set_capabilities(Some(caps));
         } else {
-            // SAFETY: Inheritable and Ambient are not necessary.
             let caps = LinuxCapabilitiesBuilder::default()
                 .bounding(HashSet::from([Capability::SysPtrace]))
                 .effective(HashSet::from([Capability::SysPtrace]))
                 .permitted(HashSet::from([Capability::SysPtrace]))
+                // SAFETY: Inheritable and Ambient are not necessary.
+                .inheritable(HashSet::new())
+                .ambient(HashSet::new())
                 .build()?;
             p.set_capabilities(Some(caps));
         }
@@ -2222,7 +2266,8 @@ fn make_root(opt: &mut GlobalOpts) -> SydResult<()> {
     // Determine root path.
     #[allow(clippy::disallowed_methods)]
     if opt.root.is_none() {
-        let is_rootless_required = rootless_required()?;
+        let syscall = create_syscall();
+        let is_rootless_required = rootless_required(&*syscall)?;
         opt.root = Some(if !is_rootless_required {
             PathBuf::from("/run/syd")
         } else if let Ok(path) = env::var("XDG_RUNTIME_DIR") {
