@@ -30,7 +30,7 @@ use nix::{
 use syd::{
     compat::epoll_ctl_safe,
     err::SydResult,
-    fs::{duprand, set_nonblock, winsize_get, winsize_set},
+    fs::{closeexcept, duprand, set_exclusive, set_nonblock, winsize_get, winsize_set},
     landlock_policy::LandlockPolicy,
 };
 
@@ -41,8 +41,7 @@ const N_TTY_BUF_SIZE: usize = 4096;
 // syd::config::PIPE_BUF may be larger...
 const PIPE_BUF: usize = N_TTY_BUF_SIZE;
 
-#[allow(clippy::disallowed_methods)]
-fn main() -> SydResult<ExitCode> {
+syd::main! {
     // Parse options.
     let opts = parse_options()?;
 
@@ -53,22 +52,18 @@ fn main() -> SydResult<ExitCode> {
     // Close all file descriptors, except:
     // 1. Standard input, output, and error.
     // 2. The PID fd and the PTY fd passed by the Syd process.
-    let max_fd = std::cmp::max(opts.fpid.as_raw_fd(), opts.fpty.as_raw_fd());
-    for fd in 3..=max_fd {
-        if fd == opts.fpid.as_raw_fd() {
-            continue;
-        }
-
-        if fd == opts.fpty.as_raw_fd() {
-            continue;
-        }
-
-        // SAFETY: Nothing can be done on close errors.
-        let _ = close(fd);
+    // Nothing can be done on closeexcept errors.
+    // We do it early here so FD randomization doesn't effect performance.
+    #[allow(clippy::cast_sign_loss)]
+    {
+        let fd1 = opts.fpid.as_raw_fd() as libc::c_uint;
+        let fd2 = opts.fpty.as_raw_fd() as libc::c_uint;
+        let _ = closeexcept(&if fd1 < fd2 {
+            [0, 1, 2, fd1, fd2]
+        } else {
+            [0, 1, 2, fd2, fd1]
+        });
     }
-    // SAFETY: nix does not have a close_range wrapper yet.
-    // Nothing can be done on close_range errors.
-    let _ = unsafe { libc::syscall(libc::SYS_close_range, max_fd + 1, RawFd::MAX, 0) };
 
     let PtyOpts {
         fpid,
@@ -109,7 +104,7 @@ fn main() -> SydResult<ExitCode> {
     //
     // SAFETY: Randomize pipe fds for hardening.
     let (pipe_pty_rd, pipe_pty_wr) = {
-        let (rd, wr) = pipe2(OFlag::O_NONBLOCK | OFlag::O_CLOEXEC)?;
+        let (rd, wr) = pipe2(OFlag::O_DIRECT | OFlag::O_NONBLOCK | OFlag::O_CLOEXEC)?;
         let rd = duprand(rd.as_raw_fd(), OFlag::O_CLOEXEC).map(|fd| {
             // SAFETY: duprand returns a valid FD on success.
             unsafe { OwnedFd::from_raw_fd(fd) }
@@ -121,7 +116,7 @@ fn main() -> SydResult<ExitCode> {
         (rd, wr)
     };
     let (pipe_std_rd, pipe_std_wr) = {
-        let (rd, wr) = pipe2(OFlag::O_NONBLOCK | OFlag::O_CLOEXEC)?;
+        let (rd, wr) = pipe2(OFlag::O_DIRECT | OFlag::O_NONBLOCK | OFlag::O_CLOEXEC)?;
         let rd = duprand(rd.as_raw_fd(), OFlag::O_CLOEXEC).map(|fd| {
             // SAFETY: duprand returns a valid FD on success.
             unsafe { OwnedFd::from_raw_fd(fd) }
@@ -143,6 +138,9 @@ fn main() -> SydResult<ExitCode> {
         unsafe { OwnedFd::from_raw_fd(fd) }
     })?;
 
+    // SAFETY: Set PTY to exclusive mode to harden against sniffing.
+    set_exclusive(&fpty, true)?;
+
     // Set PTY fd non-blocking.
     set_nonblock(&fpty, true)?;
 
@@ -275,8 +273,8 @@ fn run_pty_forwarder<
     };
     epoll_ctl_safe(&epoll.0, sig_fd.as_fd().as_raw_fd(), Some(event))?;
 
-    // TODO: MAX_EVENTS=64 move to config.rs
-    let mut events = [EpollEvent::empty(); 64];
+    // TODO: MAX_EVENTS=1024 move to config.rs
+    let mut events = [EpollEvent::empty(); 1024];
     loop {
         // Wait for events and handle EINTR.
         let n = match epoll.wait(&mut events, PollTimeout::NONE) {
@@ -285,26 +283,28 @@ fn run_pty_forwarder<
             Err(errno) => return Err(errno.into()),
         };
 
-        let mut done = false; // Handle EOF gracefully end of loop.
+        let mut is_syd = false; // Handle Syd exited?
         'eventloop: for event in events.iter().take(n) {
             let fd = event.data() as RawFd;
-            let event_flags = event.events();
+            let mut event_flags = event.events();
+
+            let is_inp = event_flags
+                .contains(EpollFlags::EPOLLIN)
+                .then(|| event_flags.remove(EpollFlags::EPOLLIN))
+                .is_some();
+            let is_out = event_flags
+                .contains(EpollFlags::EPOLLOUT)
+                .then(|| event_flags.remove(EpollFlags::EPOLLOUT))
+                .is_some();
+            let is_err = !event_flags.is_empty();
 
             if fd == pid_fd.as_fd().as_raw_fd() {
                 // Syd exited, exit gracefully.
-                done = true;
+                is_syd = true;
                 continue 'eventloop;
             }
 
-            // Exit on any hang-up or error.
-            if event_flags
-                .intersects(EpollFlags::EPOLLERR | EpollFlags::EPOLLHUP | EpollFlags::EPOLLRDHUP)
-            {
-                done = true;
-                continue 'eventloop;
-            }
-
-            if fd == sig_fd.as_raw_fd() {
+            if is_inp && fd == sig_fd.as_raw_fd() {
                 // Handle window resize event.
                 loop {
                     let sig_info = match sig_fd.read_signal() {
@@ -327,21 +327,19 @@ fn run_pty_forwarder<
                 }
             }
 
-            if event_flags.contains(EpollFlags::EPOLLIN) {
+            if is_inp {
+                // Handle readable events.
                 if fd == std_rd.as_fd().as_raw_fd() {
                     // splice from STDIN into PTY via pipe1.
-                    if splice_move(std_rd, pty_fd, pipe_pty_rd, pipe_pty_wr)? {
-                        done = true;
-                    }
+                    splice_move(std_rd, pty_fd, pipe_pty_rd, pipe_pty_wr)?;
                 } else if fd == pty_fd.as_fd().as_raw_fd() {
                     // splice from PTY into STDOUT via pipe2.
-                    if splice_move(pty_fd, std_wr, pipe_std_rd, pipe_std_wr)? {
-                        done = true;
-                    }
+                    splice_move(pty_fd, std_wr, pipe_std_rd, pipe_std_wr)?;
                 }
             }
 
-            if event_flags.contains(EpollFlags::EPOLLOUT) {
+            if is_out {
+                // Handle writable events.
                 if fd == std_wr.as_fd().as_raw_fd() {
                     // splice from pipe2 into STDOUT.
                     splice_pipe(pipe_std_rd, std_wr)?;
@@ -350,10 +348,21 @@ fn run_pty_forwarder<
                     splice_pipe(pipe_pty_rd, pty_fd)?;
                 }
             }
+
+            if is_err {
+                // Drain other side on error.
+                if fd == std_wr.as_fd().as_raw_fd() {
+                    // splice from pipe1 into PTY.
+                    splice_pipe(pipe_pty_rd, pty_fd)?;
+                } else if fd == pty_fd.as_fd().as_raw_fd() {
+                    // splice from pipe2 into STDOUT.
+                    splice_pipe(pipe_std_rd, std_wr)?;
+                }
+            }
         }
 
-        if done {
-            // Handle EOF gracefully.
+        if is_syd {
+            // Handle Syd exit gracefully.
             break;
         }
     }
@@ -539,7 +548,7 @@ fn splice_pipe<Fd1: AsFd, Fd2: AsFd>(src: Fd1, dst: Fd2) -> Result<(), Errno> {
         return match splice_data(&src, &dst) {
             Ok(0) | Err(Errno::EAGAIN) => Ok(()),
             Ok(_) | Err(Errno::EINTR) => continue,
-            Err(errno) => Err(errno.into()),
+            Err(errno) => Err(errno),
         };
     }
 }
@@ -556,7 +565,7 @@ fn splice_move<Fd1: AsFd, Fd2: AsFd, Fd3: AsFd, Fd4: AsFd>(
             Ok(_) => splice_pipe(&pipe_rd, &dst)?,
             Err(Errno::EINTR) => {}
             Err(Errno::EAGAIN) => return Ok(false),
-            Err(errno) => return Err(errno.into()),
+            Err(errno) => return Err(errno),
         }
     }
 }
