@@ -10,9 +10,11 @@ use std::{
     collections::HashMap,
     env,
     os::fd::{AsFd, AsRawFd, FromRawFd, OwnedFd, RawFd},
-    sync::{Arc, RwLock},
+    sync::{
+        atomic::{AtomicBool, Ordering},
+        Arc, RwLock,
+    },
     thread,
-    thread::JoinHandle,
 };
 
 use libseccomp::{scmp_cmp, ScmpAction, ScmpFilterContext, ScmpSyscall};
@@ -25,7 +27,7 @@ use nix::{
 use crate::{
     compat::ftruncate64,
     config::*,
-    err::{err2no, SydResult},
+    err::{err2no, SydJoinHandle, SydResult},
     error,
     fs::{lock_fd, retry_on_eintr, retry_on_intr, seal_memfd, FileInfo},
     hash::{
@@ -69,7 +71,7 @@ pub(crate) struct AesVal {
     pub(crate) spawned: bool,
 }
 
-pub(crate) type AesMap = Arc<RwLock<(HashMap<XPathBuf, AesVal, SydRandomState>, bool)>>;
+pub(crate) type AesMap = Arc<RwLock<HashMap<XPathBuf, AesVal, SydRandomState>>>;
 
 #[derive(Clone)]
 pub(crate) struct AesWorker {
@@ -77,20 +79,28 @@ pub(crate) struct AesWorker {
     files: AesMap,
     flags: Flags,
     is_memfd: bool,
+    should_exit: Arc<AtomicBool>,
 }
 
 impl AesWorker {
-    pub(crate) fn new(fdalg: (RawFd, RawFd), files: AesMap, flags: Flags, is_memfd: bool) -> Self {
+    pub(crate) fn new(
+        fdalg: (RawFd, RawFd),
+        files: AesMap,
+        flags: Flags,
+        is_memfd: bool,
+        should_exit: Arc<AtomicBool>,
+    ) -> Self {
         Self {
             fdalg,
             files,
             flags,
             is_memfd,
+            should_exit,
         }
     }
 
     #[allow(clippy::cognitive_complexity)]
-    pub(crate) fn try_spawn(self) -> Result<JoinHandle<()>, Errno> {
+    pub(crate) fn try_spawn(self) -> Result<SydJoinHandle<()>, Errno> {
         thread::Builder::new()
             .name("syd_aes".to_string())
             .stack_size(AES_STACK_SIZE)
@@ -103,12 +113,7 @@ impl AesWorker {
                     std::panic::set_hook(Box::new(|_| {}));
                 }
 
-                // SAFETY: Panic if we cannot prepare the seccomp filter.
-                #[allow(clippy::disallowed_methods)]
-                let ctx =
-                    Self::prepare_confine(self.flags).expect("prepare AES thread confinement");
-
-                #[allow(clippy::disallowed_methods)]
+                let ctx = Self::prepare_confine(self.flags)?;
                 if !dry_run {
                     // SAFETY: Logging will kill us after seccomp.
                     let safe_setid = self
@@ -117,21 +122,24 @@ impl AesWorker {
                     info!("ctx": "confine", "op": "confine_aes_thread",
                         "msg": format!("AES thread confined with{} SROP mitigation",
                             if safe_setid { "out" } else { "" }));
-
-                    // SAFETY: Panic if we cannot confine the thread.
-                    ctx.load().expect("confine AES thread");
+                    ctx.load()?;
                 } else {
                     error!("ctx": "confine", "op": "confine_aes_thread",
                         "msg": "AES threads are running unconfined in debug mode");
                 }
 
                 // Enter main loop.
-                Self::main(self.fdalg, self.files, self.is_memfd)
+                Self::main(self.fdalg, self.files, self.is_memfd, self.should_exit)
             })
             .map_err(|err| err2no(&err))
     }
 
-    fn main(setup_fds: (RawFd, RawFd), files: AesMap, is_memfd: bool) {
+    fn main(
+        fdalg: (RawFd, RawFd),
+        files: AesMap,
+        is_memfd: bool,
+        should_exit: Arc<AtomicBool>,
+    ) -> SydResult<()> {
         let mut paths = vec![];
         let mut threads = vec![];
         loop {
@@ -142,8 +150,7 @@ impl AesWorker {
             // we ensure fstat() requests keep working
             // until we're done writing.
             let my_files = files.read().unwrap_or_else(|err| err.into_inner());
-            let finished = my_files.1;
-            for (path, v) in &my_files.0 {
+            for (path, v) in my_files.iter() {
                 if !v.spawned {
                     paths.push((v.crypt_fd, path.clone()));
                 }
@@ -151,7 +158,7 @@ impl AesWorker {
             drop(my_files);
 
             if paths.is_empty() {
-                if finished {
+                if should_exit.load(Ordering::Relaxed) {
                     // main-thread signaled exit.
                     break;
                 } else {
@@ -164,7 +171,7 @@ impl AesWorker {
             // Mark entries for which we're spawning a thread.
             let mut my_files = files.write().unwrap_or_else(|err| err.into_inner());
             for (_, path) in &paths {
-                if let Some(info) = my_files.0.get_mut(path) {
+                if let Some(info) = my_files.get_mut(path) {
                     info.spawned = true;
                 }
             }
@@ -172,35 +179,30 @@ impl AesWorker {
 
             for (crypt_fd, crypt_path) in paths.drain(..) {
                 // Spawn a thread to handle the write.
-                threads.push(Self::spawn(
-                    setup_fds,
-                    crypt_fd,
-                    &crypt_path,
-                    &files,
-                    is_memfd,
-                ));
+                threads.push(Self::spawn(fdalg, &files, is_memfd, crypt_fd, &crypt_path)?);
             }
 
             // wait for a cycle.
             std::thread::sleep(AES_CYCLE_TIME);
         }
 
-        // wait for the ongoing encryption operations before exiting.
-        #[allow(clippy::disallowed_methods)]
+        // Wait for the ongoing encryption
+        // operations before exiting.
         for thread in threads {
-            thread.join().unwrap().unwrap();
+            let _ = thread.join();
         }
+
+        Ok(())
     }
 
     fn spawn(
-        setup_fds: (RawFd, RawFd),
-        crypt_fd: RawFd,
-        crypt_path: &XPath,
+        fdalg: (RawFd, RawFd),
         files: &AesMap,
         memfd: bool,
-    ) -> JoinHandle<Result<(), Errno>> {
-        #[allow(clippy::disallowed_methods)]
-        retry_on_intr(|| {
+        crypt_fd: RawFd,
+        crypt_path: &XPath,
+    ) -> SydResult<SydJoinHandle<()>> {
+        let handle = retry_on_intr(|| {
             let crypt_path = XPathBuf::from(crypt_path);
             let files = Arc::clone(files);
 
@@ -217,12 +219,12 @@ impl AesWorker {
                     lock_fd(&crypt_fd, true, true)?;
 
                     // All good, sync contents to disk.
-                    let result = Self::sync(setup_fds, &crypt_fd, &crypt_path, &files, memfd);
+                    let result = Self::sync(fdalg, &files, memfd, &crypt_fd, &crypt_path);
 
                     // Safe to remove file entry now.
                     let mut files = files.write().unwrap_or_else(|err| err.into_inner());
                     #[allow(clippy::disallowed_methods)]
-                    files.0.remove(&crypt_path).map(drop).unwrap();
+                    files.remove(&crypt_path).map(drop).unwrap();
 
                     // Close the encrypted FD.
                     drop(crypt_fd);
@@ -230,29 +232,30 @@ impl AesWorker {
                     result
                 })
                 .map_err(|err| err2no(&err))
-        })
-        .expect("spawn syd_aes")
+        })?;
+
+        Ok(handle)
     }
 
     #[allow(clippy::arithmetic_side_effects)]
     #[allow(clippy::cognitive_complexity)]
     #[allow(clippy::disallowed_methods)]
     fn sync<Fd: AsFd>(
-        setup_fds: (RawFd, RawFd),
-        crypt_fd: Fd,
-        crypt_path: &XPath,
+        fdalg: (RawFd, RawFd),
         files: &AesMap,
         memfd: bool,
-    ) -> Result<(), Errno> {
+        crypt_fd: Fd,
+        crypt_path: &XPath,
+    ) -> SydResult<()> {
         // Seal memfd to ensure no further writes happen.
         if memfd {
             seal_memfd(&crypt_fd).unwrap();
         }
 
-        let (aes_fd, mac_fd) = setup_fds;
+        let (aes_fd, mac_fd) = fdalg;
         let (enc_fd, file_mode, mut iv) = {
             let mut files = files.write().unwrap_or_else(|err| err.into_inner());
-            let v = files.0.get_mut(crypt_path).unwrap();
+            let v = files.get_mut(crypt_path).unwrap();
             #[allow(clippy::disallowed_methods)]
             (v.enc_fd.take().unwrap(), v.mode, v.iv.take().unwrap())
         };
@@ -363,7 +366,7 @@ impl AesWorker {
                         )
                     })?;
                     if n == 0 {
-                        return Err(Errno::EBADMSG);
+                        return Err(Errno::EBADMSG.into());
                     }
                     ncopy -= n;
                 }
@@ -398,10 +401,10 @@ impl AesWorker {
             while nwrite < buf.len() {
                 #[allow(clippy::arithmetic_side_effects)]
                 match write(&enc_fd, &buf[nwrite..]) {
-                    Ok(0) => return Err(Errno::EINVAL),
+                    Ok(0) => return Err(Errno::EINVAL.into()),
                     Ok(n) => nwrite += n,
                     Err(Errno::EINTR) => continue,
-                    Err(errno) => return Err(errno),
+                    Err(errno) => return Err(errno.into()),
                 }
             }
 
@@ -416,10 +419,10 @@ impl AesWorker {
             while nwrite < hmac_placeholder.len() {
                 #[allow(clippy::arithmetic_side_effects)]
                 match write(&enc_fd, &hmac_placeholder[nwrite..]) {
-                    Ok(0) => return Err(Errno::EINVAL),
+                    Ok(0) => return Err(Errno::EINVAL.into()),
                     Ok(n) => nwrite += n,
                     Err(Errno::EINTR) => continue,
-                    Err(errno) => return Err(errno),
+                    Err(errno) => return Err(errno.into()),
                 }
             }
 
@@ -429,10 +432,10 @@ impl AesWorker {
             while nwrite < buf.len() {
                 #[allow(clippy::arithmetic_side_effects)]
                 match write(&enc_fd, &buf[nwrite..]) {
-                    Ok(0) => return Err(Errno::EINVAL),
+                    Ok(0) => return Err(Errno::EINVAL.into()),
                     Ok(n) => nwrite += n,
                     Err(Errno::EINTR) => continue,
-                    Err(errno) => return Err(errno),
+                    Err(errno) => return Err(errno.into()),
                 }
             }
         }
@@ -476,7 +479,7 @@ impl AesWorker {
                     )
                 })?;
                 if n == 0 {
-                    return Err(Errno::EBADMSG);
+                    return Err(Errno::EBADMSG.into());
                 }
                 ncopy -= n;
             }
@@ -496,7 +499,7 @@ impl AesWorker {
                     )
                 })?;
                 if n == 0 {
-                    return Err(Errno::EBADMSG);
+                    return Err(Errno::EBADMSG.into());
                 }
 
                 // Duplicate data from encryption pipe to the MAC pipe using tee(2).
@@ -507,7 +510,7 @@ impl AesWorker {
                         tee(&pipe_rd_enc, &pipe_wr_mac, ntee_size, SpliceFFlags::empty())
                     })?;
                     if n_tee == 0 {
-                        return Err(Errno::EBADMSG);
+                        return Err(Errno::EBADMSG.into());
                     }
                     ntee -= n_tee;
                 }
@@ -526,7 +529,7 @@ impl AesWorker {
                         )
                     })?;
                     if n == 0 {
-                        return Err(Errno::EBADMSG);
+                        return Err(Errno::EBADMSG.into());
                     }
                     ncopy -= n;
                     nflush -= n;
@@ -546,7 +549,7 @@ impl AesWorker {
                         )
                     })?;
                     if n == 0 {
-                        return Err(Errno::EBADMSG);
+                        return Err(Errno::EBADMSG.into());
                     }
                     ncopy_mac -= n;
                 }
@@ -571,7 +574,7 @@ impl AesWorker {
                 )
             })?;
             if n == 0 {
-                return Err(Errno::EBADMSG);
+                return Err(Errno::EBADMSG.into());
             }
 
             // Duplicate data from encryption pipe to the MAC pipe using tee(2).
@@ -582,7 +585,7 @@ impl AesWorker {
                     tee(&pipe_rd_enc, &pipe_wr_mac, ntee_size, SpliceFFlags::empty())
                 })?;
                 if n_tee == 0 {
-                    return Err(Errno::EBADMSG);
+                    return Err(Errno::EBADMSG.into());
                 }
                 ntee -= n_tee;
             }
@@ -601,7 +604,7 @@ impl AesWorker {
                     )
                 })?;
                 if n == 0 {
-                    return Err(Errno::EBADMSG);
+                    return Err(Errno::EBADMSG.into());
                 }
                 ncopy -= n;
                 nflush -= n;
@@ -621,7 +624,7 @@ impl AesWorker {
                     )
                 })?;
                 if n == 0 {
-                    return Err(Errno::EBADMSG);
+                    return Err(Errno::EBADMSG.into());
                 }
                 ncopy_mac -= n;
             }
@@ -644,10 +647,10 @@ impl AesWorker {
         while nwrite < buf.len() {
             #[allow(clippy::arithmetic_side_effects)]
             match write(&enc_fd, &buf[nwrite..]) {
-                Ok(0) => return Err(Errno::EINVAL),
+                Ok(0) => return Err(Errno::EINVAL.into()),
                 Ok(n) => nwrite += n,
                 Err(Errno::EINTR) => continue,
-                Err(errno) => return Err(errno),
+                Err(errno) => return Err(errno.into()),
             }
         }
 
