@@ -19,7 +19,6 @@ use std::{
         Arc, RwLock,
     },
     thread,
-    thread::JoinHandle,
 };
 
 use libseccomp::{scmp_cmp, RawSyscall, ScmpAction, ScmpFilterContext, ScmpSyscall};
@@ -37,11 +36,12 @@ use crate::cookie::FTRUNCATE64_COOKIE_ARG3;
 #[cfg(target_arch = "x86")]
 use crate::cookie::TRUNCATE64_COOKIE_ARG3;
 use crate::{
+    alert,
     compat::epoll_ctl_safe,
     config::*,
     cookie::{
         FTRUNCATE64_COOKIE_ARG4, FTRUNCATE64_COOKIE_ARG5, FTRUNCATE_COOKIE_ARG2,
-        FTRUNCATE_COOKIE_ARG3, FTRUNCATE_COOKIE_ARG4, FTRUNCATE_COOKIE_ARG5,
+        FTRUNCATE_COOKIE_ARG3, FTRUNCATE_COOKIE_ARG4, FTRUNCATE_COOKIE_ARG5, LINKAT_COOKIE_ARG5,
         MEMFD_CREATE_COOKIE_ARG2, MEMFD_CREATE_COOKIE_ARG3, MEMFD_CREATE_COOKIE_ARG4,
         MEMFD_CREATE_COOKIE_ARG5, OPENAT2_COOKIE_ARG4, OPENAT2_COOKIE_ARG5, RENAMEAT2_COOKIE_ARG5,
         SECCOMP_IOCTL_NOTIF_ADDFD_COOKIE_ARG3, SECCOMP_IOCTL_NOTIF_ADDFD_COOKIE_ARG4,
@@ -52,7 +52,8 @@ use crate::{
         TRUNCATE_COOKIE_ARG4, TRUNCATE_COOKIE_ARG5, UNLINKAT_COOKIE_ARG3, UNLINKAT_COOKIE_ARG4,
         UNLINKAT_COOKIE_ARG5,
     },
-    err::{err2no, SydResult},
+    crit,
+    err::{err2no, scmp2no, SydJoinHandle, SydResult},
     error, extend_ioctl,
     fs::{
         seccomp_notify_respond, AT_EXECVE_CHECK, SECCOMP_IOCTL_NOTIF_ADDFD,
@@ -73,7 +74,6 @@ const EOWNERDEAD: i32 = -libc::EOWNERDEAD;
 #[derive(Clone)]
 pub(crate) struct Worker {
     fd: RawFd,
-    child: Pid,
     flags: Flags,
     epoll: Arc<Epoll>,
     cache: Arc<WorkerCache<'static>>,
@@ -101,13 +101,11 @@ impl Worker {
         crypt_map: Option<AesMap>,
     ) -> Self {
         let my_sandbox = SandboxGuard::Read(sandbox.read().unwrap_or_else(|err| err.into_inner()));
-        let child = my_sandbox.get_child_pid();
         let flags = *my_sandbox.flags;
         drop(my_sandbox); // release the read lock.
 
         Worker {
             fd,
-            child,
             flags,
             epoll,
             cache,
@@ -122,40 +120,55 @@ impl Worker {
     }
 
     #[allow(clippy::cognitive_complexity)]
-    pub(crate) fn try_spawn(self, mut confined: bool) -> Result<JoinHandle<()>, Errno> {
+    pub(crate) fn try_spawn(self, mut confined: bool) -> Result<SydJoinHandle<()>, Errno> {
         thread::Builder::new()
             .name("syd_emu".to_string())
             .stack_size(EMU_STACK_SIZE)
             .spawn(move || {
-                // Unshare CLONE_FS so cwd and umask are per-thread.
+                // Set a logging panic hook. The default panic
+                // hook calls system calls not permitted by emulators
+                // such as getcwd(2), stat(2) etc.
+                std::panic::set_hook(Box::new(|info| {
+                    let err = match info.payload().downcast_ref::<&'static str>() {
+                        Some(s) => *s,
+                        None => match info.payload().downcast_ref::<String>() {
+                            Some(s) => &**s,
+                            None => "?",
+                        },
+                    };
+
+                    let file = info.location().map(|l| l.file());
+                    let line = info.location().map(|l| l.line());
+
+                    crit!("ctx": "panic", "op": "panic_emulator_thread",
+                        "msg": err, "file": file, "line": line);
+                }));
+
+                // Unshare:
+                // 1. CLONE_FS so cwd and umask are per-thread.
+                // 2. CLONE_FILES so file descriptor table is per-thread.
                 //
-                // SAFETY: We unwrap here and crash the whole process,
-                // if this fails as this unsharing is a hard dependency.
-                #[allow(clippy::disallowed_methods)]
-                unshare(CloneFlags::CLONE_FS).expect("unshare(CLONE_FS)");
+                // Note, we cannot do 2 if Crypt sanboxing is on because
+                // emulator threads have to share memory fds with AES
+                // threads.
+                let mut unshare_flags = CloneFlags::CLONE_FS;
+                let is_crypt = self.crypt_map.is_some();
+                if !is_crypt {
+                    unshare_flags.insert(CloneFlags::CLONE_FILES);
+                }
+
+                // SAFETY: We use exit_group(2) here to bail,
+                // because this unsharing is a critical safety feature.
+                if let Err(errno) = unshare(unshare_flags) {
+                    alert!("ctx": "boot", "op": "unshare_emulator_thread",
+                        "msg": format!("failed to unshare({unshare_flags:?}): {errno}"),
+                        "err": errno as i32);
+                    std::process::exit(101);
+                }
 
                 // Create sentinel, that will handle graceful teardown.
                 let mut sentinel = Sentinel::new(&self);
 
-                let dry_run = env::var_os(ENV_SKIP_SCMP).is_some() || ExportMode::from_env().is_some();
-                if !dry_run {
-                    // Set a logging panic hook. The default panic
-                    // hook calls system calls not permitted by emulators
-                    // such as getcwd(2), stat(2) etc.
-                    std::panic::set_hook(Box::new(|info| {
-                        let err = match info.payload().downcast_ref::<&'static str>() {
-                            Some(s) => *s,
-                            None => match info.payload().downcast_ref::<String>() {
-                                Some(s) => &**s,
-                                None => "?",
-                            },
-                        };
-                        let file = info.location().map(|l| l.file());
-                        let line = info.location().map(|l| l.line());
-                        error!("ctx": "panic", "msg": err, "file": file, "line": line);
-                    }));
-                }
-
                 // Thread successfully started, increment total worker count.
                 let worker_count = self.worker_data.increment_worker_total();
 
@@ -181,16 +194,27 @@ impl Worker {
                     }
                 }
 
+                let dry_run = env::var_os(ENV_SKIP_SCMP).is_some() || ExportMode::from_env().is_some();
                 loop {
                     // SAFETY: Confine if/once locked.
                     if !confined && Sandbox::locked_once() {
-                        // SAFETY: Panic if we cannot prepare the seccomp filter.
-                        #[allow(clippy::disallowed_methods)]
-                        let ctx = Self::prepare_confine(
+                        // SAFETY: We use exit_group(2) here to bail,
+                        // because this confinement is a critical safety feature.
+                        let ctx = match Self::prepare_confine(
                             self.fd,
                             self.epoll.0.as_raw_fd(),
                             self.flags,
-                        ).expect("prepare emulator thread confinement");
+                            is_crypt,
+                        ) {
+                            Ok(ctx) => ctx,
+                            Err(error) => {
+                                let errno = error.errno().unwrap_or(Errno::ENOSYS);
+                                alert!("ctx": "boot", "op": "confine_emulator_thread",
+                                    "msg": format!("failed to confine: {error}"),
+                                    "err": errno as i32);
+                                std::process::exit(101);
+                            }
+                        };
 
                         // Honour dry-run when exporting.
                         if !dry_run {
@@ -199,9 +223,15 @@ impl Worker {
                                 "msg": format!("emulator thread confined with{} SROP mitigation",
                                     if safe_setid { "out" } else { "" }));
 
-                            // SAFETY: Panic if we cannot confine the thread.
-                            #[allow(clippy::disallowed_methods)]
-                            ctx.load().expect("confine emulator thread");
+                            // SAFETY: We use exit_group(2) here to bail,
+                            // because this confinement is a critical safety feature.
+                            if let Err(error) = ctx.load() {
+                                let errno = scmp2no(&error).unwrap_or(Errno::ENOSYS);
+                                alert!("ctx": "boot", "op": "confine_emulator_thread",
+                                    "msg": format!("failed to confine: {error}"),
+                                    "err": errno as i32);
+                                std::process::exit(101);
+                            }
                         } else {
                             error!("ctx": "confine", "op": "confine_emulator_thread",
                                 "msg": "emulator thread is running unconfined in debug mode");
@@ -241,6 +271,8 @@ impl Worker {
                         self.worker_data.decrement_worker_busy();
                     } // else process died-midway, continue.
                 }
+
+                Ok(())
             }).map_err(|err| err2no(&err))
     }
 
@@ -383,35 +415,6 @@ impl Worker {
                         );
                     };
                 }
-                Ok(1) => {
-                    // Record retired PID and try again.
-                    // These fds are added with EPOLLONESHOT.
-                    #[allow(clippy::cast_possible_truncation)]
-                    let pid = Pid::from_raw(events[0].data() as libc::pid_t);
-
-                    // Retire the PidFd.
-                    if let Some(map) = PIDFD_MAP.get() {
-                        map.del_pidfd(pid);
-                    }
-
-                    if !self.flags.contains(Flags::FL_EXIT_WAIT_ALL) && pid == self.child {
-                        // Note, threads blocked on epoll_wait will not
-                        // wake up even if we close the epoll fd or
-                        // delete the seccomp fd from epoll wait-list here.
-                        // That said, they'll never ever wake up again,
-                        // and therefore will not consume system resources.
-                        let _ = epoll_ctl_safe(&self.epoll.0, self.fd, None);
-                        let _ = close(self.fd);
-
-                        // Inform the monitor thread to exit.
-                        self.should_exit.store(true, Ordering::Relaxed);
-
-                        // Exiting with the eldest process.
-                        return Err(Errno::ESRCH);
-                    }
-
-                    // If we're waiting for all processes, let's just try again.
-                }
                 Ok(n) => unreachable!("BUG: epoll_wait returned ${n} unexpectedly!"),
                 Err(errno) => return Err(errno),
             };
@@ -424,6 +427,7 @@ impl Worker {
         seccomp_fd: RawFd,
         epoll_fd: RawFd,
         flags: Flags,
+        is_crypt: bool,
     ) -> SydResult<ScmpFilterContext> {
         // Create seccomp filter with default action.
         let mut ctx = ScmpFilterContext::new(ScmpAction::KillProcess)?;
@@ -476,6 +480,7 @@ impl Worker {
         #[allow(clippy::cast_sign_loss)]
         match ScmpSyscall::from_name(sysname) {
             Ok(syscall) => {
+                #[allow(clippy::useless_conversion)]
                 for ioctl_request in SECCOMP_IOCTL_NOTIF_LIST {
                     let mut rules = vec![scmp_cmp!($arg0 == seccomp_fd as u64)];
 
@@ -484,7 +489,6 @@ impl Worker {
                     // and aims steal file descriptors.
                     //
                     // Randomizing the seccomp-fd at startup is another mitigation against this.
-                    #[allow(clippy::useless_conversion)]
                     if restrict_cookie && *ioctl_request == SECCOMP_IOCTL_NOTIF_ADDFD {
                         rules.extend(&[
                             scmp_cmp!($arg3 == (*SECCOMP_IOCTL_NOTIF_ADDFD_COOKIE_ARG3).into()),
@@ -499,7 +503,6 @@ impl Worker {
                     // response in order to pass-through a system call to the host Linux kernel.
                     //
                     // Randomizing the seccomp-fd at startup is another mitigation against this.
-                    #[allow(clippy::useless_conversion)]
                     if restrict_cookie && *ioctl_request == SECCOMP_IOCTL_NOTIF_SEND {
                         rules.extend(&[
                             scmp_cmp!($arg3 == (*SECCOMP_IOCTL_NOTIF_SEND_COOKIE_ARG3).into()),
@@ -508,10 +511,10 @@ impl Worker {
                         ]);
                     }
 
-                    rules.push(scmp_cmp!($arg1 == *ioctl_request));
+                    rules.push(scmp_cmp!($arg1 == (*ioctl_request).into()));
                     ctx.add_rule_conditional(ScmpAction::Allow, syscall, &rules)?;
 
-                    if let Some(ioctl_request) = extend_ioctl(*ioctl_request) {
+                    if let Some(ioctl_request) = extend_ioctl((*ioctl_request).into()) {
                         rules.pop();
                         rules.push(scmp_cmp!($arg1 == ioctl_request));
                         ctx.add_rule_conditional(ScmpAction::Allow, syscall, &rules)?;
@@ -542,39 +545,16 @@ impl Worker {
             }
         }
 
-        // Allow unshare(2) with CLONE_FS only.
+        // Allow unshare(2) with CLONE_FS|CLONE_FILES only.
         let sysname = "unshare";
         match ScmpSyscall::from_name(sysname) {
             Ok(syscall) => {
-                ctx.add_rule_conditional(
-                    ScmpAction::Allow,
-                    syscall,
-                    &[scmp_cmp!($arg0 == libc::CLONE_FS as u64)],
-                )?;
-            }
-            Err(_) => {
-                info!("ctx": "confine", "op": "allow_emu_syscall",
-                    "msg": format!("invalid or unsupported syscall {sysname}"));
-            }
-        }
-
-        // SAFETY: Allow kcmp(2) with KCMP_VM or KCMP_FILE only,
-        // see users of `syd::fs::is_same_vm` and `syd::proc::proc_rand_fd`.
-        let sysname = "kcmp";
-        const KCMP_FILE: u64 = 0;
-        const KCMP_VM: u64 = 1;
-        match ScmpSyscall::from_name(sysname) {
-            Ok(syscall) => {
-                ctx.add_rule_conditional(
-                    ScmpAction::Allow,
-                    syscall,
-                    &[scmp_cmp!($arg2 == KCMP_FILE)],
-                )?;
-                ctx.add_rule_conditional(
-                    ScmpAction::Allow,
-                    syscall,
-                    &[scmp_cmp!($arg2 == KCMP_VM)],
-                )?;
+                let flags = if is_crypt {
+                    libc::CLONE_FS as u64
+                } else {
+                    (libc::CLONE_FILES | libc::CLONE_FS) as u64
+                };
+                ctx.add_rule_conditional(ScmpAction::Allow, syscall, &[scmp_cmp!($arg0 == flags)])?;
             }
             Err(_) => {
                 info!("ctx": "confine", "op": "allow_emu_syscall",
@@ -587,18 +567,21 @@ impl Worker {
         // The default is to use cross memory attach with
         // process_vm_{read,write}v(2) unconditionally since
         // version 3.32.6 unless relaxed.
-        let sysname = "write";
-        if Sandbox::memory_access() == 2 {
-            // Allow write(2) access to syd::log::LOG_FD only.
-            // If logging is disabled, deny write(2) completely.
-            // Note, we do allow the use-case where LOG_FD is
-            // intentionally set to a negative value to disable
-            // logging, hence we cast directly to an u64 and not
-            // use u64::try_from!
-            let log_fd = crate::log::LOG_FD.load(std::sync::atomic::Ordering::Relaxed);
+        let log_fd = crate::log::LOG_FD.load(std::sync::atomic::Ordering::Relaxed);
+        for sysname in ["write", "writev", "fdatasync"] {
             #[allow(clippy::cast_sign_loss)]
             match ScmpSyscall::from_name(sysname) {
+                Ok(syscall) if cfg!(feature = "prof") || Sandbox::memory_access() < 2 => {
+                    // Allow write(2) globally for proc_pid_mem(5) access.
+                    ctx.add_rule(ScmpAction::Allow, syscall)?;
+                }
                 Ok(syscall) => {
+                    // Allow write(2) access to syd::log::LOG_FD only.
+                    // If logging is disabled, deny write(2) completely.
+                    // Note, we do allow the use-case where LOG_FD is
+                    // intentionally set to a negative value to disable
+                    // logging, hence we cast directly to an u64 and not
+                    // use u64::try_from!
                     ctx.add_rule_conditional(
                         ScmpAction::Allow,
                         syscall,
@@ -610,17 +593,6 @@ impl Worker {
                         "msg": format!("invalid or unsupported syscall {sysname}"));
                 }
             }
-        } else {
-            // Allow write(2) globally for proc_pid_mem(5) access.
-            match ScmpSyscall::from_name(sysname) {
-                Ok(syscall) => {
-                    ctx.add_rule(ScmpAction::Allow, syscall)?;
-                }
-                Err(_) => {
-                    info!("ctx": "confine", "op": "allow_emu_syscall",
-                        "msg": format!("invalid or unsupported syscall {sysname}"));
-                }
-            }
         }
 
         // Allow safe fcntl(2) utility calls.
@@ -639,6 +611,22 @@ impl Worker {
             }
         }
 
+        // Deny installing new signal handlers for {rt_,}sigaction(2).
+        for sysname in ["sigaction", "rt_sigaction"] {
+            let syscall = match ScmpSyscall::from_name(sysname) {
+                Ok(syscall) => syscall,
+                Err(_) => {
+                    info!("ctx": "confine", "op": "allow_emu_syscall",
+                        "msg": format!("invalid or unsupported syscall {sysname}"));
+                    continue;
+                }
+            };
+
+            // Installs a signal handler if first argument is non-NULL.
+            // We deny this case, but allow returning the current handler.
+            ctx.add_rule_conditional(ScmpAction::Allow, syscall, &[scmp_cmp!($arg1 == 0)])?;
+        }
+
         // Allow safe system calls.
         for sysname in EMU_SYSCALLS {
             match ScmpSyscall::from_name(sysname) {
@@ -783,6 +771,29 @@ impl Worker {
                 }
             }
 
+            // linkat(2) may be used only with syscall argument cookies.
+            // We also enforce PROC_FD usage as fd argument.
+            // This fd is randomized so it further raises the bar.
+            let sysname = "linkat";
+            #[allow(clippy::cast_sign_loss)]
+            #[allow(clippy::useless_conversion)]
+            match ScmpSyscall::from_name(sysname) {
+                Ok(syscall) => {
+                    ctx.add_rule_conditional(
+                        ScmpAction::Allow,
+                        syscall,
+                        &[
+                            scmp_cmp!($arg0 == PROC_FD() as u64),
+                            scmp_cmp!($arg5 == (*LINKAT_COOKIE_ARG5).into()),
+                        ],
+                    )?;
+                }
+                Err(_) => {
+                    info!("ctx": "confine", "op": "allow_emu_syscall",
+                        "msg": format!("invalid or unsupported syscall {sysname}"));
+                }
+            }
+
             // unlinkat(2) may be used only with syscall argument cookies.
             // We also prevent AT_FDCWD usage as fd argument.
             let sysname = "unlinkat";
@@ -867,6 +878,7 @@ impl Worker {
                 "truncate64",
                 "ftruncate",
                 "ftruncate64",
+                "linkat",
                 "unlinkat",
                 "openat2",
                 "socket",
