@@ -15,9 +15,8 @@ use std::{
         fd::{AsFd, AsRawFd, FromRawFd, OwnedFd, RawFd},
         unix::net::{UnixListener, UnixStream},
     },
-    sync::{Arc, RwLock},
+    sync::{atomic::Ordering, Arc, RwLock},
     thread,
-    thread::JoinHandle,
 };
 
 use libseccomp::{scmp_cmp, ScmpAction, ScmpFilterContext, ScmpSyscall};
@@ -27,6 +26,7 @@ use nix::{
     errno::Errno,
     fcntl::OFlag,
     poll::PollTimeout,
+    sched::{unshare, CloneFlags},
     sys::{
         epoll::{Epoll, EpollCreateFlags, EpollEvent, EpollFlags},
         socket::{
@@ -36,14 +36,17 @@ use nix::{
         },
         stat::{umask, Mode},
     },
-    unistd::{Gid, Uid},
+    unistd::{write, Gid, Uid},
 };
+use serde::{ser::SerializeMap, Serialize, Serializer};
 
 use crate::{
+    alert,
     compat::{epoll_ctl_mod_safe, epoll_ctl_safe},
     config::*,
-    err::{err2no, SydResult},
-    fs::duprand,
+    crit,
+    err::{err2no, scmp2no, SydJoinHandle, SydResult},
+    fs::{closeexcept, duprand, peer_cred, retry_on_eintr},
     hash::SydRandomState,
     info,
     landlock_policy::LandlockPolicy,
@@ -65,9 +68,9 @@ pub(crate) struct IpcWorker {
     // This is set to `None` once the bind is successful.
     addr: Option<XPathBuf>,
     // Epoll file descriptor
-    epoll: Option<Epoll>,
+    pub(crate) epoll: Option<Epoll>,
     // Listener socket
-    sock: Option<UnixListener>,
+    pub(crate) sock: Option<UnixListener>,
     // Reference to the Sandbox to be configured
     sandbox: Option<Arc<RwLock<Sandbox>>>,
     // Sandbox flags specified at startup.
@@ -87,9 +90,25 @@ enum PromptMode {
     Prompt,
 }
 
+impl Serialize for PromptMode {
+    fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
+    where
+        S: Serializer,
+    {
+        let s = match *self {
+            Self::NonInteractive => "non-interactive",
+            Self::Interactive => "interactive",
+            Self::Prompt => "prompt",
+        };
+        serializer.serialize_str(s)
+    }
+}
+
 /// Per-connection state: buffering incoming bytes,
 /// pending outgoing bytes, and tracking mode.
 struct Connection {
+    /// Credentials of the other end.
+    creds: UnixCredentials,
     /// Underlying UNIX-stream socket.
     stream: UnixStream,
     /// Buffered incoming bytes until a full line arrives.
@@ -122,9 +141,46 @@ impl PartialEq for Connection {
 
 impl Eq for Connection {}
 
+// Wrapper type over UnixCredentials that implements Serialize.
+struct Creds(UnixCredentials);
+
+impl Serialize for Creds {
+    fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
+    where
+        S: Serializer,
+    {
+        let mut m = serializer.serialize_map(Some(3))?;
+        m.serialize_entry("uid", &self.0.uid())?;
+        m.serialize_entry("gid", &self.0.gid())?;
+        m.serialize_entry("pid", &self.0.pid())?;
+        m.end()
+    }
+}
+
+impl Serialize for Connection {
+    fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
+    where
+        S: Serializer,
+    {
+        let mut map = serializer.serialize_map(Some(8))?;
+
+        map.serialize_entry("fd", &self.stream.as_raw_fd())?;
+        map.serialize_entry("ilen", &self.buf.len())?;
+        map.serialize_entry("icap", &self.buf.capacity())?;
+        map.serialize_entry("olen", &self.write_buf.len())?;
+        map.serialize_entry("ocap", &self.write_buf.capacity())?;
+        map.serialize_entry("mode", &self.mode)?;
+        map.serialize_entry("close", &self.should_close)?;
+        map.serialize_entry("creds", &Creds(self.creds))?;
+
+        map.end()
+    }
+}
+
 impl Connection {
-    fn new(stream: UnixStream) -> Self {
+    fn new(creds: UnixCredentials, stream: UnixStream) -> Self {
         Self {
+            creds,
             stream,
             buf: Vec::with_capacity(1024),
             write_buf: Vec::with_capacity(1024),
@@ -316,7 +372,7 @@ impl IpcWorker {
                     )?;
                 }
                 Err(_) => {
-                    info!("ctx": "confine", "op": "allow_syscall",
+                    info!("ctx": "confine", "op": "allow_ipc_syscall",
                         "msg": format!("invalid or unsupported syscall {sysname}"));
                 }
             }
@@ -325,13 +381,21 @@ impl IpcWorker {
         // Deny open and stat family with ENOSYS rather than KillProcess.
         // We need this because Rust allocator has side effects such as
         // opening /proc/sys/vm/overcommit_memory on some architectures.
-        for sysname in ["open", "openat", "stat", "lstat", "statx", "newfstatat"] {
+        for sysname in [
+            "open",
+            "openat",
+            "openat2",
+            "stat",
+            "lstat",
+            "statx",
+            "newfstatat",
+        ] {
             match ScmpSyscall::from_name(sysname) {
                 Ok(syscall) => {
                     ctx.add_rule(ScmpAction::Errno(Errno::ENOSYS as i32), syscall)?;
                 }
                 Err(_) => {
-                    info!("ctx": "confine", "op": "allow_syscall",
+                    info!("ctx": "confine", "op": "allow_ipc_syscall",
                         "msg": format!("invalid or unsupported syscall {sysname}"));
                 }
             }
@@ -342,7 +406,7 @@ impl IpcWorker {
             let syscall = match ScmpSyscall::from_name(sysname) {
                 Ok(syscall) => syscall,
                 Err(_) => {
-                    info!("ctx": "confine", "op": "allow_syscall",
+                    info!("ctx": "confine", "op": "allow_ipc_syscall",
                         "msg": format!("invalid or unsupported syscall {sysname}"));
                     continue;
                 }
@@ -353,6 +417,73 @@ impl IpcWorker {
             }
         }
 
+        // Allow safe prctl(2) utility calls.
+        let sysname = "prctl";
+        match ScmpSyscall::from_name(sysname) {
+            Ok(syscall) => {
+                // Allow prctl(PR_SET_VMA, PR_SET_VMA_ANON_NAME, ...)
+                // for use during memory sealing of the sandbox policy.
+                ctx.add_rule_conditional(
+                    ScmpAction::Allow,
+                    syscall,
+                    &[
+                        scmp_cmp!($arg0 == libc::PR_SET_VMA as u64),
+                        scmp_cmp!($arg1 == libc::PR_SET_VMA_ANON_NAME as u64),
+                    ],
+                )?;
+            }
+            Err(_) => {
+                info!("ctx": "confine", "op": "allow_ipc_syscall",
+                    "msg": format!("invalid or unsupported syscall {sysname}"));
+            }
+        }
+
+        // Prevent executable memory.
+        const PROT_EXEC: u64 = libc::PROT_EXEC as u64;
+        for sysname in ["mmap", "mmap2", "mprotect"] {
+            match ScmpSyscall::from_name(sysname) {
+                Ok(syscall) => {
+                    ctx.add_rule_conditional(
+                        ScmpAction::Allow,
+                        syscall,
+                        &[scmp_cmp!($arg2 & PROT_EXEC == 0)],
+                    )?;
+                }
+                Err(_) => {
+                    info!("ctx": "confine", "op": "allow_ipc_syscall",
+                        "msg": format!("invalid or unsupported syscall {sysname}"));
+                }
+            }
+        }
+
+        // SAFETY: Allow write(2) to syd::log::LOG_FD only.
+        // If logging is disabled, deny write(2) completely.
+        // Note, we do allow the use-case where LOG_FD is
+        // intentionally set to a negative value to disable
+        // logging, hence we cast directly to an u64 and not
+        // use u64::try_from!
+        // Exception: Allow write(2) globally if profiling is enabled.
+        let log_fd = crate::log::LOG_FD.load(Ordering::Relaxed);
+        for sysname in ["write", "writev", "fdatasync"] {
+            #[allow(clippy::cast_sign_loss)]
+            match ScmpSyscall::from_name(sysname) {
+                Ok(syscall) if cfg!(feature = "prof") => {
+                    ctx.add_rule(ScmpAction::Allow, syscall)?;
+                }
+                Ok(syscall) => {
+                    ctx.add_rule_conditional(
+                        ScmpAction::Allow,
+                        syscall,
+                        &[scmp_cmp!($arg0 == log_fd as u64)],
+                    )?;
+                }
+                Err(_) => {
+                    info!("ctx": "confine", "op": "allow_ipc_syscall",
+                        "msg": format!("invalid or unsupported syscall {sysname}"));
+                }
+            }
+        }
+
         // Allow safe system calls.
         for sysname in IPC_SYSCALLS {
             match ScmpSyscall::from_name(sysname) {
@@ -360,7 +491,7 @@ impl IpcWorker {
                     ctx.add_rule(ScmpAction::Allow, syscall)?;
                 }
                 Err(_) => {
-                    info!("ctx": "confine", "op": "allow_syscall",
+                    info!("ctx": "confine", "op": "allow_ipc_syscall",
                         "msg": format!("invalid or unsupported syscall {sysname}"));
                 }
             }
@@ -373,7 +504,7 @@ impl IpcWorker {
                     ctx.add_rule(ScmpAction::Allow, syscall)?;
                 }
                 Err(_) => {
-                    info!("ctx": "confine", "op": "allow_syscall",
+                    info!("ctx": "confine", "op": "allow_ipc_syscall",
                         "msg": format!("invalid or unsupported syscall {sysname}"));
                 }
             }
@@ -386,7 +517,7 @@ impl IpcWorker {
                     ctx.add_rule(ScmpAction::Allow, syscall)?;
                 }
                 Err(_) => {
-                    info!("ctx": "confine", "op": "allow_syscall",
+                    info!("ctx": "confine", "op": "allow_ipc_syscall",
                         "msg": format!("invalid or unsupported syscall {sysname}"));
                 }
             }
@@ -403,7 +534,7 @@ impl IpcWorker {
                         ctx.add_rule(ScmpAction::Allow, syscall)?;
                     }
                     Err(_) => {
-                        info!("ctx": "confine", "op": "allow_syscall",
+                        info!("ctx": "confine", "op": "allow_ipc_syscall",
                             "msg": format!("invalid or unsupported syscall {sysname}"));
                     }
                 }
@@ -417,7 +548,7 @@ impl IpcWorker {
                         ctx.add_rule(ScmpAction::Allow, syscall)?;
                     }
                     Err(_) => {
-                        info!("ctx": "confine", "op": "allow_syscall",
+                        info!("ctx": "confine", "op": "allow_ipc_syscall",
                             "msg": format!("invalid or unsupported syscall {sysname}"));
                     }
                 }
@@ -428,35 +559,123 @@ impl IpcWorker {
     }
 
     // Spawn the IPC worker in a new thread named `syd_ipc'.
-    pub fn try_spawn(self) -> Result<JoinHandle<()>, Errno> {
+    pub fn try_spawn(mut self, notif_pipe: (RawFd, RawFd)) -> Result<SydJoinHandle<()>, Errno> {
         thread::Builder::new()
             .name("syd_ipc".into())
             .stack_size(IPC_STACK_SIZE)
             .spawn(move || {
-                let _ = self.main();
+                self.init(notif_pipe)?;
+                self.main()
             })
             .map_err(|err| err2no(&err))
     }
 
-    /// Main loop: waits on epoll, accepts new connections, reads commands, and writes responses.
-    fn main(&self) -> Result<(), Errno> {
+    /// Initialize IPC worker.
+    fn init(&mut self, notif_pipe: (RawFd, RawFd)) -> SydResult<()> {
+        // SAFETY: Default panic hook wont work with stdio(3) closed.
+        std::panic::set_hook(Box::new(|info| {
+            let err = match info.payload().downcast_ref::<&'static str>() {
+                Some(s) => *s,
+                None => match info.payload().downcast_ref::<String>() {
+                    Some(s) => &**s,
+                    None => "?",
+                },
+            };
+
+            let file = info.location().map(|l| l.file());
+            let line = info.location().map(|l| l.line());
+
+            crit!("ctx": "panic", "op": "panic_ipc_thread",
+                "msg": err, "file": file, "line": line);
+        }));
+
+        // SAFETY: We use exit_group(2) here to bail,
+        // because this unsharing is a critical safety feature.
+        if let Err(errno) = unshare(CloneFlags::CLONE_FS | CloneFlags::CLONE_FILES) {
+            alert!("ctx": "boot", "op": "unshare_ipc_thread",
+                "msg": format!("failed to unshare(CLONE_FS|CLONE_FILES): {errno}"),
+                "err": errno as i32);
+            std::process::exit(101);
+        }
+
+        // SAFETY: notif_pipe points to valid FDs.
+        let (pipe_rd, pipe_wr) = unsafe {
+            (
+                OwnedFd::from_raw_fd(notif_pipe.0),
+                OwnedFd::from_raw_fd(notif_pipe.1),
+            )
+        };
+        drop(pipe_rd);
+        let buf = [42u8; 1];
+        match retry_on_eintr(|| write(&pipe_wr, &buf))? {
+            0 => return Err(Errno::EIO.into()),
+            1 => {}
+            n => unreachable!("BUG: invalid pipe write of size {n}!"),
+        }
+
+        // Close the notification pipe.
+        drop(pipe_wr);
+
+        // SAFETY: The IPC worker needs to inherit only the following FDs:
+        // 1. epoll(7) FD.
+        // 2. IPC socket FD.
+        // 3. Log FD.
+        // We have to sort the set as the FDs are randomized.
+        let poll = self.epoll.as_ref().ok_or(Errno::EADDRNOTAVAIL)?;
+        let sock = self.sock.as_ref().ok_or(Errno::EADDRNOTAVAIL)?;
+        #[allow(clippy::cast_sign_loss)]
+        let mut set = vec![
+            poll.0.as_raw_fd() as libc::c_uint,
+            sock.as_raw_fd() as libc::c_uint,
+            crate::log::LOG_FD.load(Ordering::Relaxed) as libc::c_uint,
+        ];
+        set.sort_unstable();
+        closeexcept(&set)?;
+
+        Ok(())
+    }
+
+    /// Main loop: waits on epoll, accepts new connections,
+    /// reads commands, and writes responses.
+    #[allow(clippy::cognitive_complexity)]
+    fn main(&self) -> SydResult<()> {
         let epoll = self.epoll.as_ref().ok_or(Errno::EADDRNOTAVAIL)?;
         let sock = self.sock.as_ref().ok_or(Errno::EADDRNOTAVAIL)?;
         let sock_fd = sock.as_fd().as_raw_fd();
 
         // Confine `syd_ipc' thread.
-        // SAFETY: Panic if we cannot confine the thread.
-        #[allow(clippy::disallowed_methods)]
-        let ctx = Self::prepare_confine(epoll.0.as_raw_fd(), self.flags, false)
-            .expect("prepare confine ipc thread");
-        #[allow(clippy::disallowed_methods)]
-        ctx.load().expect("confine ipc thread");
+        // SAFETY: We use exit_group(2) here to bail,
+        // because this confinement is a critical safety feature.
+        let ctx = match Self::prepare_confine(epoll.0.as_raw_fd(), self.flags, false) {
+            Ok(ctx) => ctx,
+            Err(error) => {
+                let errno = error.errno().unwrap_or(Errno::ENOSYS);
+                alert!("ctx": "boot", "op": "confine_ipc_thread",
+                    "msg": format!("failed to confine: {error}"),
+                    "err": errno as i32);
+                std::process::exit(101);
+            }
+        };
+
+        // Load seccomp(2) BPF into the kernel.
+        // SAFETY: We use exit_group(2) here to bail,
+        // because this confinement is a critical safety feature.
+        if let Err(error) = ctx.load() {
+            let errno = scmp2no(&error).unwrap_or(Errno::ENOSYS);
+            alert!("ctx": "boot", "op": "confine_ipc_thread",
+                "msg": format!("failed to confine: {error}"),
+                "err": errno as i32);
+            std::process::exit(101);
+        }
+        info!("ctx": "confine", "op": "confine_ipc_thread",
+            "msg": "IPC thread confined");
 
         // Set of active connections.
         let mut connections = ConnectionMap::default();
 
         // Buffer for epoll events.
-        let mut events = [EpollEvent::empty(); 128];
+        // TODO: IPC_EPOLL_MAX_EVENTS=1024 move to config.rs
+        let mut events = [EpollEvent::empty(); 1024];
 
         // Wait for events and handle EINTR.
         loop {
@@ -464,7 +683,7 @@ impl IpcWorker {
             let nfds = match epoll.wait(&mut events, PollTimeout::NONE) {
                 Ok(n) => n,
                 Err(Errno::EINTR) => continue,
-                Err(errno) => return Err(errno),
+                Err(errno) => return Err(errno.into()),
             };
 
             for ev in events.iter().take(nfds) {
@@ -520,7 +739,8 @@ impl IpcWorker {
                     epoll_ctl_safe(&epoll.0, fd, Some(event))?;
 
                     // Create Connection state.
-                    let mut conn = Connection::new(stream);
+                    let creds = peer_cred(&stream)?;
+                    let mut conn = Connection::new(creds, stream);
 
                     // In prompt mode, send initial prompt.
                     if conn.mode == PromptMode::Prompt {
@@ -543,6 +763,7 @@ impl IpcWorker {
     }
 
     /// Handle a readable event on connection `fd`.
+    #[allow(clippy::cognitive_complexity)]
     fn handle_readable(&self, connections: &mut ConnectionMap, fd: RawFd) -> Result<(), Errno> {
         // Lookup connection by fd.
         // We need a mutable reference; use get_mut.
@@ -578,10 +799,18 @@ impl IpcWorker {
                         // 2. Reject with error message.
                         // 3. Close connection after reply.
                         let epoll = self.epoll.as_ref().ok_or(Errno::EADDRNOTAVAIL)?;
-                        conn.ctl(&epoll, EpollFlags::EPOLLOUT)?;
+                        conn.ctl(epoll, EpollFlags::EPOLLOUT)?;
                         conn.enqueue_response(IPC_RATE);
                         conn.should_close = true;
 
+                        info!("ctx": "ipc", "op": "reject_ipc_command",
+                            "msg": format!("rejected IPC command from pid:{} with uid:{}/gid:{}",
+                                conn.creds.pid(),
+                                conn.creds.uid(),
+                                conn.creds.gid()),
+                            "len": conn.buf.len() + msg.bytes,
+                            "conn": &conn);
+
                         return Ok(());
                     }
                     Ok(msg) => {
@@ -633,7 +862,7 @@ impl IpcWorker {
                             // 2. Reject with error message.
                             // 3. Close connection after reply.
                             let epoll = self.epoll.as_ref().ok_or(Errno::EADDRNOTAVAIL)?;
-                            conn.ctl(&epoll, EpollFlags::EPOLLOUT)?;
+                            conn.ctl(epoll, EpollFlags::EPOLLOUT)?;
                             conn.enqueue_response(IPC_AUTH);
                             conn.should_close = true;
 
@@ -710,7 +939,7 @@ impl IpcWorker {
 
             if conn.write_pos >= conn.write_buf.len() {
                 // All data written: clear buffer and disable EPOLLOUT.
-                conn.ctl(&epoll, EpollFlags::EPOLLIN)?;
+                conn.ctl(epoll, EpollFlags::EPOLLIN)?;
                 conn.write_buf.clear();
                 conn.write_pos = 0;
 
@@ -726,6 +955,7 @@ impl IpcWorker {
     }
 
     // Process at `pos` boundary from connection `conn` on `fd`.
+    #[allow(clippy::cognitive_complexity)]
     fn process_command(&self, conn: &mut Connection, pos: usize) -> Result<(), Errno> {
         let epoll = self.epoll.as_ref().ok_or(Errno::EADDRNOTAVAIL)?;
 
@@ -798,6 +1028,14 @@ impl IpcWorker {
                 } else if conn.mode == PromptMode::NonInteractive {
                     conn.should_close = true;
                 }
+                info!("ctx": "ipc", "op": "handle_ipc_command",
+                    "msg": format!("handled IPC command from pid:{} with uid:{}/gid:{}",
+                        conn.creds.pid(),
+                        conn.creds.uid(),
+                        conn.creds.gid()),
+                    "cmd": &cmd,
+                    "resp": XPath::from_bytes(&response),
+                    "conn": &conn);
             }
             None => {
                 // Empty line: just send prompt if in prompt mode.
@@ -809,7 +1047,7 @@ impl IpcWorker {
 
         // After enqueueing data, ensure EPOLLOUT is enabled.
         if !conn.write_buf.is_empty() {
-            conn.ctl(&epoll, EpollFlags::EPOLLIN | EpollFlags::EPOLLOUT)?;
+            conn.ctl(epoll, EpollFlags::EPOLLIN | EpollFlags::EPOLLOUT)?;
         }
 
         Ok(())
@@ -843,7 +1081,7 @@ impl IpcWorker {
 
         // Intercept display commands and send to socket
         // instead of Syd's standard error:
-        // stat: Prints Sandbox in pretty-printed JSON.
+        // stat: Prints Sandbox in compact JSON.
         // stats: Prints Sandbox in human-readable format.
         if is_equal(cmd.as_bytes(), b"stat") {
             return Ok(serde_json::to_string(&*sandbox)
