@@ -10,19 +10,19 @@
 // SPDX-License-Identifier: GPL-3.0
 
 use std::{
-    collections::{btree_map::Entry, BTreeMap, HashMap},
+    collections::{hash_map::Entry, HashMap},
+    fs::File,
     option::Option,
-    os::fd::{AsRawFd, OwnedFd, RawFd},
     sync::{
         atomic::{AtomicUsize, Ordering},
-        Arc, Mutex, RwLock,
+        Arc, RwLock,
     },
 };
 
 use nix::{
     errno::Errno,
     sys::{
-        epoll::{Epoll, EpollFlags},
+        signal::{SigSet, Signal},
         socket::UnixAddr,
     },
     unistd::{gettid, Pid},
@@ -33,13 +33,9 @@ use crate::{
         signal_map_new, sys_interrupt_map_new, sys_result_map_new, ExecResult, SignalMap,
         SigreturnResult, SysInterrupt, SysInterruptMap, SysResultMap,
     },
-    compat::epoll_ctl_safe,
-    config::*,
     elf::ExecutableFile,
-    fs::{pidfd_open, seccomp_notify_id_valid, CanonicalPath, PIDFD_THREAD},
+    fs::{block_signal, retry_on_eintr, sigtimedpoll, unblock_signal, CanonicalPath},
     hash::SydRandomState,
-    hook::RemoteProcess,
-    proc::proc_tgid,
     ScmpNotifReq, SydMemoryMap, SydSigSet,
 };
 
@@ -55,10 +51,6 @@ pub(crate) mod emu;
 /// A cache for worker threads.
 #[derive(Debug)]
 pub(crate) struct WorkerCache<'a> {
-    // Shared epoll instance
-    pub(crate) poll: Arc<Epoll>,
-    // Seccomp-notify fd
-    pub(crate) scmp: RawFd,
     // Signal handlers map
     pub(crate) signal_map: SignalMap,
     // System call interrupt map
@@ -68,10 +60,8 @@ pub(crate) struct WorkerCache<'a> {
 }
 
 impl<'a> WorkerCache<'a> {
-    pub(crate) fn new(poll: Arc<Epoll>, scmp: RawFd) -> Self {
+    pub(crate) fn new() -> Self {
         Self {
-            poll,
-            scmp,
             signal_map: signal_map_new(),
             sysint_map: sys_interrupt_map_new(),
             sysres_map: sys_result_map_new(),
@@ -147,59 +137,50 @@ impl<'a> WorkerCache<'a> {
     }
 
     // Record a chdir result.
-    pub(crate) fn add_chdir<'b>(&'b self, process: RemoteProcess, path: CanonicalPath<'a>) {
+    pub(crate) fn add_chdir<'b>(&'b self, pid: Pid, path: CanonicalPath<'a>) {
         self.sysres_map
             .trace_chdir
             .lock()
             .unwrap_or_else(|err| err.into_inner())
-            .insert(process, path);
+            .insert(pid, path);
     }
 
     // Query, remove and return a chdir result.
     #[allow(clippy::type_complexity)]
-    pub(crate) fn get_chdir<'b>(&'b self, pid: Pid) -> Option<(RemoteProcess, CanonicalPath<'a>)> {
-        let p = RemoteProcess {
-            pid,
-            pid_fd: libc::AT_FDCWD,
-        };
-
+    pub(crate) fn get_chdir<'b>(&'b self, pid: Pid) -> Option<(Pid, CanonicalPath<'a>)> {
         self.sysres_map
             .trace_chdir
             .lock()
             .unwrap_or_else(|err| err.into_inner())
-            .remove_entry(&p)
+            .remove_entry(&pid)
     }
 
     // Record an error result.
-    pub(crate) fn add_error(&self, process: RemoteProcess, errno: Option<Errno>) {
+    pub(crate) fn add_error(&self, pid: Pid, errno: Option<Errno>) {
         self.sysres_map
             .trace_error
             .lock()
             .unwrap_or_else(|err| err.into_inner())
-            .insert(process, errno);
+            .insert(pid, errno);
     }
 
     // Query, remove and return a error result.
     #[allow(clippy::type_complexity)]
-    pub(crate) fn get_error(&self, pid: Pid) -> Option<(RemoteProcess, Option<Errno>)> {
-        let p = RemoteProcess {
-            pid,
-            pid_fd: libc::AT_FDCWD,
-        };
-
+    pub(crate) fn get_error(&self, pid: Pid) -> Option<(Pid, Option<Errno>)> {
         self.sysres_map
             .trace_error
             .lock()
             .unwrap_or_else(|err| err.into_inner())
-            .remove_entry(&p)
+            .remove_entry(&pid)
     }
 
     // Record a execv result.
     #[allow(clippy::too_many_arguments)]
     pub(crate) fn add_exec(
         &self,
-        process: RemoteProcess,
-        file: ExecutableFile,
+        pid: Pid,
+        exe: ExecutableFile,
+        file: File,
         arch: u32,
         ip: u64,
         sp: u64,
@@ -209,6 +190,7 @@ impl<'a> WorkerCache<'a> {
         memmap: Option<Vec<SydMemoryMap>>,
     ) {
         let result = ExecResult {
+            exe,
             file,
             arch,
             ip,
@@ -223,28 +205,23 @@ impl<'a> WorkerCache<'a> {
             .trace_execv
             .lock()
             .unwrap_or_else(|err| err.into_inner())
-            .insert(process, result);
+            .insert(pid, result);
     }
 
     // Query, remove and return a exec result.
-    pub(crate) fn get_exec(&self, pid: Pid) -> Option<(RemoteProcess, ExecResult)> {
-        let p = RemoteProcess {
-            pid,
-            pid_fd: libc::AT_FDCWD,
-        };
-
+    pub(crate) fn get_exec(&self, pid: Pid) -> Option<(Pid, ExecResult)> {
         self.sysres_map
             .trace_execv
             .lock()
             .unwrap_or_else(|err| err.into_inner())
-            .remove_entry(&p)
+            .remove_entry(&pid)
     }
 
     // Record a sigreturn entry.
     #[allow(clippy::too_many_arguments)]
     pub(crate) fn add_sigreturn(
         &self,
-        process: RemoteProcess,
+        pid: Pid,
         is_realtime: bool,
         ip: u64,
         sp: u64,
@@ -264,21 +241,16 @@ impl<'a> WorkerCache<'a> {
             .trace_sigret
             .lock()
             .unwrap_or_else(|err| err.into_inner())
-            .insert(process, result);
+            .insert(pid, result);
     }
 
     // Query, remove and return a sigreturn entry info.
-    pub(crate) fn get_sigreturn(&self, pid: Pid) -> Option<(RemoteProcess, SigreturnResult)> {
-        let p = RemoteProcess {
-            pid,
-            pid_fd: libc::AT_FDCWD,
-        };
-
+    pub(crate) fn get_sigreturn(&self, pid: Pid) -> Option<(Pid, SigreturnResult)> {
         self.sysres_map
             .trace_sigret
             .lock()
             .unwrap_or_else(|err| err.into_inner())
-            .remove_entry(&p)
+            .remove_entry(&pid)
     }
 
     // Add a restarting signal.
@@ -336,24 +308,56 @@ impl<'a> WorkerCache<'a> {
         ignore_restart: bool,
     ) -> Result<(), Errno> {
         let handler_tid = gettid();
-        let request_tgid = proc_tgid(Pid::from_raw(request.pid as libc::pid_t))?;
-        let interrupt = SysInterrupt::new(request, request_tgid, handler_tid, ignore_restart)?;
+        let interrupt = SysInterrupt::new(request, handler_tid, ignore_restart)?;
 
         let (ref lock, ref cvar) = *self.sysint_map.sys_block;
         let mut map = lock.lock().unwrap_or_else(|err| err.into_inner());
 
+        map.retain(|_, interrupt| interrupt.handler != handler_tid);
         map.insert(request.id, interrupt);
 
         cvar.notify_one();
 
-        Ok(())
+        // Discard spurious pending signals.
+        // Note, SIGALRM is only queued once unlike realtime signals,
+        // therefore we do not need a while loop here for sigtimedpoll.
+        let mut mask = SigSet::empty();
+        mask.add(Signal::SIGALRM);
+        let _ = retry_on_eintr(|| sigtimedpoll(&mask, None));
+
+        unblock_signal(Signal::SIGALRM)
     }
 
     // Remove a blocked fifo.
-    pub(crate) fn del_sys_block(&self, request_id: u64) {
-        let (ref lock, ref _cvar) = *self.sysint_map.sys_block;
-        let mut map = lock.lock().unwrap_or_else(|err| err.into_inner());
-        map.remove(&request_id);
+    pub(crate) fn del_sys_block(&self, request_id: u64, interrupted: bool) -> Result<(), Errno> {
+        block_signal(Signal::SIGALRM)?;
+
+        if !interrupted {
+            let (ref lock, ref _cvar) = *self.sysint_map.sys_block;
+            let mut map = lock.lock().unwrap_or_else(|err| err.into_inner());
+            map.remove(&request_id);
+        }
+
+        Ok(())
+    }
+
+    // Remove a PID completely from the cache.
+    pub(crate) fn del_pid(&self, pid: Pid) {
+        // Retire TGID from signal maps.
+        self.retire_sig_handle(pid);
+        self.retire_sig_restart(pid);
+
+        // Remove preexisting error record for pid.
+        let _ = self.get_error(pid);
+
+        // Remove preexisting chdir record for pid.
+        let _ = self.get_chdir(pid);
+
+        // Remove preexisting exec record for pid.
+        let _ = self.get_exec(pid);
+
+        // Remove preexisting sigreturn record for pid.
+        let _ = self.get_sigreturn(pid);
     }
 }
 
@@ -450,121 +454,3 @@ impl WorkerData {
 // or mount id so unfortunately we cannot check
 // for that here.
 pub(crate) type BindMap = Arc<RwLock<HashMap<u64, UnixAddr, SydRandomState>>>;
-
-/// PidFd map, used to store pid file descriptors.
-#[derive(Debug)]
-#[allow(clippy::type_complexity)]
-pub struct PidFdMap {
-    /// Inner PidFd concurrent dash map.
-    pub pidfd: Arc<Mutex<BTreeMap<Pid, OwnedFd>>>,
-    /// A reference to the WorkerCache to clean relevant data on process exit.
-    pub(crate) cache: Arc<WorkerCache<'static>>,
-}
-
-impl PidFdMap {
-    /// Create a new PidFd map.
-    pub(crate) fn new(cache: Arc<WorkerCache<'static>>) -> Self {
-        Self {
-            cache,
-            pidfd: Arc::new(Mutex::new(BTreeMap::new())),
-        }
-    }
-
-    /*
-    #[inline]
-    pub(crate) fn get_pidfd(&self, pid: Pid) -> Option<RawFd> {
-        self.pidfd
-            .lock()
-            .unwrap_or_else(|err| err.into_inner())
-            .get(&pid)
-            .map(|fd| fd.as_raw_fd())
-    }
-
-    #[inline]
-    pub(crate) fn add_pidfd(&self, pid: Pid, pid_fd: OwnedFd) {
-        self.pidfd
-            .lock()
-            .unwrap_or_else(|err| err.into_inner())
-            .insert(pid, pid_fd);
-    }
-    */
-
-    #[inline]
-    pub(crate) fn del_pidfd(&self, pid: Pid) {
-        // Retire TGID from signal maps.
-        self.cache.retire_sig_handle(pid);
-        self.cache.retire_sig_restart(pid);
-
-        // Remove preexisting error record for pid.
-        let _ = self.cache.get_error(pid);
-
-        // Remove preexisting chdir record for pid.
-        let _ = self.cache.get_chdir(pid);
-
-        // Remove preexisting exec record for pid.
-        let _ = self.cache.get_exec(pid);
-
-        // Remove preexisting sigreturn record for pid.
-        let _ = self.cache.get_sigreturn(pid);
-
-        // Finally, remove the PidFd from the map.
-        self.pidfd
-            .lock()
-            .unwrap_or_else(|err| err.into_inner())
-            .remove(&pid);
-    }
-
-    pub(crate) fn pidfd_open(
-        &self,
-        request_pid: Pid,
-        tgid: bool,
-        request_id: Option<u64>,
-    ) -> Result<RawFd, Errno> {
-        let mut pidfd = self.pidfd.lock().unwrap_or_else(|err| err.into_inner());
-        if let Some(fd) = pidfd.get(&request_pid) {
-            return Ok(fd.as_raw_fd());
-        }
-
-        // Use PIDFD_THREAD if available.
-        let (pid, flags) = if *HAVE_PIDFD_THREAD {
-            (request_pid, PIDFD_THREAD)
-        } else if tgid {
-            (request_pid, 0)
-        } else {
-            (proc_tgid(request_pid)?, 0)
-        };
-
-        // Open the PIDFd.
-        let pid_fd = pidfd_open(pid, flags)?;
-
-        if let Some(request_id) = request_id {
-            // SAFETY:
-            // 1. Validate the PIDFd by validating the request ID if submitted.
-            // 2. EAGAIN|EINTR is handled.
-            // 3. ENOENT means child died mid-way.
-            if seccomp_notify_id_valid(self.cache.scmp, request_id).is_err() {
-                return Err(Errno::ESRCH);
-            }
-        }
-
-        // SAFETY: Add the PIDFd to the epoll instance.
-        //
-        // Note: EPOLLEXCLUSIVE|EPOLLONESHOT is invalid!
-        #[allow(clippy::cast_sign_loss)]
-        let event = libc::epoll_event {
-            events: (EpollFlags::EPOLLIN | EpollFlags::EPOLLONESHOT).bits() as u32,
-            u64: request_pid.as_raw() as u64,
-        };
-
-        let pid_fd_raw = pid_fd.as_raw_fd();
-
-        // SAFETY: In epoll(7) we trust.
-        #[allow(clippy::disallowed_methods)]
-        epoll_ctl_safe(&self.cache.poll.0, pid_fd_raw, Some(event))
-            .expect("BUG: Failed to add PidFd to Epoll!");
-
-        pidfd.insert(request_pid, pid_fd);
-
-        Ok(pid_fd_raw)
-    }
-}
